/*
*	Prototypes of FINDER functions.
*	Functions in TR.c are not listed here.
*/

/*------ The following are in file futils.c -----------------------------*/

void get_options(int argc, char *argv[],
		 boolean *debugging,
		 char *filename,
		 char *diag_output,
		 boolean *pipe_out
		 );
void go(int x, boolean persists);
boolean clause_is_long(void);
void timesup();
void short_timesup(int x);
void intrupt();
void print_synopsis(void);
void print_with_default(char c, char *s, boolean b);



/*------ The following are in file FINDER.c -----------------------------*/

int found(const dynstring *s, int add_if_true, int seek_model);
void init_init_finder(void);
boolean init_finder(char filename[], int tolerated);
boolean initFinder(ifile i, ifile o, int tolerated);
boolean _init_finder(FILE *if_fil, ifile ifili, ifile ifilo, int tolerated);
ifile transfer_job(char *source);
void refute_current(void);
int theory_size(int x);
void finder_input(const dynstring *s);
boolean initial_model_found(void);
void set_no_semantics(void);
int no_semantics_set(void);
void set_job_only(void);
void unset_job_only(void);
void FINDER_cleanup(void);
void print_model_anyway(void);


/*------ The following are NOT in file FINDER.c -----------------------------*/

boolean model(char clause[], int in_theory, boolean test_only);
void init_theory(int theory);
void print_theory(int i);
int get_last_backtracks(void);
int get_last_default_size(int theory);
boolean get_old_optimal(int theory);
int get_clauses_modelled(int theory);
int get_failed_searches(int theory);
int get_max_backtracks_in(int theory);
int get_max_time_in(int theory);
int get_searches(int theory);
int get_size(int theory);
int get_successful_searches(int theory);
int get_tests(int theory);
int get_time_outs(void);
int get_time_outs_in_test(void);
int get_updates(int theory);
boolean equality_is_present(void);
void output_the_model(int new_theory);
void clause_to_string(clause cl, char *str);
void wff_to_str(wff w, char *str);
void set_min_prob_extn(float min_prob_extn);
void set_min_time_limit(int limit);
void set_max_time_limit(int limit);
void set_tolerance(float limit);
void set_up_last_model(void);


/*------ The following are in file initial.c ----------------------------*/

void initial_initial(void);
void initialize(void);
void canon(kind k, char new_name[]);
int lex_pos(char word[]);
kind newkind(char new_name[]);
functor newfunctor(char new_name[], syntype category, int adicity, boolean dec);
void free_tables(void);
void free_table(functor f);

/*------ The following are in file input.c ------------------------------*/

char next_char(void);
char get_a_char(void);
void get_word(dynstring *word);
void input_to_terminator(char *s);
statement_type get_key_word(void);
void read_job(FILE *infil, ifile ifil);
void check_surjections(void);
void get_incl_file(void);
int get_int(void);
float get_float(void);
void get_item(void (*item)(), kind k, functor f);
void get_sort(kind k, functor f, dynstring *name, int x);
void sort_spec(kind k, functor f, dynstring *spec, int x);
void de_enumerate(kind k);
void free_canon(name_list lp);
void get_func(kind k, functor f, dynstring *name, int x);
void get_func_math(kind k, functor f, dynstring *name, int x);
void get_pred(kind k, functor f, dynstring *name, int x);
void get_pred_math(kind k, functor f, dynstring *name, int x);
void get_const(kind k, functor f, dynstring *name, int x);
void get_func_comp(kind k, functor f, dynstring *name, int x);
void get_pred_comp(kind k, functor f, dynstring *name, int x);
void get_vocab(kind k, functor f, dynstring *name, int x);
kind set_up_type(dynstring *type);
void func_spec(kind k, functor f, dynstring *key_word, int x);
boolean add_clause(dynstring *clause_string);
void add_declaration(dynstring *declaration);
void get_input_cl(kind k, functor f, dynstring *first_word, int softweight);
boolean get_cl(dynstring *lead_word, int softweight);
void re_clausify(void);
boolean isint(char *s);
void get_set(kind k, functor f, dynstring *set, int x);
void verb_spec(kind k, functor f, dynstring *s, int x);
void verb_spec_spec(print_mode printing);
void event_subscription(kind k, functor f, dynstring *s, int x);
void event_subs_spec(debug_event e);
boolean got_partial(void);
boolean got_function(void);
void multiread(functor f, int x, int theArguments[]);
void func_sanity_check(functor f);
void inerror_break_line(char *s1, char *s2, boolean drilldown);
void inerror(char *s1, char *s2, boolean drilldown);
void inwarning(char *s1, char *s2);


/*------ The following are in file output.c -----------------------------*/

void xprint(FILE *f, ifile i, char *s);
void xprintfc(FILE *f, ifile i, const char *format, char c);
void xprintfd(FILE *f, ifile i, const char *format, int x);
void xprintfld(FILE *f, ifile i, const char *format, long int x);
void xprintff(FILE *f, ifile i, const char *format, double x);
void xprintfs(FILE *f, ifile i, const char *format, char *s);
void xprintfsd(FILE *f, ifile i, const char *format, char *s, int x);
void xprintfsf(FILE *f, ifile i, const char *format, char *s, double x);
void print_the_job(void);
void print_job(void);
void xprint_the_job(FILE *fil, ifile ifil);
void xprint_job(FILE *fil, ifile ifil);
void iprint_the_job(ifile ifil);
void fprint_job(FILE *fil);
void iprint_job(ifile ifil);
void print_sorts(FILE *fil, ifile ifil);
void print_one_sort(FILE *fil, ifile ifil, kind k);
void print_functions(FILE *fil, ifile ifil);
void print_one_function(FILE *fil, ifile ifil, functor f);
void print_clauses(FILE *fil, ifile ifil);
void cprint_one_clause(FILE *fil, ifile ifil, clause c, boolean plain);
void print_one_clause(FILE *fil, ifile ifil, clause c, boolean plain);
void FOFprint_clause (FILE *fil, ifile ifil, clause c);
void printwff(wff w);
void outwff(FILE *fil, ifile ifil, wff w);
void print_without_type(FILE *fil, ifile ifil, wff w);
void outsubwff(FILE *fil, ifile ifil, wff w);
void indentwff(FILE *f, wff w, int i, boolean suppress);
void print_indent(FILE *f, int i);
void print_setting(FILE *fil, ifile ifil);
void prints_int(FILE *fil, ifile ifil, char *s, int x);
void prints_bool(FILE *fil, ifile ifil, char *strue, char *sfalse, boolean b);
void prints_verb(FILE *fil, ifile ifil, char *s, print_mode pmd);
void print_the_model(void);
void fprint_the_model(FILE *fil);
void xprint_the_model(FILE *fil, ifile ifil);
void print_model(void);
void fprint_model(FILE *fil);
void xprint_model(FILE *fil, ifile ifil);
void zeroprint(FILE *fil, ifile ifil, functor f);
void oneprint(FILE *fil, ifile ifil, functor f);
void multiprint(FILE *fil, ifile ifil, functor f, int x, int theArguments[]);
void indent(FILE *fil, ifile ifil, int x);
int get_monadic_indentation(functor f);
int *get_monadic_columnwidth(functor f);
int get_indent(functor f, int theArguments[]);
int *get_column_width(functor f, int theArguments[]);
void get_the_value(functor f, int theArguments[], dynstring *dest);
void get_the_name(kind k, int x, dynstring *dest);
int canonlength(kind k);
void printvalue(FILE *fil, ifile ifil, kind k, int x, int len, char nullval);
void print_violated_clauses(FILE *fil, ifile ifil);
void printstats(void);
void dump_report(trin tr);
void ref_size_report(void);
void report_branches(trin tr);
void dump_refs(int rows, int columns);
void print_domains(set possibles[], set impossibles[]);
void dump_space(trs T);

/*------ The following are in file prepare.c ----------------------------*/

int vec_pos(functor f, int argument[]);
int getoffset(functor f, int argument[]);
int commutative_offset(int n, int argument[]);
int commutative_upto(int n, int lead);
functor un_vec_pos(int x, int argument[]);
void get_weight(clause c);
void normalize(void);
boolean is_after(clause c1, clause c2);
void set_up_cc(void);
void clean_up_cc(void);
void set_matrix(clause c);
void set_my_matrix(wff_stack ws);
void set_matrices(wff w);
void set_table(functor f);
void unset_table(functor f);
void set_sizmask(void);
void set_vlength(functor f);
void set_forbidden(functor f);
void set_configurable( functor f );
void set_sum(void);
void set_minus(void);
void set_pred(void);
void set_succ(void);
void set_dif(void);
void set_E(void);
void set_max(void);
void set_fats(functor f);
void set_cc(void);
void set_len(void);
functor un_vec_pos(int x, int theArguments[]);
void print_set(functor f);
void writeref(int x, int y, int z);
void set_max_penalty(int x);
void demote_hiddens(void);
boolean too_long_to_be_soft(int weight, int vars, int tuples);


/*------ The following are in file parser.c ----------------------------*/

void free_wff_list(wff_stack ptr);
void begin_parse(void);
boolean is_a_variable(char *s);
void add_a_variable(char *s);
wff parse(dynstring *symbol, binding_list b);
void pw_merge(psw *pp, psw qq, functor mf);
wff pf_transcribe(psw pp);
wff sparse(dynstring *symbol, binding_list b, boolean ignore_FOF);
wff locate(functor main_func, wff subform[ARGMAX]);
void typecheck(void);
void type_needed(wff w);
void add_type(wff w, kind k);
void AT_plus(wff w);
void AT_up(wff w, char *s);
functor newvar(char s[], int q, kind k);
wff type_complete(wff w);
void type_consistent(wff w);
boolean trivial_cl(void);
void end_parse(clause c);
boolean disjoint_wffstacks(wff_stack w1, wff_stack w2);
wff_stack get_variables_in_list(wff_stack wp, wff_stack varlist);
wff_stack get_variables_in_wff(wff w, wff_stack varlist);
wff_stack get_testables_in_list(wff_stack wp, wff_stack testlist);
wff_stack get_testables_in_wff(wff w, wff_stack testlist);
void inner_first(wff_stack testlist);
void set_valptrs(wff wf);
void positive_last(clause c);
wff_stack sort_wfflist(wff_stack ws);
wff_stack sort_wfflistn(wff_stack ws, int n);
void rewrite_variables(void);
int wffcmp(wff_stack w1, wff_stack w2);
void add_to_theory(void);
wff copy_to_form(wff w);


/*------ The followintg are in file drilldown.c -------------------------*/

char *undecorated(char *s_name);
void bad_variable_drilldown(functor f);
void type_mismatch_drilldown(wff w, int i);
void type_missing_drilldown(wff w, wff bad);
void infix_drilldown(wff w, char *s);
void right_parenthesis_drilldown(wff w, char *s);
void formula_drilldown(char *s);
void quantifier_drilldown(functor f, wff w);
void quantifier_type_mismatch_drilldown(wff w);
void last_argument_drilldown(functor f, wff args[], char *s);
void nonlast_argument_drilldown(functor f, wff args[], char *s);
void parenthesis_after_function_drilldown(functor f, char *s);
wff vacuous_quant(wff w, wff bad);
void free_variable_drilldown(wff w, wff v);
void list_vocabulary(wff w);
f_stack extract_vocabulary(wff w, f_stack vocab);
f_stack reverse_list(f_stack source);
void free_list(f_stack s);


/*------ The followintg are in file clausify.c -------------------------*/

void clausify_report(char *side, char *operator, CLAUSE dest[], int dest_size, boolean replaced[]);
int clausify(clause source, CLAUSE dest[]);
void copy_clause(clause c, clause d);
void free_clause_parts(clause c);
void free_clauses(void);
wff_stack remove_wff(wff_stack thelist, wff w);
wff_stack add_wff(wff_stack thelist, wff w);
void skolemise(wff v, wff w);
int collect_free_vars(wff v, wff w, wff freevar[]);
void substitute(wff body, wff old, wff new);
wff copy_and_substitute(wff body, wff old, wff new);
wff rewrite_bound_variables(wff w);

/*------ The following are in file set_space.c ----------------------*/

void get_sizes(int icomvec[], int vlength);
boolean setposs(set info[], TRS *T);
void set_possval(set info[], functor f, TRS *T);
boolean another_argument(functor f, int argument[]);
int set_priority(functor f, TRS *T);
boolean good_one(set info[], clause c, trs T);
boolean setposs_test(set info[], clause c, wff_stack tf, trs T);

/*----- The following are in file pretest.c ----------------------------*/

boolean do_smallrefs(set info[], TRS *T);
boolean twt_test(set info[], clause c, int lwt, wff plit, TRS *T);
boolean alive(clause c, wff_stack *tf, wff_stack *ltf, 
	      int *maximum, wff_stack wf,
	      set info[], wff plit, TRS *T);
boolean good_all(clause c, wff_stack tf[], wff plit, set info[], TRS *T);
boolean poss_val(wff w, set info[]);
void correct_order(wff_stack tf[]);
int builtin(functor f, int sbvals[]);
void do_injectives(TRS *T);
void set_dinfo(functor f, int x, int argument[], boolean dinfo[]);
boolean do_surjectives(TRS *T);
int localweight(set info[], clause c, wff *plit, TRS *T);
wff_stack firstopen(set info[], wff_stack wfl, int opens);
boolean is_open(functor f);
boolean occurs_in_list(wff fmla, wff_stack wfl);
boolean occurs_inside(wff little, wff big);

/*----- The following are in file test.c -------------------------------*/

boolean test(int info[], TRS *T);
void print_best_model(int info[], trs T);
void vect_into_model(int vector[]);
void set_last_model(int thisvector[]);
boolean is_good(clause cl);
int find_badness(clause cl);
void sanity_check(wff w);
void report_insanity(char error_message[], wff w, int code);
boolean clause_fails(clause c);
boolean next_assignment(clause c);
void record_refutation(clause c, TRS *T);
void lift_clause(clause c);


/*-----	The following are in file newtest.c ----------------------------*/

int tisgood(clause c);
int tset_cc(void);
void untset_cc(void);
void sizes_t(int icomvec[], int vlength);
boolean model_t(int info[], TRS *T);
boolean single_t(set info[], TRS *T);
boolean tuple_t(int info[], TRS *T);
void dump_wcc(void);
boolean next_instance(wff w, int subvalues[]);


/*-----	The following are in file debugger.c ----------------------------*/

void init_events(void);
void standard_masks(debug_event e);
void standard_inmask(debug_event e);
void standard_outmask(debug_event e);
debug_event event_called(char *s);
void collect_runtime_stats(trin tr, boolean mask[]);
void printTheStats(debug_event e);
void collect_and_print_stats(trin tr, debug_event e);
void signal_continue(void);
void wait_for_continue_signal(void);


/*================================= END ================================*/
