#include "FINDER.h"

static int theStats[MAX_REPORT];
static int breakpoints = 0;
static char fstring[100];
static char statNames[MAX_REPORT][100];
static DEBUG_EVENT theEvents[MAX_EVENT];
static boolean continue_flag = false;



/*
 * Sets the contents of the event objects. Called from finitial.c
 */

void init_events(void)
{
  event_id i;
  report_code rc;

  for (i=0; i<MAX_EVENT; i++) {
    theEvents[i].uid = i;
    theEvents[i].subscribed = unsubscribed;    
    for (rc=0; rc<MAX_REPORT; rc++) {
      theEvents[i].in_mask[rc] = false;
      theEvents[i].out_mask[rc] = false;
    }
  }
  strcpy(theEvents[enterSearch].name,"Enter Search");
  strcpy(theEvents[enterSearch].shortname,"enterSearch");
  for (rc=0; rc<MAX_REPORT; rc++)
    theEvents[enterSearch].in_mask[rc] = true;
  theEvents[enterSearch].out_mask[SETPOSSES] = true;
  theEvents[enterSearch].out_mask[DOSMALLS] = true;
  theEvents[enterSearch].out_mask[SEARCHES] = true;
  theEvents[enterSearch].out_mask[SOLUTIONS] = true;
  theEvents[enterSearch].out_mask[VECMAX] = true;
  theEvents[enterSearch].out_mask[PREREFS] = true;
  theEvents[enterSearch].out_mask[F_SUBSUMED] = true;
  theEvents[enterSearch].out_mask[B_SUBSUMED] = true;
  theEvents[enterSearch].out_mask[B1_SUBSUMED] = true;
  theEvents[enterSearch].out_mask[SOFT_DOUBLE] = true;
  
  strcpy(theEvents[leaveSearch].name,"Leave Search");
  strcpy(theEvents[leaveSearch].shortname,"leaveSearch");
  for (rc=0; rc<MAX_REPORT; rc++) {
    theEvents[leaveSearch].in_mask[rc] = true;
    theEvents[leaveSearch].out_mask[rc] = true;
  }

  strcpy(theEvents[enterNode].name,"Enter Node");
  strcpy(theEvents[enterNode].shortname,"enterNode");
  standard_masks(theEvents+enterNode);

  strcpy(theEvents[leaveNode].name,"Leave Node");
  strcpy(theEvents[leaveNode].shortname,"leaveNode");
  standard_masks(theEvents+leaveNode);

  strcpy(theEvents[unitPropagate].name,"Unit Propagate");
  strcpy(theEvents[unitPropagate].shortname,"unitPropagate");
  standard_masks(theEvents+unitPropagate);

  strcpy(theEvents[enterTest].name,"Enter Test");
  strcpy(theEvents[enterTest].shortname,"enterTest");
  standard_masks(theEvents+enterTest);
  theEvents[enterTest].in_mask[SOLUTIONS] = true;
  theEvents[enterTest].out_mask[SOLUTIONS] = true;

  strcpy(theEvents[leaveTest].name,"Leave Test");
  strcpy(theEvents[leaveTest].shortname,"leaveTest");
  standard_masks(theEvents+leaveTest);
  theEvents[leaveTest].in_mask[SOLUTIONS] = true;
  theEvents[leaveTest].out_mask[SOLUTIONS] = true;

  strcpy(theEvents[solution].name,"Solution");
  strcpy(theEvents[solution].shortname,"gotSolution");
  standard_masks(theEvents+solution);
  theEvents[solution].in_mask[SOLUTIONS] = true;
  theEvents[solution].out_mask[SOLUTIONS] = true;

  strcpy(theEvents[abortSearch].name,"Abort Search");
  strcpy(theEvents[abortSearch].shortname,"abortSearch");
  standard_masks(theEvents+abortSearch);
  for (rc=0; rc<MAX_REPORT; rc++) {
    theEvents[leaveSearch].in_mask[rc] = true;
    theEvents[leaveSearch].out_mask[rc] = true;
  }
}

void standard_masks(debug_event e)
{
  standard_inmask(e);
  standard_outmask(e);
}

void standard_inmask(debug_event e)
{
  e->in_mask[BAD_TESTS] = true;
  e->in_mask[SURJ_BACKTRACKS] = true;
  e->in_mask[NULL_BACKTRACKS] = true;
  e->in_mask[INVALUES] = true;
  e->in_mask[FLIPS] = true;
  e->in_mask[REFSIZE] = true;
  e->in_mask[TOPLESS] = true;
  e->in_mask[LOC_MEMORY] = true;
  e->in_mask[MEMORY] = true;
}


void standard_outmask(debug_event e)
{
  e->out_mask[BAD_TESTS] = true;
  e->out_mask[SURJ_BACKTRACKS] = true;
  e->out_mask[NULL_BACKTRACKS] = true;
  e->out_mask[INVALUES] = true;
  e->out_mask[FLIPS] = true;
  e->out_mask[REFSIZE] = true;
  e->out_mask[TOPLESS] = true;
  e->out_mask[LOC_MEMORY] = true;
  e->out_mask[MEMORY] = true;
}


/*
 * Find the event with the given short name and return a pointer to it
 */

debug_event event_called(char *s)
{
  event_id eid;

  for (eid=0; eid<MAX_EVENT; eid++)
    if (!strcmp(theEvents[eid].shortname,s))
      return theEvents+eid;
  return 0;
}


/*
 * Performs action when an event happens
 */

void EVENT(event_id i, trin tr)
{
  debug_event e = theEvents+i;

  switch (e->subscribed) {
  case unsubscribed:
    return;
  case collect_only:
    collect_runtime_stats(tr,e->in_mask);
    return;
  case non_blocking_send:
    collect_and_print_stats(tr,e);
    return;
  case blocking_send:
    collect_and_print_stats(tr,e);
    wait_for_continue_signal();
    return;
  }
}


/*
 * Copies the numbers from the report list of the job, representing
 * floats as ints 100 times bigger (which is a hack, but works OK in
 * practice for this application). The numbers are recorded but not
 * printed.
*/

void collect_runtime_stats(trin tr, boolean mask[])
{
  report_code rc;
  char *colon;
  int x;

  if (!breakpoints)
    for (rc=0; rc<MAX_REPORT; rc++) {
      theStats[rc] = 0;
      for (x=0; isspace(tr->reps[rc].s_report[x]); x++) ;
      strcpy(statNames[rc],tr->reps[rc].s_report+x);
      colon = strchr(statNames[rc],':');
      if (colon) *colon = 0;
    }
  breakpoints++;

  for (rc=0; rc<MAX_REPORT; rc++)
    if (mask[rc]) {
      if (tr->reps[rc].report_type & INT_REPORT)
	theStats[rc] = tr->reps[rc].i_report;
      else if (tr->reps[rc].report_type & FLOAT_REPORT)
	theStats[rc] = (int)(tr->reps[rc].f_report * 100);
    }
}


/*
 * Prints out the most recent recorded values for the report
 * data. This is separate from updating the record.
 */

void printTheStats(debug_event e)
{
  report_code rc;

  printf("<array id=\"%s\">\n",e->name);
  printf("  <int id=\"UID\"> %d </int>\n",breakpoints);
  for (rc=0; rc<MAX_REPORT; rc++)
    if (e->out_mask[rc])
      printf("  <int id=\"%s\"> %d </int>\n",statNames[rc],theStats[rc]);
  printf("</array>\n");
}


/*
 * Self-explanatory
 */

void collect_and_print_stats(trin tr, debug_event e)
{
  collect_runtime_stats(tr, e->in_mask);
  printTheStats(e);
}



void signal_continue(void)
{
  continue_flag = true;
}

void wait_for_continue_signal(void)
{
  while (!continue_flag)
    sleep(1);
  continue_flag = false;
}



/*
 * This is crude but guaranteed to work
 */

int f2i(double f)
{
  int i;
  sprintf(fstring,"%0.0f",f);
  sscanf(fstring,"%d",&i);
  return i;
}

