/*
 * Functions to change the list of clauses, either by adding and
 * deleting or by changing the weights. Drill down, which will replace
 * a clause by some of its instances, is also in here.
 */

#include "diagnose.h"


void add_clause_action(dialogue D, event_list E)
// Read a new clause from stdin and call finder to parse it and
// add it to the problem description. This has to be done via a
// system call, as finder crashes (gracefully) if the clause
// contains any syntax error.
{
  ifile tempfile;
  ifile clausefile = iopen();
  dclause c;

  D->nclauses++;
  D->theClauses = (dclause) realloc(D->theClauses,D->nclauses);
  c = D->theClauses + D->nclauses - 1;
  dclause_init(c);
  printf("\nClause:  "); fflush(stdout);
  dynstring_read_to(&(c->s),'.',stdin);
  tempfile = make_finderfile(D,0,1);
  iprintf(tempfile,"clause { %s }\n", c->s);
  iprintf(tempfile,"setting verbosity {job:full stats:none}\n");
  iprintf(tempfile,"setting no_process\nend\n");
  initFinder(tempfile,clausefile,0);
  if (no_error(clausefile)) {      // New clause parses OK in context
    E->iparams = (int*) malloc(sizeof(int));
    E->iparams[0] = D->nclauses-1;
    E->sparams[0] = (char*) malloc(strlen(c->s.c_str));
    strcpy(E->sparams[0],c->s.c_str);
    free_the_clauses(D);
    D->theClauses = problem_clauses(clausefile,&(D->nclauses));
  }
  else {
    initFinder(tempfile,0,0);                     // Info for the user
    dclause_free(c);
    E->e = no_op;
  }
  iclose(tempfile);
  iclose(clausefile);
}


int no_error(ifile f)
// File f should hold the output from FINDER with verbosity set to
// null. Look at the first line to see whether it reported an error
// during problem input or not. If it did, the first line in the file
// will be "Input error on line ..." or "Warning: ...".
{
  char s[100];

  iscanf(f,"%s %n",s,IEND);
  if (!strncmp(s,"Input",5) || !strncmp(s,"Warning",7))
    return 0;
  return 1;
}


void edit_clause_action(dialogue D, event_list E)
{
  printf("Edit clause action not yet implemented\n");
}


void delete_clause_action(dialogue D, event_list E)
{
  int x;
  int del;
  dclause c;

  show_problem_action(D,0);
  printf("\nDelete which clause?  "); fflush(stdout);
  scanf("%d",&del);  c = D->theClauses+del;
  E->iparams = (int*) malloc(sizeof(int));
  E->iparams[0] = del;
  E->sparams[0] = (char*) malloc(strlen(c->s.c_str));
  strcpy(E->sparams[0],c->s.c_str);
  dynstring_free(&(c->s));
  if (c->var_name) free(c->var_name);
  if (c->var_sort) free(c->var_sort);
  c->var_name = c->var_sort = 0;
  for (x=del; x<D->nclauses; x++)
    copy_dclause(D->theClauses+x+1,D->theClauses+x);
  D->nclauses--;
}


void weight_clause_action(dialogue D, event_list E, int w)
{
  int cid;
  dclause c;

  show_problem_action(D,0);
  printf("\nAdjust which clause?  "); fflush(stdout);
  scanf("%d",&cid);  c = D->theClauses+cid;
  E->iparams = (int*) malloc(2*sizeof(int));
  E->iparams[0] = cid;
  E->sparams[0] = (char*) malloc(strlen(c->s.c_str));
  strcpy(E->sparams[0],c->s.c_str);
  switch(w) {
  case 2:
    printf("Current weight = %d.  New weight = ", c->softweight);
    fflush(stdout);
    scanf("%d", &w);
  case 0:
  case 1:
    E->iparams[1] = c->softweight = w;
    c->sacred = (w? 0: 1);
  }
}


void change_time_limit_action(dialogue D, event_list E)
{
  printf("Current time limit (seconds): %d\nNew time limit:  ",D->time);
  fflush(stdout);
  scanf("%d",&(D->time));
  E->iparams = (int*) malloc(sizeof(int));
  E->iparams[0] = D->time;
}


void drill_down_action(dialogue D, event_list E)
{
  printf("NYI. It is still not quite clear what this option will do.\n");
  PAUSE();
}


