#include "dual.h"
#include <string.h>
#include <strings.h>

static SET SIGMA;   // The universe
static set Sigma;   // Pointer to the universe
static D_parameters D_PARAM;
static int D_initialised = 0;
family D_restrict(family source, set H, family dest);


set D_OPT(D_problem p,
	  int (*sol)(D_problem,set),
	  set (*inc_sol)(D_problem,set))
// Find and return a minmum-cost hiting set for the given problem. The
// function (*sol) tests whether a set is a solution to a
// D_problem. If (inc_sol) is specified, it returns the next set to be
// hit. If not, iteration of (*sol) over all members of Sigma is used
// instead.
{
  set D_Starter();
  family sofar = 0;
  set H;
  set D;
  int x;
  int count;
  int bound = 0;

  if (!D_initialised)
    D_initial();
  if (D_PARAM.interval && D_PARAM.incremental_hs == minimum) {
    fprintf(stderr,"WARNING: setting interval to zero for minimum hitting sets\n");
    D_PARAM.interval = 0;
  }
  if (D_PARAM.seeded)
    H = D_Starter(p,sol);
  else
    H = S_new(1,0);
  Forallmembers(H,x) {
    D = S_new(Sigma->n,0);
    S_into(x,D);
    sofar = F_into(D,sofar);
  }

  for (;;) {
    // Test the optimal hitting set
    D = HS_test(p,H,&sofar,sol,inc_sol);
    if (!D) {
      D_initialised = 0;       // In case called again
      return H;
    }
    if (D_PARAM.verbosity) print_one_set(D,1);

    // Try fast and dirty first. Doesn't happen if count is zero
    for (count=0; count<D_PARAM.interval; count++) {
      if (D_PARAM.incremental_hs == anything)
	S_into(S_first_member(D,-1),H);
      else
	H = really_dirty_hittingset(sofar);
      D = HS_test(p,H,&sofar,sol,inc_sol);
      if (!D)
	break;
      if (D_PARAM.verbosity) print_one_set(D,1);
    }
    H = HS(sofar,bound);
    if (H) {
      if (S_size(H) < bound) {
	fprintf(stderr,"Hitting set ");
	fprint_one_set(stderr,H,1);
	HS_abort("Bound decreased");
      }
    }
    else {
      H = HS(sofar,0);
      bound = S_size(H);
    }
  }
}


void D_initial(void)
// Called before D_OPT starts, to initialise Sigma and D_PARAM
{
  Sigma = &SIGMA;
  S_empty(Sigma);
  D_PARAM.seeded = seeded_default;
  D_PARAM.incremental_hs = incremental_hs_default;
  D_PARAM.goal_hs = goal_hs_default;
  D_PARAM.interval = interval_default;
  D_initialised = 1;
}


void D_add_element(int x, int cost)
// Simply adds one integer to Sigma
{
  S_into(x,Sigma);
}


void D_set_flag(char *name)
// Set the named boolean flag in D_PARAM to true
{
  if (!strcmp(name,"seeded"))
    D_PARAM.seeded = 1;
}


void D_unset_flag(char *name)
// Set the named boolean flag in D_PARAM to false
{
  if (!strcmp(name,"seeded"))
    D_PARAM.seeded = 0;
}


void D_set_hs_type(char *name, HS_type t)
// Set the named search type in D_PARAM to the given value
{
  if (!strcmp(name,"during"))
    D_PARAM.incremental_hs = t;
  else if (!strcmp(name,"after"))
    D_PARAM.goal_hs = t;
}


void D_set_verbosity(int i)
{
  D_PARAM.verbosity = i;
}


void D_set_interval(int i)
{
  if (i <1 ) i = 1;
  if (i > 100) i = 100;
  D_PARAM.interval = i;
}


set HS_test(D_problem p, set H, family *sofar,
	int (*sol)(D_problem,set), set (*inc_sol)(D_problem,set))
{
  set D_Sel();
  set D;

  if ((*sol)(p,H))
    return 0;
  if (inc_sol)
    D = (*inc_sol)(p,H);
  else
    D = D_Sel(p,sol,H);
  *sofar = F_into(D,*sofar);
  return D;
}

/*
#ifdef DEBUG
    printf("Hitting set: "); Forallmembers(H,x) printf(" %d",x); printf("\n");
#endif
*/

set D_OPT_new(D_problem p, int (*sol)(D_problem,set), set (*inc_sol)(D_problem,set))
// Find and return a minmum-cost hiting set for the given problem. The
// function (*sol) tests whether a set is a solution to a
// D_problem. If (inc_sol) is specified, it returns the next set to be
// hit. If not, iteration of (*sol) over all members of Sigma is used
// instead.
{
  set D_Sel();
  set D_Starter();
  family sofar = 0;
  family sofar_core = 0;
  set H;
  set D;
  int x;

  printf("OPT new\n"); fflush(stdout);
  /*
   *  H = D_Starter(p,sol);
   */
  H = S_new(1,0);
  Forallmembers(H,x) {
    D = S_new(Sigma->n,0);
    S_into(x,D);
    sofar = F_into(D,sofar);
  }

  for (;;) {
    if ((*sol)(p,H))
      return H;
    if (inc_sol)
      D = (*inc_sol)(p,H);
    else
      D = D_Sel(p,sol,H);

    // H is a hitting set for everything in sofar except D
    printf("%d/%d sets to hit.  Latest one size %d.  HS size %d\n",
	   F_size(sofar_core), F_size(sofar), S_size(D), S_size(H));

    sofar_core = D_restrict(sofar,H,sofar_core);
    sofar = F_into(D,sofar);
    do {
      sofar_core = F_into(D,sofar_core);
      H = HS(sofar_core,0);
      D = F_disjoint_from(sofar,H);
    }
    while (D);
  }
}


set D_Sel(D_problem p, int (*sol)(D_problem,set), set H)
// Just a subroutine of the above, to get the next inclusion-minimal
// member of the dual set which will be disjoint from the currently
// suggested hitting set.
{
  set A;
  set Abar;
  int x;

  A = S_copy(H,0);
  Abar = S_new(S_size(Sigma),0);
  Forallmembers(Sigma,x) {
    if (!S_is_member(x,H)) {
      S_into(x,A);
      if ((*sol)(p,A)) {
	S_outof(x,A);
	S_into(x,Abar);
      }
    }
  }
  return Abar;
}


set D_Starter(D_problem p, int (*sol)(D_problem,set))
// Another subroutine. Find and return the set of singletons requiring
// to be hit
{
  set A;
  set Allbut;
  int x;

  A = S_new(Sigma->n,0);
  Allbut = S_copy(Sigma,0);
  Forallmembers(Sigma,x) {
    S_outof(x,Allbut);
    if (!(*sol)(p,Allbut))
      S_into(x,A);
    S_into(x,Allbut);
  }
  S_free(Allbut);
  return A;
}


family D_restrict(family source, set H, family dest)
{
  set s;
  int x;

  F_free(dest);
  dest = 0;
  Forallmembers(H,x) {
    S_outof(x,H);
    s = F_disjoint_from(source,H);
    dest = F_into(s,dest);
    S_into(x,H);
  }
  return dest;
}
