#include "FINDER.h"

static wff CLAUSE_ERROR;

void clausify_report(char *side, char *operator, CLAUSE dest[], int dest_size, boolean replaced[])
// All calls to this protected by DEBUG definition
{
  int x;

  printf("Treated %s on %s side:\n",operator,side);
  for (x=0; x<dest_size; x++) {
    if (!replaced[x]) {
      printf("    ");
      cprint_one_clause(0,0,dest+x,false);
    }
    printf("\n");
  }
}

int clausify(clause source, CLAUSE dest[])
{
  int dest_size = 1;
  boolean replaced[CL_MAX];
  int x,y;
  wff_stack wptr;
  wff w_rewritten;

#ifdef DEBUG
  printf("Clausifying "); cprint_one_clause(0,source,false); printf("\n");
#endif
  CLAUSE_ERROR = 0;
  for (x=0; x<CL_MAX; x++) replaced[x] = false;
  copy_clause(source,dest);
  for (x=0; !setting.parse_only && x<dest_size; x++) {
    /* Clausify literals on the premise side */
    for (wptr=dest[x].prem; wptr && !replaced[x]; wptr=wptr->nxt) {
      if (!strcmp(wptr->w->sym->fname,"NOT")) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].prem = remove_wff(dest[dest_size].prem,wptr->w);
	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[0]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("prem","NOT",dest,dest_size,replaced);
#endif
      }
      else if (!strcmp(wptr->w->sym->fname,"AND")) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].prem = remove_wff(dest[dest_size].prem,wptr->w);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[0]);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("prem","AND",dest,dest_size,replaced);
#endif
      }
      else if (!strcmp(wptr->w->sym->fname,"OR")) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].prem = remove_wff(dest[dest_size].prem,wptr->w);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[0]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	copy_clause(dest+x,dest+dest_size);
	// ALSO REWRITE ALL BOUND VARIABLES ??
	dest[dest_size].prem = remove_wff(dest[dest_size].prem,wptr->w);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("prem","OR",dest,dest_size,replaced);
#endif
      }
      else if (!strcmp(wptr->w->sym->fname,"IMP")) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].prem = remove_wff(dest[dest_size].prem,wptr->w);
	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[0]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	copy_clause(dest+x,dest+dest_size);
	// ALSO REWRITE ALL BOUND VARIABLES ??
	dest[dest_size].prem = remove_wff(dest[dest_size].prem,wptr->w);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("prem","IMP",dest,dest_size,replaced);
#endif
      }
      else if (wptr->w->sym->fname[0] == '=' &&
	       wptr->w->sb[0]->sym->val_type == Bool) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].prem = remove_wff(dest[dest_size].prem,wptr->w);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[0]);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	copy_clause(dest+x,dest+dest_size);
	// ALSO REWRITE ALL BOUND VARIABLES ??
	dest[dest_size].prem = remove_wff(dest[dest_size].prem,wptr->w);

	// Reinstated code
       	w_rewritten = rewrite_bound_variables(wptr->w);
	//	if (CLAUSE_ERROR) {
	//	  free_variable_drilldown(wptr->w, CLAUSE_ERROR);
	//	}
       	dest[dest_size].conc = add_wff(dest[dest_size].conc,w_rewritten->sb[0]);
       	dest[dest_size].conc = add_wff(dest[dest_size].conc,w_rewritten->sb[1]);
	// End reinstated section

	// Alternative code removed
	//	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[0]);
	//       	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("prem","IFF",dest,dest_size,replaced);
#endif
      }
      else if (wptr->w->sym->category == Quant) {
	copy_clause(dest+x,dest+dest_size);
	wptr->w->sb[0]->sym->binder = 0;
	if (!strcmp(wptr->w->sym->fname,"ALL"))
	  skolemise(wptr->w->sb[0],wptr->w->sb[1]);
	dest[dest_size].prem = remove_wff(dest[dest_size].prem,wptr->w);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("prem","QUANT",dest,dest_size,replaced);
#endif
      }
    }

    /* Now clausify on the conclusion side */
    for (wptr=dest[x].conc; wptr && !replaced[x]; wptr=wptr->nxt) {
      if (!strcmp(wptr->w->sym->fname,"NOT")) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].conc = remove_wff(dest[dest_size].conc,wptr->w);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[0]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("conc","NOT",dest,dest_size,replaced);
#endif
      }
      else if (!strcmp(wptr->w->sym->fname,"OR")) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].conc = remove_wff(dest[dest_size].conc,wptr->w);
	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[0]);
	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("conc","OR",dest,dest_size,replaced);
#endif
      }
      else if (!strcmp(wptr->w->sym->fname,"IMP")) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].conc = remove_wff(dest[dest_size].conc,wptr->w);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[0]);
	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("conc","IMP",dest,dest_size,replaced);
#endif
      }
      else if (!strcmp(wptr->w->sym->fname,"AND")) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].conc = remove_wff(dest[dest_size].conc,wptr->w);
	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[0]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	copy_clause(dest+x,dest+dest_size);
	// ALSO REWRITE ALL BOUND VARIABLES ??
	dest[dest_size].conc = remove_wff(dest[dest_size].conc,wptr->w);
	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("conc","AND",dest,dest_size,replaced);
#endif
      }
      else if (wptr->w->sym->fname[0] == '=' &&
	       wptr->w->sb[0]->sym->val_type == Bool) {
	copy_clause(dest+x,dest+dest_size);
	dest[dest_size].conc = remove_wff(dest[dest_size].conc,wptr->w);
	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[0]);
	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	copy_clause(dest+x,dest+dest_size);
	// ALSO REWRITE ALL BOUND VARIABLES ??
	dest[dest_size].conc = remove_wff(dest[dest_size].conc,wptr->w);

	// Reinstated code
       	w_rewritten = rewrite_bound_variables(wptr->w);
	//       	if (CLAUSE_ERROR) {
	//       	  free_variable_drilldown(wptr->w, CLAUSE_ERROR);
	//       	}
       	dest[dest_size].conc = add_wff(dest[dest_size].conc,w_rewritten->sb[0]);
       	dest[dest_size].prem = add_wff(dest[dest_size].prem,w_rewritten->sb[1]);
	// End reinstated section

	// Alternative code removed
	//	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[0]);
	//	dest[dest_size].prem = add_wff(dest[dest_size].prem,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("conc","IFF",dest,dest_size,replaced);
#endif
      }
      else if (wptr->w->sym->category == Quant) {
	copy_clause(dest+x,dest+dest_size);
	wptr->w->sb[0]->sym->binder = 0;
	if (!strcmp(wptr->w->sym->fname,"SOME"))
	  skolemise(wptr->w->sb[0],wptr->w->sb[1]);
	dest[dest_size].conc = remove_wff(dest[dest_size].conc,wptr->w);
	dest[dest_size].conc = add_wff(dest[dest_size].conc,wptr->w->sb[1]);
	dest_size++;
	if (dest_size >= CL_MAX)
	  inerror("Clause limit exceeded during clausification","",0);
	replaced[x] = true;
#ifdef DEBUG
	clausify_report("conc","QUANT",dest,dest_size,replaced);
#endif
      }
    }
    if (replaced[x])  free_clause_parts(dest+x);
  }
  replaced[dest_size] = true;
  for (x=0; x<dest_size; x++)
    if (replaced[x]) {
      for (y=x+1; y<dest_size; y++)
	if (!replaced[y]) {
	  copy_clause(dest+y,dest+x);
	  free_clause_parts(dest+y);
	  replaced[y] = true;
	  replaced[x] = false;
	  break;
	}
    }
  for (x=0; x<dest_size; x++) {
    if (replaced[x])
      return x;
    end_parse(dest+x);
    get_weight(dest+x);
  }
  return dest_size;
}


/*
 * Simply make a copy of the contents of c
 */

void copy_clause(clause c, clause d)
{
  wff_stack wptr;

  free_clause_parts(d);
  for(wptr=c->prem; wptr; wptr=wptr->nxt)
    d->prem = add_wff(d->prem,wptr->w);
  for(wptr=c->conc; wptr; wptr=wptr->nxt)
    d->conc = add_wff(d->conc,wptr->w);
  for(wptr=c->varlist; wptr; wptr=wptr->nxt)
    d->varlist = add_wff(d->varlist,wptr->w);
  for(wptr=c->testlist; wptr; wptr=wptr->nxt)
    d->testlist = add_wff(d->testlist,wptr->w);
  d->weight = c->weight;
  //  printf("Copy weight %d to %d\n",c->weight,d->weight);
  d->varcount = c->varcount;
  d->softweight = c->softweight;
  //  printf("Copy softweight %d to %d\n",c->softweight,d->softweight);
  d->badness = c->badness;
  d->preprocessed = c->preprocessed;
  d->tautology = c->tautology;
  d->pid = c->pid;
  d->cid = c->cid;
  //  printf("Copied clause to %d: ",(int)d); cprint_one_clause(0,d,false);
}


void free_clause_parts(clause c)
{
  if (!c) return;
  free_wff_list(c->prem);
  c->prem = 0;
  free_wff_list(c->conc);
  c->conc = 0;
  free_wff_list(c->varlist);
  c->varlist = 0;
  free_wff_list(c->testlist);
  c->testlist = 0;
}


void free_clauses(void)
{
  clause c;

  for (c=TH[current_theory]; c; c=c->nextclause)
    free_clause_parts(c);
}


wff_stack remove_wff(wff_stack thelist, wff w)
{
  wff_stack ptr, trail;

  if (!thelist) return 0;
  if (thelist->w == w) {
    ptr = thelist->nxt;
    free(thelist);
    return ptr;
  }
  for (ptr=thelist->nxt,trail=thelist;
       ptr;
       ptr=ptr->nxt,trail=trail->nxt)
    if (ptr->w == w) {
      trail->nxt = ptr->nxt;
      free(ptr);
      break;
    }
  return thelist;
}


wff_stack add_wff(wff_stack thelist, wff w)
{
  wff_stack ptr;

  for (ptr=thelist; ptr; ptr=ptr->nxt) 
    if (ptr->w == w)
      return thelist;
  ptr = (wff_stack) malloc(sizeof(WFF_STACK));  /* Where's this get freed? */
  ptr->nxt = thelist;
  ptr->w = w;
  return ptr;
}



/*
 * Create a hidden functor sf with same value type as variable v and
 * argument types of the free variables x1, ..., xn in w, not counting
 * v. Create a new wff sw being sf(x1, ..., xn). Go recursively
 * through w replacing all [free] ocurrences of v by sw.
 */

void skolemise(wff v, wff w)
{
  functor sf;
  wff sw;
  wff freevar[ARGMAX];
  int x;
  int n;
  char nonce[LEXLENGTH];

  for (x=0; x<ARGMAX; x++) freevar[x] =0;
  n = collect_free_vars(v,w,freevar);
  sprintf(nonce,"skolem_%d",next_qid++);
  sf = newfunctor(nonce,Func,n,false);
  for (x=0; x<n; x++)
    sf->arg_type[x] = freevar[x]->sym->val_type;
  sf->val_type = v->sym->val_type;
  sf->pmode = null;
  sf->cut_it = true;
  sw = locate(sf,freevar);
  if (w == v)
    inerror(v->sym->fname,"skolemised in itself",0);
  substitute(w,v,sw);
}


/*
 * Go recursively through w adding all the (distinct) free variables
 * except for v to the array freevar. Return the number of variables
 * collected. Crash if ARGMAX exceeded.
 */

int collect_free_vars(wff v, wff w, wff freevar[])
{
  int sofar = 0;
  int x;

  if (!w) return 0;
  if (w->sym->category == Var && w->sym->binder == 0 && v != w) {
    for (x=0; freevar[x]; x++) {
      if (x == ARGMAX)
	inerror("Skolem function has too many arguments","",0);
      if (freevar[x] == w)
	break;
    }
    if (!freevar[x]) {
      freevar[x] = w;
      return 1;
    }
  }
  for (x=0; x<w->sym->adicity; x++)
    sofar += collect_free_vars(v,w->sb[x],freevar);
  return sofar;
}


/*
 * Recursive call to substitute new expression for old throughout the
 * body of a formula.
 */

void substitute(wff body, wff old, wff new)
{
  int x;

  if (!body || !old || !new) return;
  for (x=0; x<body->sym->adicity; x++)
    if (body->sb[x] == old)
      body->sb[x] = new;
    else
      substitute(body->sb[x],old,new);
}


/*
 * As above, but make a new formula rather than changing the old one
 */

wff copy_and_substitute(wff body, wff old, wff new)
{
  int x;
  wff args[ARGMAX];

  if (!old) return body;
  if (!body || !new) return 0;
  if (body == old)
    return new;
  for (x=0; x<ARGMAX; x++) args[x] = 0;
  for (x=0; x<body->sym->adicity; x++)
    args[x] = copy_and_substitute(body->sb[x],old,new);
  return locate(body->sym,args);
}


/*
 * Change all bound variables in a formula to new ones, to avoid
 * possible clashes with the same variables bound somewhere
 * else. Return a new formula with the variables rewritten, to avoid
 * overwriting the old one.
 */

wff rewrite_bound_variables(wff w)
// Probably isn't needed. In any case, the "free variable" check is nonsense.
// Removed for now. If bad things happen, correct and reactivate it.
{
  int x;
  wff old;
  wff new;
  char varname[LEXLENGTH];
  char subscript[LEXLENGTH];
  char typename[LEXLENGTH];
  char *ptr;
  wff args[ARGMAX];
  wff dummy_args[ARGMAX];
  wff result;

  if (!w || CLAUSE_ERROR) return 0;
  for (x=0; x<ARGMAX; x++) dummy_args[x] = args[x] = 0;
  for (x=0; x<w->sym->adicity; x++)
    args[x] = rewrite_bound_variables(w->sb[x]);
  if (CLAUSE_ERROR) return 0;
  result = locate(w->sym,args);
  if (result->sym->category == Quant) {
    old = result->sb[0];
    strcpy(varname,old->sym->fname);
    ptr = strchr(varname,':');
    if (ptr) {
      strcpy(typename,ptr);
      *ptr = 0;
    }
    else *typename = 0;
    ptr = (strstr(varname,"__"));
    /*
    if (!ptr) {
      if (setting.drilldown) {
	CLAUSE_ERROR = old;
	return 0;
      }
      else
	skipout("Quantifier binds free variable",SKIP);
    }
    */
    if (ptr) *ptr = 0;
    sprintf(subscript,"__%d",next_qid++);
    strcat(varname,subscript);
    if (*typename)
      strcat(varname,typename);
    new = locate(newvar(varname,next_qid,old->sym->val_type),dummy_args);
    result = copy_and_substitute(result,old,new);
  }
  return result;
}


/*
Parse - typecheck - clausify needs a rethink. The parser needs to allocate subscripts to guard against type errors where the same expression is used as a variable of two types in one formula. These can be stripped off following type allocation, except where the clausifier needs them to make prenex form work properly. Alternatively, leave the numbers as before, but just don't print them. Give thought to better names for the skolem functions as well: users are confused.
 */
