/***************************** DRILLDOWN STUFF ***************************/

#include "FINDER.h"


static char temporary_string[1000];         // For nonce uses only

char *undecorated(char *s_name)
// If the string ends in "__<number>" strip off the ending.
// Do not change the original, but return a pointer to the copy.
{
  int x;

  strcpy(temporary_string,s_name);
  if (!isdigit(s_name[strlen(s_name)-1]))
    return s_name;
  for (x=strlen(s_name)-1; x>=0 && isdigit(s_name[x]); x--) ;
  if (!x || s_name[x] != '_' || s_name[x-1] != '_')
    return s_name;

  strcpy(temporary_string,s_name);
  temporary_string[x-1] = 0;
  return temporary_string;
}


void bad_variable_drilldown(functor f)
{
  fprintf(WARNING_STREAM,
	  "The last item parsed was \"%s\".\n",f->fname);
  fprintf(WARNING_STREAM,
	  "This occurs earlier in the formula as a variable bound by a\n");
  fprintf(WARNING_STREAM,
	  "quantifier (ALL or SOME), but here it is no longer in the scope\n");
  fprintf(WARNING_STREAM,
	  "of the quantifier, so it counts as a name. It is not allowed to\n");
  fprintf(WARNING_STREAM,
	  "be both a variable and a name in the same formula.\n\n");
  fprintf(WARNING_STREAM,
	  "Check the parentheses - maybe there are some missing.\n\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}


void infix_drilldown(wff w, char *s)
{
  boolean itsaname = false;

  if (w) {
    fprintf(WARNING_STREAM,"The last item parsed was\n    ");
    printwff(w);
    if (w->sym->adicity == 0) {
      if (w->sym->declared)
	fprintf(WARNING_STREAM,"\nwhich is a ");
      else
	fprintf(WARNING_STREAM,"\nwhich is an undeclared ");
      if (w->sym->category == Var) 
	fprintf(WARNING_STREAM,"variable ");
      else {
	fprintf(WARNING_STREAM,"name ");
	itsaname = true;
      }
    }
    else
      fprintf(WARNING_STREAM,"\nwhich is an expression ");
    if (w->sym->val_type)
      fprintf(WARNING_STREAM,"of type %s.\n",w->sym->val_type->sname);
    else
      fprintf(WARNING_STREAM,"of unknown type.\n");

    fprintf(WARNING_STREAM,"The only symbol that can come next is a binary operator in infix position.\n\n");
    if (itsaname)
      fprintf(WARNING_STREAM,"If %s isn't supposed to be a name, check the spelling.\n\n", w->sym->fname);
    else if (w->sym->adicity) {
      list_vocabulary(w);
      fprintf(WARNING_STREAM,"\nParsed [sub]formula structure is:\n\n");
      indentwff(WARNING_STREAM,w,2,false);
      fprintf(WARNING_STREAM,"\n");
    }
  }
  else
    fprintf(WARNING_STREAM,"No more detailed information available\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}


void right_parenthesis_drilldown(wff w, char *s)
{
  fprintf(WARNING_STREAM,
	  "Detailed diagnostics: parsing a formula in parentheses. ");
  fprintf(WARNING_STREAM, "Just parsed:\n  (");
  printwff(w);
  fprintf(WARNING_STREAM, "\nThe next symbol should be the closing ");
  fprintf(WARNING_STREAM,"parenthesis \")\" but \"%s\" was\n",s);
  fprintf(WARNING_STREAM,"found instead.\n");

  fprintf(WARNING_STREAM,"\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}


void quantifier_drilldown(functor f, wff w)
{
  boolean itsaname = false;

  fprintf(WARNING_STREAM, "Detailed diagnostics: the quantifier symbol %s ",
	  f->fname);
  fprintf(WARNING_STREAM, "should be followed by a\nvariable, but instead was followed by ");
  printwff(w);
  fprintf(WARNING_STREAM,"\n");

  if (w->sym->adicity == 0) {
    if (w->sym->declared)
      fprintf(WARNING_STREAM," which is a ");
    else
      fprintf(WARNING_STREAM," which is an undeclared ");
    if (w->sym->category == Var) 
      fprintf(WARNING_STREAM,"variable ");
    else {
      fprintf(WARNING_STREAM,"name ");
      itsaname = true;
    }
  }
  else
    fprintf(WARNING_STREAM," which is an expression ");
  if (w->sym->val_type)
    fprintf(WARNING_STREAM,"of type %s.\n",w->sym->val_type->sname);
  else
    fprintf(WARNING_STREAM,"of unknown type.\n");
  fprintf(WARNING_STREAM,"\n");

  if (itsaname)
    fprintf(WARNING_STREAM,"If it isn't supposed to be a name, check how it's used on previous lines.\n\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}


void quantifier_type_mismatch_drilldown(wff w)
{
 fprintf(WARNING_STREAM, "Detailed diagnostics: the [sub]formula\n  ");
  printwff(w);
  fprintf(WARNING_STREAM,
	  "\nhas a quantifier \"%s %s\" as its main operator. ",
	  w->sym->fname, w->sb[0]->sym->fname);
  fprintf(WARNING_STREAM, "For that to make sense, the\n");
  fprintf(WARNING_STREAM, "body of the formula should be a boolean ");
  fprintf(WARNING_STREAM, "expression, (containing \"%s\") but\n  ",
	  w->sb[0]->sym->fname);
  printwff(w->sb[1]);
  fprintf(WARNING_STREAM,
	  "\nis an expression of type %s, not of type bool.\n\n",
	  w->sb[1]->sym->val_type->sname);

  list_vocabulary(w);

  fprintf(WARNING_STREAM,"\nParsed [sub]formula structure is:\n\n");
  indentwff(WARNING_STREAM,w,2,false);

  fprintf(WARNING_STREAM,"\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}

void last_argument_drilldown(functor f, wff args[], char *s)
{
  int x;

  fprintf(WARNING_STREAM, "Detailed diagnostics: the function symbol ");
  fprintf(WARNING_STREAM, "%s in prefix position should be\nfollowed by ",
	  f->fname);
  if (f->adicity > 1)
    fprintf(WARNING_STREAM, "a comma-separated list of %d arguments",
	    f->adicity);
  else if (f->adicity == 1)
    fprintf(WARNING_STREAM, "one argument");
  else if (f->adicity == 0)
    fprintf(WARNING_STREAM, "nothing, or zero arguments");
  fprintf(WARNING_STREAM, " enclosed in parentheses.\n");
  fprintf(WARNING_STREAM,"The symbol after \"%s(", f->fname);
  for (x=0; args[x]; x++) {
    if (x) fprintf(WARNING_STREAM, ", ");
    printwff(args[x]);
  }
  fprintf(WARNING_STREAM, "\" should be closing parenthesis \")\",\n");
  fprintf(WARNING_STREAM, "but \"%s\" was found instead.\n", s);

  fprintf(WARNING_STREAM,"\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}


void nonlast_argument_drilldown(functor f, wff args[], char *s)
{
  int x;

  fprintf(WARNING_STREAM, "Detailed diagnostics: the function symbol ");
  fprintf(WARNING_STREAM, "%s in prefix position should be\n",
	  f->fname);
  fprintf(WARNING_STREAM, "followed by a comma-separated list of %d arguments",
	  f->adicity);
  fprintf(WARNING_STREAM, " enclosed in parentheses.\n");
  fprintf(WARNING_STREAM,"The symbol after \"%s(", f->fname);
  for (x=0; args[x]; x++) {
    if (x) fprintf(WARNING_STREAM, ", ");
    printwff(args[x]);
  }
  fprintf(WARNING_STREAM, "\" should be a comma \",\",\n");
  fprintf(WARNING_STREAM, "but \"%s\" was found instead.\n", s);

  fprintf(WARNING_STREAM,"\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}


void parenthesis_after_function_drilldown(functor f, char *s)
{
  fprintf(WARNING_STREAM, "Detailed diagnostics: the function symbol ");
  fprintf(WARNING_STREAM, "%s in prefix position should be\n",
	  f->fname);
  fprintf(WARNING_STREAM, "followed by %d arguments enclosed in ",
	  f->adicity);
  fprintf(WARNING_STREAM, "parentheses, so the symbol after\n");
  fprintf(WARNING_STREAM,"%s should be \"(\", but \"%s\" was found instead.\n",
	  f->fname, s);

  fprintf(WARNING_STREAM,"\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}


void type_mismatch_drilldown(wff w, int i)
{
  fprintf(WARNING_STREAM, "Detailed diagnostics: in the formula\n  ");
  printwff(w);
  if (strchr("=<>",w->sym->fname[0]) || 
      !strcmp(w->sym->fname,"DIF") || 
      !strcmp(w->sym->fname,"!=")) {
    fprintf(WARNING_STREAM,
	    "\nthe arguments of the main %s ",
	    w->sym->fname);
    fprintf(WARNING_STREAM, "should be of the same type, but\n  ");
    printwff(w->sb[0]);
    fprintf(WARNING_STREAM, "\nis of type \"%s\", and\n  ",
	    w->sb[0]->sym->val_type->sname); 
    printwff(w->sb[1]);
    fprintf(WARNING_STREAM, "\nis of type \"%s\".\n\n",
	    w->sb[1]->sym->val_type->sname); 
  }
  else {
    fprintf(WARNING_STREAM,"\nthe main operator \"%s\" ",w->sym->fname);
    fprintf(WARNING_STREAM,"expects argument %d to be of type %s\n",
	    i+1, w->sym->arg_type[i]->sname);
    fprintf(WARNING_STREAM,"but argument %d is\n  ",i+1);
    printwff(w->sb[i]);
    fprintf(WARNING_STREAM,"\nwhich is of type %s.\n\n",
	    w->sb[i]->sym->val_type->sname);
    fprintf(WARNING_STREAM, "Hints for this kind of error: check\n");
    fprintf(WARNING_STREAM, "  (1) parentheses;\n");
    fprintf(WARNING_STREAM, "  (2) possible typos (e.g. misspelling);\n");
    fprintf(WARNING_STREAM,
	    "  (3) how variables and names get their sorts assigned.\n\n");
  }

  list_vocabulary(w);

  fprintf(WARNING_STREAM,"\nParsed [sub]formula structure is:\n\n");
  indentwff(WARNING_STREAM,w,2,false);

  fprintf(WARNING_STREAM,"\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}

void type_missing_drilldown(wff w, wff bad)
{
  wff vq;

  fprintf(WARNING_STREAM, "Detailed diagnostics: in the formula\n  ");
  printwff(w);
  // Vacuous quantifiers cause this error to appear: check that first
  vq = vacuous_quant(w,bad);
  if (vq) {
    fprintf(WARNING_STREAM, "\nthe variable ");
    printwff(bad);
    fprintf(WARNING_STREAM, " occurs in the quantifier %s %s\n",
	    vq->sym->fname, bad->sym->fname);
    fprintf(WARNING_STREAM, "but does not subsequently occur in the scope ");
    fprintf(WARNING_STREAM, "of that quantifier:\n");
    printwff( vq->sb[1]);
    fprintf(WARNING_STREAM, "\n\n");
  }
  else {
    fprintf(WARNING_STREAM, "\nthe ");
    if (bad->sym->category == Var)
      fprintf(WARNING_STREAM, "variable\n  ");
    else if (bad->sym->category == Const)
      fprintf(WARNING_STREAM, "constant\n  ");
    else if (bad->sym->category == Func && bad->sym->adicity == 0)
      fprintf(WARNING_STREAM, "name\n  ");
    else
      fprintf(WARNING_STREAM, "symbol\n ");
    printwff(bad);
    fprintf(WARNING_STREAM,
	    "\noccurs, but not in a position from which its sort can be ");
    fprintf(WARNING_STREAM,
	    "deduced.\nThis is not allowed, as every well-formed expression ");
    fprintf(WARNING_STREAM, "must have a type.\n\n");
  }

  list_vocabulary(w);

  fprintf(WARNING_STREAM,"\nParsed formula structure is:\n\n");
  indentwff(WARNING_STREAM,w,2,false);

  fprintf(WARNING_STREAM,"\n");
  fflush(WARNING_STREAM);
  exit(current_line);
}


wff vacuous_quant(wff w, wff bad)
{
  int x;
  wff v;

  if (w) {
    if (w->sym->category == Quant &&
	w->sb[0] == bad &&
	!occurs_inside(bad,w->sb[1]))
      return w;
    for (x=0; x<w->sym->adicity; x++) {
      v = vacuous_quant(w->sb[x],bad);
      if (v) return v;
    }
  }
  return 0;
}


void free_variable_drilldown(wff w, wff v)
// This one from the clausifier is a skipout, not an inerror
{
  fprintf(WARNING_STREAM, "Error: variable occurs bound and free.\n\n");

  fprintf(WARNING_STREAM, "Detailed diagnostics: in the formula\n  ");
  printwff(w);
  fprintf(WARNING_STREAM, "\nthe variable ");
  printwff(v);
  fprintf(WARNING_STREAM, " occurs outside the scope of a quantifier.\n");
  fprintf(WARNING_STREAM, "This is usually caused by a misplaced parenthesis.\n\n");
  list_vocabulary(w);
  
  fprintf(WARNING_STREAM,"\nParsed formula structure is:\n\n");
  indentwff(WARNING_STREAM,w,2,false);

  fprintf(WARNING_STREAM,"\n");
  skipout(0,SKIP);
}


void list_vocabulary(wff w)
// For error message drilldown purposes, list all the symbols in w
{
  f_stack vocab = 0;
  f_stack ptr;
  int x;

  ptr = extract_vocabulary(w,vocab);
  vocab = reverse_list(ptr);
  free_list(ptr);
  fprintf(WARNING_STREAM,"Vocabulary used in the formula is:\n");
  for (ptr=vocab; ptr; ptr=ptr->nxt) {
    if (ptr->f->category == Var) {
      fprintf(WARNING_STREAM,"  variable %s: ", ptr->f->fname);
      if (ptr->f->val_type)
	fprintf(WARNING_STREAM,"%s.\n",ptr->f->val_type->sname);
      else
	fprintf(WARNING_STREAM,"sort unknown.\n");
    }
    else if (ptr->f->adicity == 0) {
      fprintf(WARNING_STREAM,"  name %s: ", ptr->f->fname);
      if (ptr->f->val_type)
	fprintf(WARNING_STREAM,"%s.\n",ptr->f->val_type->sname);
      else
	fprintf(WARNING_STREAM,"sort unknown.\n");
    }
    else if (ptr->f->val_type == Bool) {
      if (ptr->f->category == Quant)
	fprintf(WARNING_STREAM,"  quantifier %s.\n", ptr->f->fname);
      else {
	for (x=0; x<ptr->f->adicity; x++)
	  if (ptr->f->arg_type[x] != Bool) break;
	if (x >= ptr->f->adicity)
	  fprintf(WARNING_STREAM,"  connective %s(", ptr->f->fname);
	else
	  fprintf(WARNING_STREAM,"  predicate %s(", ptr->f->fname);
	for (x=0; x<ptr->f->adicity; x++) {
	  if (x) fprintf(WARNING_STREAM,",");
	  if (ptr->f->arg_type[x])
	    fprintf(WARNING_STREAM,"%s",ptr->f->arg_type[x]->sname);
	  else
	    fprintf(WARNING_STREAM,"any_sort");
	}
	fprintf(WARNING_STREAM,").\n");
      }
    }
    else {
      fprintf(WARNING_STREAM,"  function %s(", ptr->f->fname);
      for (x=0; x<ptr->f->adicity; x++) {
	if (x) fprintf(WARNING_STREAM,",");
	if (ptr->f->arg_type[x])
	  fprintf(WARNING_STREAM,"%s",ptr->f->arg_type[x]->sname);
	else
	  fprintf(WARNING_STREAM,"any_sort");
      }
      fprintf(WARNING_STREAM,"): ");
      if (ptr->f->val_type)
	fprintf(WARNING_STREAM,"%s.\n",ptr->f->val_type->sname);
      else
	fprintf(WARNING_STREAM,"sort unknown.\n");
    }
  }
  free_list(vocab);
}

f_stack extract_vocabulary(wff w, f_stack vocab)
{
  f_stack ptr;
  int x;

  if (!w) return 0;
  for (ptr=vocab; ptr; ptr=ptr->nxt)
    if (ptr->f == w->sym)
      break;
  if (ptr)
    ptr = vocab;
  else {
    ptr = (f_stack) malloc(sizeof(F_STACK));
    ptr->f = w->sym;
    ptr->nxt = vocab;
    vocab = ptr;
  }
  for (x=0; x<w->sym->adicity; x++)
    vocab = extract_vocabulary(w->sb[x],vocab);
  return vocab;
}

f_stack reverse_list(f_stack source)
// Make a copy of the function stack in reverse order.
// Do not destroy or change the original.
{
  f_stack dest = 0;
  f_stack ptr;

  while (source) {
    ptr = (f_stack) malloc(sizeof(F_STACK));
    ptr->f = source->f;
    ptr->nxt = dest;
    dest = ptr;
    source = source->nxt;
  }
  return dest;
}

void free_list(f_stack s)
{
  if (s) {
    free_list(s->nxt);
    free(s);
  }
}

