#include "FINDER.h"

#ifdef SOFTIE_SCOTT
#include "SOFTIE_SCOTT.h"
#endif

void initial_initial(void)
// Called _right_ at the start, before even the command line switches
// are read. This is needed, as some things must be set by means of
// switches before reading the input, as they affect parts of the input
// parsing etc possibly before the settings section is read.
{
  not_searching_yet = true;
  piping = false;
  pipe_open = false;
  negative = false;
  dynstring_initialise(&command_line,"");
#ifdef WORLDWIDE
  setting.drilldown = true;
  setting.single_literal = true;
  setting.function_spec_mode = comp;
#else
  setting.drilldown = false;
  setting.single_literal = false;
  setting.function_spec_mode = math;
#endif
}


/*------------------------------------------------------------------------*/
void initialize(void)

     /*Called by init_finder().
      *The first function is called from main right at the start.
      *The name says it all.*/

{
  int i, j;
  functor f;
  kind k;

  /* Start at the beginning */
  current_line = 0;

  /* Ensure sets exist */
  S_initialise(10000);

  /* Initialise debugging information */
  init_events();

  /*---First zero all the array parts of array elements which are ------
   ----------to be null iff this element is undefined.-----------------*/

  for (i = 0; i < LEX_MAX; i++) {
    lexicon[i].word[0] = 0;
    lexicon[i].iskind = false;
    lexicon[i].offset = 0;
  }
  for (i = 0; i < ST_MAX; i++) sort[i].sname = 0;
  for (i = 0; i < FN_MAX; i++) {
    fn[i].fname = 0;
    fn[i].table = 0;
  }
  for (i = 0; i < WF_MAX; i++) form[i].sym = 0;
  for (i = 0; i < MAXISET_ARRAY; i++) {
    for (j = 0; j < CL_MAX; j++) {
      theory[i][j].prem = theory[i][j].conc = 0;
      theory[i][j].nextclause = theory[i][j].lastclause = 0;
    }
    size[i] = 0;
  }
  current_theory = 0;
  latest_clause.cid = latest_clause.pid = 0;
  latest_clause.prem = 0;
  latest_clause.conc = 0;
  latest_clause.testlist = 0;    
  latest_clause.varlist = 0;
  latest_clause.nextclause = 0;
  latest_clause.lastclause = 0;
  latest_clause.softweight = 0;
  latest_clause.preprocessed = false;
  latest_clause.tautology = false;
  FN = 0;
  next_qid = 1;     // because 0 is for free variables
  next_cid = 1;
  next_pid = 1;

  /*-------Set up the pre-defined types bool and int, and the ----------
   ------pre-defined relation and function symbols =, <, > and +.------*/

  k = newkind("bool");
  k->siz = k->sizmin = k->sizmax = 2;
  k->is_val = true;
  k->pred = (int*) malloc(3 * sizeof(int));
  k->pred[0] = k->pred[1] = 2; k->pred[2] = 0;
  k->succ = (int*) malloc(3 * sizeof(int));
  k->succ[0] = k->succ[2] = 2; k->succ[1] = 1;
  k->Ex = (int*) malloc(3 * sizeof(int));
  k->Ex[0] = 0; k->Ex[1] = k->Ex[2] = 1;
  k->len = (int*) malloc(2 * sizeof(int));
  canon(k, "FALSE");  canon(k, "TRUE");

  f = newfunctor("NOT", Func, 1, true);
  f->val_type = f->arg_type[0] = k;
  f->scope = 1;	f->table = neg;

  f = newfunctor("AND", Func, 2, true);
  f->val_type = f->arg_type[0] = f->arg_type[1] = k;
  f->scope = 10; f->table = cjctn;

  f = newfunctor("OR", Func, 2, true);
  f->val_type = f->arg_type[0] = f->arg_type[1] = k;
  f->scope = 10; f->table = djctn;

  f = newfunctor("IMP", Func, 2, true);
  f->val_type = f->arg_type[0] = f->arg_type[1] = k;
  f->scope = 10; f->table = impl;

  f = newfunctor("=", Func, 2, true);
  f->val_type = k;	f->scope = 9;	f->table = 0; // id;

  f = newfunctor("!=", Func, 2, true);
  f->val_type = k;	f->scope = 9;	f->table = 0; // nid;

  f = newfunctor("<", Func, 2, true);
  f->val_type = k;	f->scope = 9;	f->table = 0; // less;

  f = newfunctor(">", Func, 2, true);
  f->val_type = k;	f->scope = 9;	f->table = 0; // greater;

  k = newkind("int");
  k->sizmin = k->sizmax = (((WSZ-1)/10)*10+1);
  k->is_val = true;

  f = newfunctor("DIF", Func, 2, true);
  f->val_type = k;	f->scope = 5;	f->table = 0; // difference;	

  f = newfunctor("+", Func, 2, true);        /* The dummy typeless addition */
  f = newfunctor("-", Func, 2, true);        /* The dummy typeless subtraction */
  f = newfunctor("PRED", Func, 1, true);     /* The dummy typeless predecessor */
  f = newfunctor("SUCC", Func, 1, true);     /* The dummy typeless successor */
  f = newfunctor("EST", Func, 1, true);      /* The dummy typeless "exists" */
  f->val_type = Bool;
  f = newfunctor("LEN", Func, 1, true);      /* The dummy typeless length */
  f->val_type = Integer;
  f = newfunctor("MIN", Const, 0, true);     /* The dummy typeless zero */
  f = newfunctor("MAX", Const, 0, true);     /* The dummy typeless last element */
  f = newfunctor("SOME", Quant, 2, true); /* The existential quantifier */
  f->arg_type[1] = Bool;
  f->val_type = Bool;
  f = newfunctor("ALL", Quant, 2, true);  /* The universal quantifier */
  f->arg_type[1] = Bool;
  f->val_type = Bool;

  /*----------Defualt values for parameters ----------------------------*/

  setting.max_backtracks = 0;
  setting.max_invalues = 0;
  setting.max_flips = 0;
  setting.random = false;
  setting.dump_rate = 0;

#ifdef SOFTIE_SCOTT
  setting.pretest_limit = SHORT_CLAUSE_CUTOFF;  // synchronize short clause cutoff as set in SOFTIE_SCOTT.h
#else
  setting.pretest_limit = 3;  // the standard SFINDER pretest limit
#endif

#ifdef WORLDWIDE
  setting.max_models = 3;
  setting.batches = S_LARGE;
  setting.maxtime = 3;
  setting.pm.verbosity[pstats] = null;
  setting.print_unsat = true;
#else
  setting.max_models = 0;      /*solutions within a theory*/
  setting.batches = S_MAXIMAL;
  setting.maxtime = 0;
  setting.pm.verbosity[pstats] = brief;
  setting.print_unsat = false;
#endif
  setting.extra_batches = EXTRA_DEFAULT;
  setting.maxref = MAX_REF_SIZE;
  setting.test_maxref = MAX_REF_SIZE;
  setting.max_cellvals = CELLVAL_MAX;
  setting.target_bound = MAX_PENALTY;
  setting.hardweight = HARDWEIGHT;
  setting.search_is_local = false;
  setting.topsoff = true;
  setting.parse_only = false;
  setting.clausify_only = false;
  setting.vars_over_wt = true;
  setting.report_breaks = false;
  setting.print_headers = true;
  setting.noisy = false;
  setting.sortcut = true;
  setting.timestamps = false;
  setting.profile = false;
  setting.dumpdepth = false;
  setting.awnty_enabled = false;
  setting.cgv_enabled = false;
  setting.debug_enabled = false;
  setting.Nosecs = false;
  setting.invert_change = false;
  setting.pm.number_only = false;
  setting.pm.verbosity[pmodels] = full;
  setting.pm.verbosity[pjob] = null;
  setting.pm.verbosity[pspace] = null;
  setting.pm.pretty = true;
  setting.pm.incremental = false;
  setting.configmode = false;
  setting.configheuristic = true;   /* Does nothing if not in config mode */
  setting.outf = 0;
  setting.outi = 0;
 
  /*----------Place the addresses of the test functions in the -----------
   --------------- structures to be passed to transref ------------------*/

  init_trin(&tr_parameters);
  tr_parameters.Test = &test;
  tr_parameters.Onerefs = &setposs;
  tr_parameters.Smallrefs = &do_smallrefs;
  tr_parameters.Sizes = &get_sizes;
  tr_parameters.Printdomains = &print_domains;
  
  init_trin(&ttr_par);
  ttr_par.Test = model_t;
  ttr_par.Onerefs = single_t;
  ttr_par.Smallrefs = tuple_t;
  ttr_par.Sizes = sizes_t;
  ttr_par.maxref = setting.test_maxref;
  ttr_par.max_cellvals = 0;

  neg[0] = 2; neg[1] = 1; neg[2] = 0;
  cjctn[0] = 2; djctn[0] = 2; impl[0] = 2;
  for (i=0; i<2; i++)
    for (j=0; j<2; j++) {
      cjctn[(2*i)+j+1] = (i && j)? 1: 0;
      djctn[(2*i)+j+1] = (i || j)? 1: 0;
      impl[(2*i)+j+1] = ((!i) || j)? 1: 0;
    }
  zero = 0;

  for (i=0; i< MAXISET_ARRAY; i++) {
    for (j = 0; j < ST_MAX; j++) {
      last_sizes[i][j] = 0;
      last_sizmasks[i][j] = 0;
    }
    last_model[i] = 0;
    last_vlength[i] = 0;
    TH[i] = 0;
  }
  
  over_argmax = false;
  cmd_line_input = false;
  normal_input = true;
  single_model = false;
  cc_badguy = false;
  first_newval = true;
}
 /* initialize() */


/*------------------------------------------------------------------------*/
void canon(kind k, char new_name[])

     /*Called by initialise() and sort_spec().
      *This function adds a canonical name new_name to the enumeration
      *of the objects of sort k, setting sizmin and sizmax.*/

{
  name_list *lpp;
  functor f;

  f = newfunctor(new_name, Const, 0, true);
  f->val_type = k;

    /*count the number of elements in the enumeration -- the hard way*/
  k->siz = 1;
  for (lpp = &(k->canonical); *lpp; lpp = &((*lpp)->nextname))
    k->siz++;

  if (!(f->table)) f->table = (int*) malloc(sizeof(int));
  *(f->table) = k->siz-1;

  *lpp = (name_list) malloc(sizeof(NAME_LIST));

  (*lpp)->name = f->fname;
  (*lpp)->nextname = 0;

  k->sizmin = k->sizmax = k->siz;
} /* canon() */



/*------------------------------------------------------------------------*/
int lex_pos(char word[])

     /*Called by newkind() and newfunctor() (in this file)
      *and various functions in parser.c and input.c.
      
      *Lex_pos scans the lexicon for the given "word", returning
      *its position in the lexicon. 
      *If it is new to the lexicon then returns -1.*/
     
{
  int i;
  
  for (i = 0; lexicon[i].word[0]; i++)
    if (!strcmp(word,lexicon[i].word))
      return i;
  return -1; /* indicates that word is new */

} /* lex_pos */



/*------------------------------------------------------------------------*/
kind newkind(char new_name[])

     /*Called by initialise() and get_sort().
      *If the sort identifier with name "new_name" already exists, 
      *return a pointer to it. Otherwise add it to the list, put its name
      *in the lexicon and initialize it to the default values.*/

{
  int i,k;
  functor f;
  dynstring string;

  dynstring_initialise(&string, "");
  
  if ((i = lex_pos(new_name)) != -1){
    if (lexicon[i].iskind)
      return(sort+lexicon[i].offset);  /*return pointer to existing sort*/
    else
      inerror(new_name,"doubly defined as a sort and a functor",0);
  }
  /*find first undefined word*/
  for (i=0; (i < LEX_MAX) && lexicon[i].word[0]; i++);

  /*find first undefined sort*/
  for (k=0; (k < ST_MAX) && sort[k].sname; k++);

  lexicon[i].offset = k;
  lexicon[i].iskind = true;
  strcpy(lexicon[i].word,new_name);
  sort[k].sname = lexicon[i].word; /*sort points to its name in the lexicon*/
  sort[k].sizmin = 1;
  sort[k].siz = sort[k].maxvalue = 0;
  sort[k].sizmax = WSZ-1;
  sort[k].sizmask = 0;
  sort[k].canonical = 0;
  sort[k].len = 0;
  sort[k].is_val = false;
  sort[k].sum = sort[k].minus = 0;
  sort[k].pred = sort[k].succ = 0;
  sort[k].Ex = sort[k].len = 0;
  sort[k].piped = pipe_open;

  /* Define builtins for new sort */

  dynstring_strcpyc(&string,"+_"); dynstring_strcatc(&string,new_name);
  f = newfunctor(string.c_str, Func, 2, true);
  f->total = false;
  f->arg_type[0] = &sort[k];
  f->arg_type[1] = Integer;
  f->val_type = &sort[k];
  f->constant_table = true;

  dynstring_strcpyc(&string,"-_"); dynstring_strcatc(&string,new_name);
  f = newfunctor(string.c_str, Func, 2, true);
  f->total = false;
  f->arg_type[0] = &sort[k];
  f->arg_type[1] = Integer;
  f->val_type = &sort[k];
  f->constant_table = true;

  dynstring_strcpyc(&string,"PRED_"); dynstring_strcatc(&string,new_name);
  f = newfunctor(string.c_str, Func, 1, true);
  f->arg_type[0] = sort+k;
  f->val_type = sort+k;
  f->constant_table = true;
 
  dynstring_strcpyc(&string,"SUCC_"); dynstring_strcatc(&string,new_name);
  f = newfunctor(string.c_str, Func, 1, true);
  f->arg_type[0] = sort+k;
  f->val_type = sort+k;
  f->constant_table = true;
 
  dynstring_strcpyc(&string,"EST_"); dynstring_strcatc(&string,new_name);
  f = newfunctor(string.c_str, Func, 1, true);
  f->arg_type[0] = &sort[k];
  f->val_type = Bool;
  f->constant_table = true;
 
  dynstring_strcpyc(&string,"LEN_"); dynstring_strcatc(&string,new_name);
  f = newfunctor(string.c_str, Func, 1, true);
  f->arg_type[0] = sort+k;
  f->val_type = Integer;
  f->constant_table = true;

  dynstring_strcpyc(&string,"MIN_"); dynstring_strcatc(&string,new_name);
  f = newfunctor(string.c_str, Const, 0, true);
  f->val_type = sort+k;
  f->table = &zero;
  f->constant_table = true;

  dynstring_strcpyc(&string,"MAX_"); dynstring_strcatc(&string,new_name);
  f = newfunctor(string.c_str, Const, 0, true);
  f->val_type = &sort[k];
  f->table = &(sort[k].maxvalue);
  f->constant_table = true;

  dynstring_free(&string);

  return &sort[k];
} /* newkind() */




/*------------------------------------------------------------------------*/
functor newfunctor(char new_name[], syntype category, int adicity, boolean dec)

     /*Called by initialize(), newsort(), and conon() (in this file)
      *and by input.c and parser.c.
      *If the functor identifier with name "new_name" already exists, 
      *in ther functor stack fn[] return a pointer to it. 
      *Otherwise add it to the stack, put its name in the lexicon and 
      *initialize it to the default values. */
{
  int i, j;
  functor f;

  if (category != Const && isdigit(new_name[0]))
    inerror(new_name,
	    "defined as a function symbol but starts numerically",0);

  if (adicity > ARGMAX) inerror(new_name,"given too many argument places",0);
  
  if ((i = lex_pos(new_name)) != -1){
    if (!lexicon[i].iskind)
      return(fn+lexicon[i].offset);
    else
      inerror(new_name,"doubly defined as a sort and a functor",0);
  }
  /*find first undefined word   */
  for (i = 0; (i < LEX_MAX) && lexicon[i].word[0]; i++); 

  /*find first undefined functor*/
  for (j = 0; (j < FN_MAX) && fn[j].fname; j++);

  lexicon[i].offset = j;
  lexicon[i].iskind = false;
  strcpy(lexicon[i].word,new_name);
  fn[j].fname = lexicon[i].word;
  fn[j].category = category;
  fn[j].scope = adicity? adicity-1: 0;
  fn[j].adicity = adicity;
  fn[j].constant_table = false;

  if (category == Var)
    inerror(new_name,"both a variable and a functor",0);
  else if (category==Const &&
	   strncmp(fn[j].fname,"MIN_",4) &&
	   strcmp(fn[j].fname,"MIN_") &&
	   strncmp(fn[j].fname,"MAX_",4) &&
	   strcmp(fn[j].fname,"MAX")) {
    fn[j].table = (int*) malloc(sizeof(int));
    *fn[j].table = 0;
  }
  else
    fn[j].table = 0;
  for (i=0; i<ARGMAX; i++) fn[j].fats[i] = 1;
  if (!strcmp(new_name,"AND") ||
      !strcmp(new_name,"OR") ||
      !strcmp(new_name,"IMP")) {
    fn[j].fats[0] = 2;
    fn[j].fats[1] = 1;
  }

  fn[j].fcc = 0;
  for (i=0; i < MAXISET_ARRAY; i++) fn[j].old_fcc[i] = 0;
  fn[j].total = true;
  fn[j].injective = fn[j].leftin = fn[j].rightin = false;
  fn[j].surjective = fn[j].leftsur = fn[j].rightsur = false;
  fn[j].surj_set = false;
  fn[j].comm = false;
  fn[j].nop = false;
  fn[j].cut_it = false;
  fn[j].piped = pipe_open;
  fn[j].labels = true;
  fn[j].cgv = false;
  fn[j].configurable = false;
  fn[j].declared = dec;
  for (i=0; i<ARGMAX; i++) fn[j].arg_type[i] = 0;
  fn[j].val_type = 0;
  fn[j].pmode = null;
  fn[j].nextfunct = 0;

  /*Append it to the testable functor stack FN[] */
  if (category != Const &&
      category != Quant &&
      !strchr("+-<>=",new_name[0])) {
    if (strcmp(new_name,"EST")      && strcmp(new_name,"LEN")      &&
	strncmp(new_name,"EST_",4)  && strncmp(new_name,"LEN_",4)  &&
	strcmp(new_name,"PRED")     && strcmp(new_name,"SUCC")     &&
	strncmp(new_name,"PRED_",5) && strncmp(new_name,"SUCC_",5) &&
	strcmp(new_name,"NOT")      && strcmp(new_name,"DIF")      &&
	strcmp(new_name,"AND")      && strcmp(new_name,"OR")       &&
	strcmp(new_name,"IMP")      && strcmp(new_name,"!=")) {
      if (!FN) FN = &fn[j];
      else {
	for (f = FN; f->nextfunct; f = f->nextfunct);
	f->nextfunct = &fn[j];
      }
    }
  }
  
  return &fn[j];
} /* newfunctor() */


/*
 * ...and because the above allocates memory...
 */

void free_tables(void)
{
  int i;

  for (i=0; i<FN_MAX; i++)
    free_table(fn+i);
}


void free_table(functor f)
{
  if (!f->fname) return;
  if (f->constant_table) return;
  if (strchr("=><+-",f->fname[0])) return;
  if (!strcmp("!=",f->fname) ||
      !strcmp("DIF",f->fname) ||
      !strcmp("NOT",f->fname) ||
      !strcmp("AND",f->fname) ||
      !strcmp("OR",f->fname) ||
      !strcmp("IMP",f->fname)) return;
  if (f->table) {
    free(f->table);
    f->table = 0;
  }
}

int hardweight(void){return HARDWEIGHT;}

/*=============================== END =====================================*/
