// This file and fclausify.c need major overhaul. Variables should be
// subscripted with their type, so that the same one is referenced if
// the clause is re-parsed. Maybe use a colon x:t as is standard. In
// foutput.c there needs to be a function returning the clause as a
// string with or without connectives and quantifiers. Clauses need to
// be identified by id number. Also keep the input formulae in case
// that's useful.
// 
// The clausifier needs to do some renaming, especially for cases
// where there is a threat of exponential blow-out and where
// quantifiers occur as subformulae. NOTE: MAYBE NOT.

/*
*			parser.c
*/


#include "FINDER.h"

static WFF newcl[NEWCL_MAX]; /*a parse wff stack for the new clause */
static boolean type_added;

static PSW parse_wffs[NEWCL_MAX]; /*parse() fcn wff stack*/
static int psw_next;    /* index of next (sub)formula to be parsed */
static int previous_end;
static char* current_vars[LEX_MAX];  /* all variables in this clause */

/*------------------------------------------------------------------------*/
void free_wff_list(wff_stack ptr)

/*Called by end_parse()* and free_clause_parts()*/

/* Don't free the formulae, but free the list of pointers to them.
   Should zero the list head after this has been called */

{
  if (ptr) {
    free_wff_list(ptr->nxt);
    ptr->nxt = 0;
    free(ptr);
  }

} /* free_wff_list() */



/*------------------------------------------------------------------------*/
void begin_parse(void)

     /*Called by get_cl().
      *Initializes the temporary wff stack and the various lists of the 
      *current clause before parsing starts.*/

{
  int i;

  for (i = 0; i < NEWCL_MAX; i++) {
    newcl[i].sym = 0;
    newcl[i].val = -1; /* dummy value */
    parse_wffs[i].m = false;
  }
  free_clause_parts(&latest_clause);
  psw_next = 0;

  /* zero the variable names */
  for (i=0; i<LEX_MAX; i++) {
    if (current_vars[i])
      free(current_vars[i]);
    current_vars[i] = 0;
  }

  /* get last name on the functor stack */
  for (i=0; fn[i].fname; i++);
  previous_end = i;
} /* begin_parse() */


boolean is_a_variable(char *s)
{
  int i;

  for (i=0; i<LEX_MAX && current_vars[i]; i++)
    if (!strcmp(s,current_vars[i]))
      return true;
  return false;
}

void add_a_variable(char *s)
{
  int i;

  for (i=0; i<LEX_MAX && current_vars[i]; i++)
    if (!strcmp(s,current_vars[i]))
      return;
  if (i >= LEX_MAX)
    skipout("Too many variables in clause",SKIP);
  current_vars[i] = (char*)malloc(sizeof(char) * strlen(s)+1);
  // size of char should be 1, but justr in case...
  strcpy(current_vars[i],s);
}


/*------------------------------------------------------------------------*/
wff parse(dynstring *symbol, binding_list b)
     
     /*Called by sparse() and get_cl().
      *Parsing routine which allows infix binary function symbols.

      *A general wff (gwf) consists of one or more simple wffs (swf) 
      *separated by binary functors.  
      *An swf is either a functor followed by the appropriate number 
      *of swf to form its arguments or a gwf in parentheses.

      *Hence we need two interacting parse functions, one for 
      *gwf (parse) and one for swf (sparse).  
      *In each case the symbol[] parameter holds the first symbol of 
      *the sequence to be parsed. After the parse it holds the symbol 
      *which follows the formula. 

      *Returns the fully parsed wff*/

{
  psw pp, qq;
  functor mf;
  int i;
  pp = &(parse_wffs[psw_next]);

  if (++psw_next >= NEWCL_MAX) inerror("Too many formulae parsed","",0);

  pp->leftsub = pp->rightsub = 0;
  pp->main = 0;
  pp->actual = sparse(symbol,b,false); /* get first swf */

  get_word(symbol);
    /* if next word is not a terminator,
       then it must be a binary operator */

  while (!strchr(".,;)",symbol->c_str[0]) && strncmp("->",symbol->c_str,2)) { 
    i = lex_pos(symbol->c_str);
    if (i < 0) {
      inerror_break_line(symbol->c_str,
			 "found where ',' or '.' or a function symbol was expected",true);
      infix_drilldown(pf_transcribe(pp),symbol->c_str);
    }
    mf = &(fn[lexicon[i].offset]);
    if (mf->category != Func || mf->adicity != 2) {
      inerror_break_line(symbol->c_str,
			 "found where a binary operator was expected",true);
      infix_drilldown(pp->actual,symbol->c_str);
    }
    qq = (psw) malloc(sizeof(PSW));
    qq->leftsub = qq->rightsub = 0;
    qq->main = 0;
    qq->m = true;

    get_word(symbol);
    qq->actual = sparse(symbol,b,false); /* get the second argument of the 
				      binary operator */

    pw_merge(&pp, qq, mf);     /* merge these structures into 
				    something more resonable */

    get_word(symbol);
  }
  return pf_transcribe(pp);

} /* parse() */




/*-----------------------------------------------------------------------*/
void pw_merge(psw *pp, psw qq, functor mf)

     /*Called by parse().
       
      *Recursive routine for converting infix to prefix forms.
      *Puts the next functor and subformula into the tree of 
      *partially parsed subformulas. 
      *Where the tree so far is, say
      *                     pp
      *	                   /  \
      *	                  L    R
      *		              / \
      *		            Rl   Rr
      *
      *the result of merging (...qq mf) with it will be either
      *
      *		pp        or else       mf
      *	       /  \                    /  \
      *	      L    R                  pp   qq
      *		  / \                / \
      *		 Rl  mf             L   R
      *		    / \                / \
      *		   Rr  qq            Rl   Rr
      *
      *Note we pass pp by address since memory has already been 
      *allocated and this is what we want to return.*/

{
  psw rr;

    /*right diagram*/
  if (!(*pp)->main || mf->scope >= (*pp)->main->scope) {
    rr = (psw) malloc(sizeof(PSW));
    rr->leftsub = *pp;
    rr->rightsub = qq;
    rr->main = mf;
    rr->actual = 0;
    rr->m = true;
    *pp = rr;
    return;
  }

    /*else left diagram*/
  pw_merge(&((*pp)->rightsub), qq, mf);
  return;

} /* pw_merge()*/



/*-------------------------------------------------------------------------*/
wff pf_transcribe(psw pp)

     /*Called by parse() and itself.
      *Pf_transcribe is part of getting the semi-parsed inside of a pair of
      *parentheses into fully transcribed form. It returns the fully parsed 
      *subformula whose partially parsed version is rooted at pp.*/

{
  wff subforms[2];
  wff rc;

  if (!pp) return 0;
  if (pp->actual) rc = pp->actual;
  else {
    subforms[0] = pf_transcribe(pp->leftsub);
    subforms[1] = pf_transcribe(pp->rightsub);
    rc = locate(pp->main, subforms);
  }
  if (pp->m) free(pp);
  return rc;

} /* pf_transcribe() */



/*-------------------------------------------------------------------------*/
wff sparse(dynstring *symbol, binding_list b, boolean ignore_FOF)

     /*Called by parse() and itself.
      *Parse the cases other than that of a parenthesised subformula.
      *Returns the fully parsed simple wff. */

{    
  wff parsed_subform, args[ARGMAX];
  functor f = 0;
  int i, pos;
  binding_list bptr;

  /*if it's in parentheses call parse() to deal with it*/
  if (symbol->c_str[0] == '(') {
    get_word(symbol);
    parsed_subform = parse(symbol,b);
    if (symbol->c_str[0] != ')') {
      inerror_break_line(symbol->c_str,
			 "found where a right parenthesis was expected",true);
      right_parenthesis_drilldown(parsed_subform,symbol->c_str);
    }
    return parsed_subform;
  }

  /*otherwise it's either a quantifier, a function, a variable or a number.*/
  for (i=0; i<ARGMAX; i++) args[i] = 0;
  pos = lex_pos(symbol->c_str);
  if (pos < 0) {
    if (isalpha(symbol->c_str[0])) {                /* ok, its a varaiable */
      for (bptr=b; bptr; bptr=bptr->next)
	if (!strcmp(bptr->v,symbol->c_str)) {
	  add_a_variable(symbol->c_str);            /* should be redundant */
	  return locate(newvar(symbol->c_str,bptr->q,0),args);
	}
      if (!bptr) {
	if (setting.single_literal && !ignore_FOF) {
                                                   /* FOF: no free variables */
	  f = newfunctor(symbol->c_str, Func, 0, false);
	  f->val_type = 0;                         /* type TBA */
	  f->pmode = full;
	  if (is_a_variable(f->fname)) {
	    inerror_break_line(f->fname,
			       "used as both a variable and a name",true);
	    bad_variable_drilldown(f);
	  }
	  return locate(f, args);
	}
	else {
	  add_a_variable(symbol->c_str);
	  return locate(newvar(symbol->c_str,0,0),args);
	}
      }
    }
    
    else if (isdigit(symbol->c_str[0])) {         /* ok, its a number    */
      for (i = 0; symbol->c_str[i]; i++)
	if (!isdigit(symbol->c_str[i]))
	  inerror_break_line(symbol->c_str,
			     "starts numerically but isn't a number",false);
      /* make a new constant for that number */
      f = newfunctor(symbol->c_str, Const, 0, false);
      f->val_type = Integer;
      *(f->table) = 0;
      for (i=0; symbol->c_str[i]; i++) {
	*(f->table) *= 10;
	*(f->table) += (symbol->c_str[i] - '0');
      }
      return locate(f, args);
    }
    else inerror_break_line("Formula expected, but a formula cannot start with",
			    symbol->c_str,false);
  }
  
  if (lexicon[pos].iskind)
    inerror("Expected function or predicate but found sort name",
	    symbol->c_str,false);
  f = &fn[lexicon[pos].offset];                 /* function or quantifier */

  if (f->category == Quant) {                   /* quantifier */
    get_word(symbol);
    args[0] = sparse(symbol,b,true);
    if (args[0]->sym->category != Var) {
      inerror_break_line("","expected variable after quantifier",true);
      quantifier_drilldown(f,args[0]);
    }
    bptr = (binding_list) malloc(sizeof(BINDING_LIST));
    bptr->q = next_qid++;
    strcpy(bptr->v,args[0]->sym->fname);
    args[0] = locate(newvar(bptr->v,bptr->q,0),0);
    bptr->next = b;
    get_word(symbol);
    args[1] = sparse(symbol,bptr,false);           // Recursive call
    free(bptr);                                    // bptr's job is now done
    {
      char extension[10];                          // Limit 10^7 variables (!)
      sprintf(extension,"__%d",args[0]->sym->binder);
      strcat(args[0]->sym->fname,extension);
    }
  }

  else if (f->adicity > 0) {                    /* function */
    for (i=0; i<f->adicity; i++) args[i] = 0;
    get_word(symbol);
    if (symbol->c_str[0] == '(') {
      for (i=0; i<f->adicity; i++) {
	get_word(symbol);
	args[i] = parse(symbol,b);
	if (i == f->adicity-1) {
	  if (symbol->c_str[0] != ')') {
	    inerror_break_line(symbol->c_str,
			       "found where a right parenthesis was expected",true);
	    last_argument_drilldown(f,args,symbol->c_str);
	  }
	}
	else {
	  if (symbol->c_str[0] != ',') {
	    inerror_break_line(symbol->c_str,
			       "found where a comma was expected",true);
	    nonlast_argument_drilldown(f,args,symbol->c_str);
	  }
	}
      }
    }
    else if (f->adicity == 1) {
      args[0] = sparse(symbol,b,ignore_FOF);
    }
    else {
      inerror_break_line(symbol->c_str,
			 "found where a left parenthesis was expected",true);
      parenthesis_after_function_drilldown(f,symbol->c_str);
    }
  }

  return locate(f, args);

} /* sparse() */


/*-------------------------------------------------------------------------*/
wff locate(functor main_func, wff subform[ARGMAX])

     /*Called by sparse() and pf_transcribe().
      *Returns a pointer to the (sub)formula with main functor main_func
      *and argument list subform[].
      *This is especialy important for finding pre-existing variables.

      *Note that the formula is located in the "parse wff stack" newcl[], 
      *not in form[]. This is where newcl[] gest built.
     
      *This is sometime called with a null 2nd argument. Would prefer to
      *do something more elegant for this case.*/

{
  int i, j;
  boolean subs_match;
  
  /* search for formula in the stack */
  for (i = 0; newcl[i].sym; i++){
    if (i >= NEWCL_MAX) inerror("Clause is too long","",0);
    if (newcl[i].sym->fname == main_func->fname) {

      if (main_func->category == Var || main_func->category == Const) 
	return &(newcl[i]); /*found it!!*/

      /*compare subformulae*/
      subs_match = true;
      for (j=0; j<main_func->adicity; j++)
	if (subform[j] != newcl[i].sb[j])
	  subs_match = false;
      if (subs_match) return &(newcl[i]); /* found it!! */
    }
  }

  /*Otherwise we have to add it to the stack -- the i-th element*/

  if (i >= NEWCL_MAX) inerror("Clause is too long","",0);
  newcl[i].sym = main_func;
  newcl[i].matrix = main_func->table;
  newcl[i].wcc = -1;

  /* This is not enough, as MAX_* doesn't have a value yet */
  if (main_func->category == Const && main_func->table)
    newcl[i].val = *(main_func->table);

  /* set any subformulae */
  for (j=0; j<ARGMAX; j++) {
    newcl[i].sb[j] = (j<main_func->adicity? subform[j]: 0);
    newcl[i].sbval[j] = (j<main_func->adicity? &(subform[j]->val): &zero);
  }

  return &(newcl[i]);

} /* locate() */



/*-------------------------------------------------------------------------*/
void typecheck(void)
     
     /*Called by get_cl().
      *Checking for agreement of types, especially for variables and for
      *typically ambiguous symbols like + and =, is a messy business.*/
     
{
  wff_stack wf;
  wff bad_formula = 0;
  wff bad_subformula = 0;


  /* add any missing types */
  do {
    type_added = false;
    for (wf = latest_clause.prem; wf; wf = wf->nxt) type_needed(wf->w);
    for (wf = latest_clause.conc;wf; wf = wf->nxt) type_needed(wf->w);
  }
  while (type_added);

  /*check for completness */
  for (wf = latest_clause.prem; wf && !bad_subformula; wf = wf->nxt) {
    bad_subformula = type_complete(wf->w);
    if (bad_subformula) bad_formula = wf->w;
  }
  for (wf = latest_clause.conc; wf && !bad_subformula; wf = wf->nxt) {
    bad_subformula = type_complete(wf->w);
    if (bad_subformula) bad_formula = wf->w;
  }
  if (bad_subformula)
    type_missing_drilldown(bad_formula,bad_subformula);

  /*check for consistency */
  for (wf = latest_clause.prem;wf; wf = wf->nxt) type_consistent(wf->w);
  for (wf = latest_clause.conc;wf; wf = wf->nxt) type_consistent(wf->w);

} /* typecheck() */



/*-------------------------------------------------------------------------*/
void type_needed(wff w)

     /*Called by typecheck() and itself.
      *Not every symbol is born with a type. This is the function for 
      *adding types to symbols without. There are various cases, for
      *sort-relative symbols like + and EST, and then for variables.
      *Types have to be deduced from the context.*/

{
  int i;

  if (!w) return;
  if (!strcmp(w->sym->fname,"+")) AT_plus(w);
  if (!strcmp(w->sym->fname,"-")) AT_plus(w);
  if (!strcmp(w->sym->fname,"EST")) AT_up(w, "EST_");
  if (!strcmp(w->sym->fname,"LEN")) AT_up(w, "LEN_");
  if (!strcmp(w->sym->fname,"PRED")) AT_up(w, "PRED_");
  if (!strcmp(w->sym->fname,"SUCC")) AT_up(w, "SUCC_");

  /*if argument types are known, check each argument
    --- what about outermost symbol? 
    --- does equality go thru' here? No, "=" has no argument types.*/
  for (i=0; i<w->sym->adicity; i++)
    if (w->sym->arg_type[i] && !w->sb[i]->sym->val_type)
      add_type(w->sb[i],w->sym->arg_type[i]);

  /* types of predefined binary ops must match !!*/
  if (w->sym->adicity == 2 && 
      (strchr("=<>",w->sym->fname[0]) || 
       !strcmp("DIF",w->sym->fname) ||
       !strcmp("!=",w->sym->fname))) {
    if (w->sb[0]->sym->val_type && !w->sb[1]->sym->val_type)
      add_type(w->sb[1], w->sb[0]->sym->val_type);
    if (w->sb[1]->sym->val_type && !w->sb[0]->sym->val_type)
      add_type(w->sb[0], w->sb[1]->sym->val_type);
  }

  /*check any subformulae */
  for (i=0; i<w->sym->adicity; i++)
    type_needed(w->sb[i]);

}/* type_needed() */



/*-------------------------------------------------------------------------*/
void add_type(wff w, kind k)
     
     /*Called by type_needed() and AT_plus().
      *Gives w the  type k. 
      *Variables get their types this way -- same as kind. */

{
  dynstring string;

  dynstring_initialise(&string, "");
  
  if (!strcmp(w->sym->fname,"MIN") ||
      !strcmp(w->sym->fname,"MAX") ||
      !strcmp(w->sym->fname,"+") ||
      !strcmp(w->sym->fname,"-") ||
      !strcmp(w->sym->fname,"EST") ||
      !strcmp(w->sym->fname,"LEN") ||
      !strcmp(w->sym->fname,"PRED") ||
      !strcmp(w->sym->fname,"SUCC")) {
    dynstring_sprintf(&string, "%s_%s", w->sym->fname, k->sname);
    w->sym = fn+lexicon[lex_pos(string.c_str)].offset;
    w->matrix = w->sym->table;
    if (w->sym->category == Const)
      w->val = *(w->sym->table);
  }
  else if (w->sym->category == Var) {
    if (w->sym->binder)
      w->sym->val_type = k;
    else
      w->sym = newvar(w->sym->fname,w->sym->binder,k); /* get variable types */
  }
  else if (w->sym->adicity == 0) {
    w->sym->val_type = k;
  }
  else {
    inerror("Apparently missing type for",undecorated(w->sym->fname),0);
    /* No drill down for this one, as I don't understand it! */
  }
  type_added = true;

  dynstring_free(&string);
} /* add_type() */



/*-------------------------------------------------------------------------*/
void AT_plus(wff w)

     /*Called by type_needed().
      *Ensure that in formula like (A + B) the + is of the same type 
      *as the A.*/
     
{
  if (w->sb[0]->sym->val_type)
    add_type(w, w->sb[0]->sym->val_type);

} /* AT_plus() */



/*-------------------------------------------------------------------------*/
void AT_up(wff w, char *s)

     /*Called by type_needed().
      *Ensure that in formulae such as LEN(A) or EST(A), the type of
      *the predicate is determined by that of its argument.*/

{
  char string[LEXLENGTH];

  if (w->sb[0]->sym->val_type) {
    strcpy(string, s);
    strcat(string, w->sb[0]->sym->val_type->sname);
    w->sym = fn+lexicon[lex_pos(string)].offset;
    w->matrix = w->sym->table;
    type_added = true;
  }
} /* AT_up() */



/*-------------------------------------------------------------------------*/
functor newvar(char s[], int q, kind k)
     
     /*Called by sparse() and add_type(). 
      *Returns the variable with given name, binder and sort.*/

{
  functor f;
  int i;

  for (i=0; (i < FN_MAX) && fn[i].fname; i++) {
    if (!strcmp(s,fn[i].fname)) {
      if (fn[i].category != Var)
	inerror(undecorated(s),"used as a variable and a function symbol",0);
      if (fn[i].binder == q && fn[i].val_type == k)
	return &(fn[i]);
    }
  }

  if (i >= FN_MAX) inerror("FN_MAX exceeded: too many functors!","",0);
  f = &(fn[i]); /* set to next open position */

  /*copy variable to the list of functors but not to the lexicon*/
  f->fname = (char*) malloc(strlen(s)+1);
  strcpy(f->fname,s);
  f->category = Var;
  f->scope = 0;
  f->adicity = 0;
  f->table = 0;
  f->fcc = 0;
  for (i=0; i<MAXISET_ARRAY; i++) fn->old_fcc[i] = 0;
  for (i=0; i<ARGMAX; i++) f->fats[i] = 0;
  f->total = true;
  f->injective = false;
  f->cut_it = false;
  for (i=0; i<ARGMAX; i++)
    f->arg_type[i] = 0;    /* variables have no arguments */
  f->val_type = k;         /* same as the kind passed if any */
  f->binder = q;           /* and bound by the same quantifier if any */
  f->pmode = null;
  return f;
  
} /* newvar() */



/*-------------------------------------------------------------------------*/
wff type_complete(wff w)

     /*Called by typecheck() and itself.
      *Checks that every symbol in the current formula has a
      *value type. Some (like `=') have no argument types.
      *Return pointer to a subformula which lacks a type.*/

{
  int i;
  wff rc = 0;

  if (!w) return 0;
  if (!w->sym->val_type) {
    inerror("Type missing for",undecorated(w->sym->fname),true);
    rc = w;
  }
  for (i=0; i<w->sym->adicity && !rc; i++)
    rc = type_complete(w->sb[i]);
  return rc;

} /* type_complete() */


/*-------------------------------------------------------------------------*/
void type_consistent(wff w)

     /*Called by typecheck() and itself.
      *After everything has a type, there may be errors. This function 
      *goes recursively through a formula looking for type mismatches.*/

{
  int i;

  if (strchr("=<>",w->sym->fname[0]) || 
      !strcmp("DIF",w->sym->fname) || 
      !strcmp("!=",w->sym->fname)) {
    if (w->sb[0]->sym->val_type != w->sb[1]->sym->val_type) {
      inerror("Type mismatch: arguments of",w->sym->fname,true);
      type_mismatch_drilldown(w,0);
    }
  }
  else if (w->sym->category == Quant) {
    if (w->sym->arg_type[1] != w->sb[1]->sym->val_type) {
      inerror("Type mismatch: non-boolean follows",w->sym->fname,true);
      quantifier_type_mismatch_drilldown(w);
    }
  }
  else
    for (i=0; i<w->sym->adicity; i++)
      if (w->sym->arg_type[i] != w->sb[i]->sym->val_type) {
	inerror("Type mismatch with argument of",w->sym->fname,true);
	type_mismatch_drilldown(w,i);
      }
  
  for (i=0; i<w->sym->adicity; i++)
    type_consistent(w->sb[i]);

} /* type_consistent() */



/*-------------------------------------------------------------------------*/
boolean trivial_cl(void)

     /*Called by get_cl().
      *Return TRUE iff the functors in the new clause are all
      *trivially satisfied.*/

{
  int i;
  
  for (i=0 ; newcl[i].sym ;i++) {
    if (i >= NEWCL_MAX) inerror("Clause is too long","",0);
    if (newcl[i].sym->category == Var) continue;
    if (strchr("+=<>",newcl[i].sym->fname[0])) continue;
    if (!strcmp("equal",newcl[i].sym->fname)) continue;
    if (!strcmp("!=",newcl[i].sym->fname)) continue;
    return false;
  }
  return true;

}/* trivial_cl() */


/*-------------------------------------------------------------------------*/
void end_parse(clause c)

     /* Called by get_cl() and clausify().
      * After the current clause has been parsed, this is called to
      * set up the various lists (varlist and testlist) and to set the
      * pointers to subformula values correctly. Also called after
      * conversion to pure clause form, with each of the resulting
      * clauses.*/

{
  wff_stack wp;
  int i;

  if (!disjoint_wffstacks(c->prem,c->conc))
    c->tautology = true;

  free_wff_list(c->varlist);
  c->varlist = 0;
  free_wff_list(c->testlist);
  c->testlist = 0;

  for (wp=c->prem; wp; wp=wp->nxt)
    c->varlist = get_variables_in_list(wp,c->varlist);
  for (wp=c->conc; wp; wp=wp->nxt)
    c->varlist = get_variables_in_list(wp,c->varlist);
  for (wp=c->prem; wp; wp=wp->nxt)
    c->testlist = get_testables_in_list(wp,c->testlist);
  for (wp=c->conc; wp; wp=wp->nxt)
    c->testlist = get_testables_in_list(wp,c->testlist);
  inner_first(c->testlist);

  if (c==&latest_clause)
    for (i=0; i < NEWCL_MAX && newcl[i].sym; i++)
      newcl[i].matrix = newcl[i].sym->table;

  /* set the value pointers */
  for (wp = c->prem; wp; wp = wp->nxt)
    set_valptrs(wp->w);
  for (wp = c->conc; wp; wp = wp->nxt)
    set_valptrs(wp->w);

  /* do something obscure with positive last clauses */
  if (c->conc &&
       !(c->conc->nxt) &&
       *(c->conc->w->sym->fname) == '=' &&
       is_open(c->conc->w->sb[0]->sym) &&
       !occurs_inside(c->conc->w->sb[0],
		      c->conc->w->sb[1]) &&
       !occurs_in_list(c->conc->w->sb[0],
		       c->prem))
    positive_last(c);


} /* end_parse() */


boolean disjoint_wffstacks(wff_stack w1, wff_stack w2)
// Check that there is no wff in common
// This is the stupid quadratic time method, but the lists are fairly short...
{
  wff_stack w1p, w2p;

  for (w1p=w1; w1p; w1p=w1p->nxt)
    for (w2p=w2; w2p; w2p=w2p->nxt)
      if (w1p->w == w2p->w)
	return false;
  return true;
}


wff_stack get_variables_in_list(wff_stack wp, wff_stack varlist)
{
  if (wp) {
    varlist = get_variables_in_list(wp->nxt,varlist);
    varlist = get_variables_in_wff(wp->w,varlist);
  }
  return varlist;
}

wff_stack get_variables_in_wff(wff w, wff_stack varlist)
{
  int x;

  if (w->sym->category == Var)
    varlist = add_wff(varlist,w);
  else
    for (x=0; x<w->sym->adicity; x++)
      varlist = get_variables_in_wff(w->sb[x],varlist);
  return varlist;
}


wff_stack get_testables_in_list(wff_stack wp, wff_stack testlist)
{
  if (wp) {
    testlist = get_testables_in_list(wp->nxt,testlist);
    testlist = get_testables_in_wff(wp->w,testlist);
  }
  return testlist;
}

wff_stack get_testables_in_wff(wff w, wff_stack testlist)
{
  int x;

  if (w) {
    if (w->sym->category != Var && w->sym->category != Const)
      testlist = add_wff(testlist,w);
    for (x=0; x<w->sym->adicity; x++)
      testlist = get_testables_in_wff(w->sb[x],testlist);
  }
  return testlist;
}


/*
 * Ensure that the subformula relation is embedded in the testlist.
 * Depends on the fact that all formulae are testable except for
 * variables and constants (which do not have subformulae).
 */

void inner_first(wff_stack testlist)
{
  wff_stack ptr1, ptr2;
  int i;
  wff tempwff;

  for (ptr1=testlist; ptr1; ptr1=ptr1->nxt) {
    for (ptr2=testlist; ptr2 != ptr1; ptr2=ptr2->nxt) {
      for (i=0; ptr2 && i<ptr2->w->sym->adicity; i++)
	if (ptr2->w->sb[i] == ptr1->w) {
	  tempwff = ptr1->w;
	  ptr1->w = ptr2->w;
	  ptr2->w = tempwff;
	  ptr1=ptr2;
	  ptr2 = 0;
	}
      if (!ptr2) break;
    }
  }
}


/*-------------------------------------------------------------------------*/
void set_valptrs(wff wf)

     /*Called by end_parse().
      *Sets the subformula value pointers.*/

{
  int i;

  if (!wf) return;
  if (wf->val >= 0) return;

  /* foreach subformula */
  for (i=0; i<wf->sym->adicity; i++)
    set_valptrs(wf->sb[i]);

  for (i=0; i<ARGMAX; i++)
    wf->sbval[i] = (i<wf->sym->adicity? &(wf->sb[i]->val): &zero);

} /* set_valptrs() */



/*-------------------------------------------------------------------------*/
void positive_last(clause c)

     /*Called by end_parse().
      *Certain Horn clauses can be treated specially with one entry in
      *the list of refutation points counted as "positive". The positive
      *point must be moved to the end of the testlist so that any values
      *needed to evaluate it (which need not be in its subformulae) will
      *have been calculated before it is encountered.
      *
      *This is rather ugly code, but inefficiency here is of no great
      *significance since the testlist is always fairly short.
      */

{
  wff_stack lc, cls, wp_up, wp_down;
  /* latest clause conclusion, its left sub, two pointers */
  
  for (lc=c->testlist; lc; lc=lc->nxt)
    if (lc->w == c->conc->w)
      break;
  if (!lc) return;
  /*
   * Left subformula of lc->w is known not to occur on the premise
   * side, so nothing dominates lc, so it can be moved to the end
   */
  if (lc == c->testlist && lc->nxt) { /* Ensure testlist exists */
    c->testlist = lc->nxt;
    lc->nxt = c->testlist->nxt;
    c->testlist->nxt = lc;
  }
  for (wp_down=c->testlist; wp_down; wp_down=wp_down->nxt)
    if (wp_down->nxt == lc)
      break;
  while (lc->nxt) {
    wp_up = lc->nxt;
    lc->nxt = wp_up->nxt;
    wp_up->nxt = lc;
    wp_down->nxt = wp_up;
    wp_down = wp_up;
  }

  for (cls = c->testlist; cls; cls = cls->nxt)
    if (cls->w == lc->w->sb[0])
      break;
  if (!cls) return;
  /*
   * Nothing except lc dominates cls, so it can be moved up too
   */
  if (cls == c->testlist && cls->nxt != lc) {
    c->testlist = cls->nxt;
    cls->nxt = c->testlist->nxt;
    c->testlist->nxt = cls;
  }
  for (wp_down=c->testlist; wp_down; wp_down=wp_down->nxt)
    if (wp_down->nxt == cls)
      break;
  while (cls->nxt != lc) {
    wp_up = cls->nxt;
    cls->nxt = wp_up->nxt;
    wp_up->nxt = cls;
    wp_down->nxt = wp_up;
    wp_down = wp_up;
  }
  
} /* positive_last() */



/*-------------------------------------------------------------------------*/
void rewrite_variables(void)
/*
Called from add_to_theory.

Variables in the latest clause are subscripted with integer
identifiers in order to avoid clashes of variable name during
parsing. Now the subscripts should be removed, except that variables
of the same name and type in the latest clause should have local
identifying subscripts, and the types of the variables should be
added. So a variable v__globalID of type t becomes v:t unless there
are two or more variables v in the latest clause, in which case it
becomes v_localID:t
*/
{
  wff_stack wptr;
  int x;
  int subscript;      // Subscript to be applied here
  boolean sb_next;
  char extension[LEXLENGTH];

  latest_clause.varlist = sort_wfflist(latest_clause.varlist);
  // Strip off the old subscripts
  for (wptr = latest_clause.varlist; wptr; wptr = wptr->nxt) {
    x = strlen(wptr->w->sym->fname)-1;
    while(isdigit(wptr->w->sym->fname[x])) wptr->w->sym->fname[x--] = 0;
    while(wptr->w->sym->fname[x] == '_') wptr->w->sym->fname[x--] = 0;
  }
  // Add any new subscripts that may be required
  // Add the type extensions as well
  subscript = 0;
  for (wptr = latest_clause.varlist; wptr; wptr = wptr->nxt) {
    if (wptr->nxt && !strcmp(wptr->w->sym->fname,wptr->nxt->w->sym->fname)) {
      if (subscript == 0)
	subscript = 1;
      sb_next = true;
    }
    else
      sb_next = false;
    if (subscript)
      sprintf(extension,"_%d:%s",subscript,wptr->w->sym->val_type->sname);
    else
      sprintf(extension,":%s",wptr->w->sym->val_type->sname);
    strcat(wptr->w->sym->fname,extension);
    if (sb_next)
      subscript++;
    else
      subscript = 0;
  }
}


wff_stack sort_wfflist(wff_stack ws)
// Sort a singly-linked list of wffs. Return a pointer to the new head.
{
  wff_stack wptr;
  int x =0;

  for (wptr=ws; wptr; wptr=wptr->nxt) x++;
  return sort_wfflistn(ws,x);
}



wff_stack sort_wfflistn(wff_stack ws, int n)
// Sort as above, given length of list is known to be n
{
  wff_stack new;
  wff_stack wptr;
  wff_stack topstart;
  wff_stack bottomend;
  int x =0;

  if (n > 1) {
    if (n < 7) {
      // Very short list: stupid quadratic time sort is faster
      for (wptr=ws; wptr && wptr->nxt; wptr=wptr->nxt)
	while (wffcmp(ws,wptr->nxt) > 0) {
	  new = wptr->nxt;
	  wptr->nxt = new->nxt;
	  new->nxt = ws;
	  ws = new;
	}
      // Now have first things first
      ws->nxt = sort_wfflistn(ws->nxt,n-1);
    }
    else {
      //Genuinely needs computation: merge sort in O(n.log(n))
      for (bottomend=ws; x+1<n/2; bottomend=bottomend->nxt) x++;
      topstart = bottomend->nxt;
      bottomend->nxt = 0;
      // left sublist of x items starts with ws
      // right sublist of n-x items starts with top
      ws = sort_wfflistn(ws,x);
      topstart = sort_wfflistn(topstart,n-x);
      if (wffcmp(ws,topstart) > 1) {
	wptr = topstart;
	topstart = ws;
	ws = wptr;
      }
      // wptr is the last item in the sorted (low) fragment
      // ws is the start of the sorted (low) fragment
      // topstrt is the strt of the sublist not reachable from ws
      while (topstart) {
	if (wptr->nxt) {
	  if (wffcmp(wptr->nxt,topstart) > 1) {
	    new = wptr->nxt;
	    wptr->nxt = topstart;
	    topstart = new;
	  }
	  wptr = wptr->nxt;
	}
	else {
	  wptr->nxt = topstart;
	  topstart = 0;
	}
      }
    }
  }
  return ws;
}


int wffcmp(wff_stack w1, wff_stack w2)
{
  if (!(w1 && w2)) return 0;
  if (!(w1->w && w2->w)) return 0;    // Null result if not real
  return strcmp(w1->w->sym->fname,w2->w->sym->fname);
}



/*-------------------------------------------------------------------------*/
void add_to_theory(void)
     
     /*Called by model() and get_cl().
      *Move current clause from the temporary stack newcl[] to the 
      *permanent one form[], and add it to current_theory.*/

{
  wff_stack wptr;
  clause target;      /* The new clause as part of the theory */
  CLAUSE all_latest_clauses[CL_MAX];   /* Clausified version */
  int clause_id;
  int alc_size = 1;

  /* take id subscripts off variables and add type names */
  rewrite_variables();

  /* initialise, in case these have junk values */
  for (clause_id = 0; clause_id < CL_MAX; clause_id++) {
    all_latest_clauses[clause_id].prem = 0;
    all_latest_clauses[clause_id].conc = 0;
    all_latest_clauses[clause_id].varlist = 0;
    all_latest_clauses[clause_id].testlist = 0;
  }
  alc_size = clausify(&latest_clause,all_latest_clauses);

  /* zero latest clause, ready for the next */
  free_clause_parts(&latest_clause);

  for (clause_id = 0; clause_id < alc_size; clause_id++)
    all_latest_clauses[clause_id].varlist =
      sort_wfflist(all_latest_clauses[clause_id].varlist);

  /* New stuff is now in all_latest_clauses.
     Transfer each one to the persistent data structures */

  for (clause_id = 0; clause_id < alc_size; clause_id++) {
    /* copy clause to form[] */
    for (wptr = all_latest_clauses[clause_id].prem; wptr; wptr = wptr->nxt)
      wptr->w = copy_to_form(wptr->w);
    for (wptr = all_latest_clauses[clause_id].conc; wptr; wptr = wptr->nxt)
      wptr->w = copy_to_form(wptr->w);
    for (wptr = all_latest_clauses[clause_id].varlist; wptr; wptr = wptr->nxt)
      wptr->w = copy_to_form(wptr->w);
    for (wptr = all_latest_clauses[clause_id].testlist; wptr; wptr = wptr->nxt)
      wptr->w = copy_to_form(wptr->w);
  
    /* check sanity */
    if (size[current_theory] >= CL_MAX)
      skipout("Too many clauses (whew!)",1);

    /* get the next vacant spot in current theory */
    target = theory[current_theory] + size[current_theory];
    target->pid = all_latest_clauses[clause_id].pid;
    target->cid = next_cid++;
    target->prem = all_latest_clauses[clause_id].prem;
    target->conc = all_latest_clauses[clause_id].conc;
    target->varlist =  all_latest_clauses[clause_id].varlist;
    target->testlist = all_latest_clauses[clause_id].testlist;
    target->weight = all_latest_clauses[clause_id].weight;
    target->varcount = all_latest_clauses[clause_id].varcount;
    target->tuples = all_latest_clauses[clause_id].tuples;
    target->softweight = all_latest_clauses[clause_id].softweight;
    target->preprocessed = false;
    target->tautology = all_latest_clauses[clause_id].tautology;

    /* add it to the list of clauses and position it appropriately */
    target->lastclause = 0;
    target->nextclause = TH[current_theory];
    if (target->nextclause) target->nextclause->lastclause = target;
    TH[current_theory] = target;
    size[current_theory]++;
    normalize();
  }
}


/*-------------------------------------------------------------------------*/
wff copy_to_form(wff w)

     /*Called by add_to_theory().
      *Recursively copies the wff from the given structure to the form[],
      *identifying subformulae that already occur as it goes.*/

{
  int i,j;
  boolean subs_match;

  if (!w) return 0;

  /* for each argument of the main functor */
  for (i=0; i<w->sym->adicity; i++)
    w->sb[i] = copy_to_form(w->sb[i]);

  /* check to see if its already in the stack form[] */
  for (i=0; form[i].sym; i++) {
    if (form[i].sym->fname == w->sym->fname
	&& form[i].sym->val_type == w->sym->val_type) {     
      if (w->sym->category == Var || w->sym->category == Const) 
	return &(form[i]); /*found it!!*/

      /*compare subformulae*/
      subs_match = true;
      for (j=0; j<w->sym->adicity; j++)
	if (form[i].sb[j] != w->sb[j])
	  subs_match = false;
      if (subs_match) return &(form[i]);
    }
  }

  /* doesn't already exist so make a new one */
  if (i >= WF_MAX) skipout("Too many formulas parsed",1);
  form[i].sym = w->sym;
  for (j=0; j<ARGMAX; j++) {
    form[i].sb[j] = (j<w->sym->adicity? w->sb[j]: 0);
    form[i].sbval[j] = (w->sb[j]? &(w->sb[j]->val): &zero);
  }
  form[i].matrix = w->sym->table;
  if (w->sym->category == Const) form[i].val = w->val;
  form[i].wcc = -1;
  return &(form[i]); /* this is for the recursive call on subformulae */
}

/*================================ END ===================================*/
