#include "FINDER.h"


/*
*  Do_smallrefs is called after the search space has been prepared, to 
*  record refutations known to be of cardinality setting.pre_test or less.
*  Injective and surjective functions are also treated at this stage. 
*
*  The first bit of this function sets the matrix values of the nulladic 
*  functions which are fixed and which therefore don't count in refutations.
*/

boolean do_smallrefs(set info[], TRS *T)
{
  int twt, i;
  clause c;
  functor f;
  boolean do_one_small_clause();

  if (setting.pm.verbosity[pspace] == full)
    dump_space(T);

  for (f=FN; f; f=f->nextfunct) if (!f->adicity)
    if (S_size(info[f->fcc]) == 1) {
      for (i=0; !S_is_member(i,info[f->fcc]); i++) ;
      *(f->table) = i;
    }

  do_injectives(T);
  if (!do_surjectives(T)) return false;

  for (twt=0; twt <= setting.pretest_limit && !T->job->done; twt++) {
    for (c=TH[current_theory]; c && !T->job->done; c=c->nextclause) 
      if (!do_one_small_clause(c,twt,info,T))
	return false;
    if ((latest_clause.prem || latest_clause.conc)
	&& !do_one_small_clause(&latest_clause,twt,info,T))
      return false;
  }
  if (setting.pm.verbosity[pspace] != null)
    dump_space(T);
  return true;
}



boolean do_one_small_clause(clause c, int twt, set info[], TRS *T)
{
  int lwt;
  wff plit;

  if (!c->preprocessed && !T->cut_flag) {
    lwt = localweight(info,c,&plit,T);
    if (lwt <= twt || (plit && lwt == twt+1)) {
      c->preprocessed = true;
#ifdef DEBUG
      if (1) {                     /* Set more interesting condition */
	printf("Preprocessing ");
	print_one_clause(stdout,c,false);
	printf("  lwt = %d",lwt);
	printf("\n");
      }
#endif
      if (!twt_test(info,c,lwt,plit,T))
	return false;
    }
  }
  return true;
}


/*
*  The test routine for a clause whose local weight is lwt.
*  Return value is false iff a 0-refutation has been found.
*/

boolean twt_test(set info[], clause c, int lwt, wff plit, TRS *T)
{
  wff_stack wf;
  wff_stack tf[PRE_MAX];    /* The test formulae, in testlist order */
  int maximum[PRE_MAX];     /* Max value assignable to each formula */
  int i;
  functor f;

  tf[lwt] = 0;
  maximum[lwt] = 0;
  if (lwt)
    tf[0] = firstopen(info,c->testlist,lwt);
  for (i=1; i < lwt; i++)
    tf[i] = firstopen(info,tf[i-1]->nxt,1);
  for (i=0; i < lwt; i++)
    maximum[i] = tf[i]->w->sym->val_type->siz;
  correct_order(tf);

/*
*  First put in the values of any fixed cells
*/
  for (wf=c->testlist; wf; wf=wf->nxt) {
    f = wf->w->sym;
    if (is_open(f)) {
      if (!f->adicity && S_size(info[f->fcc]) == 1) {
	for (i=0; !S_is_member(i,info[f->fcc]); i++) ;
	*(f->table) = i;
      }
    }
  }

/*
*  Next initialize all the variables to the first (highest) value.
*/
  for (wf=c->varlist; wf; wf=wf->nxt)
    wf->w->val = wf->w->sym->val_type->siz-1;
/*
*  Now repeat the test until no more variable assignments remain.
*/
  do {

#ifdef DEBUG
    if (0) {
      for (wf=c->varlist; wf; wf=wf->nxt)
	printf(" %d", wf->w->val);
      printf("\n");
    }
#endif
    if (!alive(c,
	       tf,
	       tf,
	       maximum,
	       c->testlist,
	       info,
	       plit,
	       T)
	)
      return false;
  }
  while (!T->cut_flag && !T->job->done && next_assignment(c));
  return true;
}



boolean alive(clause c, wff_stack *tf, wff_stack *ltf, int *maximum,
	      wff_stack wf, set info[], wff plit, TRS *T)
{
  int i;
  int subvalues[ARGMAX];
  wff ww;

  for (; wf != *ltf; wf=wf->nxt) {
    ww = wf->w;
    for (i=0; i<ww->sym->adicity; i++)
      subvalues[i] = *(ww->sbval[i]);
    if (ww->matrix)
      ww->val = *(ww->matrix + getoffset(ww->sym,subvalues));
    else
      ww->val = builtin(ww->sym,subvalues);
  }

  if (!*ltf) return good_all(c,tf,plit,info,T);

  if (wf->w == plit)
    return alive(c,
		 tf,
		 ltf+1,
		 maximum+1,
		 (*ltf)->nxt,
		 info,
		 plit,
		 T);

  for ((*ltf)->w->val = 0; (*ltf)->w->val <= *maximum; (*ltf)->w->val++)
    if (poss_val((*ltf)->w,info) &&
	!alive(c,
	       tf,
	       ltf+1,
	       maximum+1,
	       (*ltf)->nxt,
	       info,
	       plit,
	       T))
      return false;
  return true;
}




/*
*  If the clause fails, inform TRANSREF. Return false iff the result
* refutes the entire search space (this will cause exit from do_smallrefs).
*/

boolean good_all(clause c, wff_stack tf[], wff plit, set info[], TRS *T)
{
  boolean failure = true;
  wff_stack wf;
  wff W;
  int cells[PRE_MAX], vals[PRE_MAX];
  boolean valency[PRE_MAX];
  int i,j;
  int subvalues[ARGMAX];

  for (wf = c->prem; wf; wf = wf->nxt)
    if (!wf->w->val) failure = false;
  if (plit) {
    if (!plit->sym->adicity && S_size(info[plit->sym->fcc]) == 1 &&
	c->conc->w->sb[1]->val == *(plit->matrix))
      failure = false;
    else
      plit->val = c->conc->w->sb[1]->val;
  }
  else for (wf=c->conc; wf; wf=wf->nxt)
    if (wf->w->val) failure = false;

  if (failure) {
    for (i=0; tf[i]; i++) {
      W = tf[i]->w;
      for (j=0; j<W->sym->adicity; j++)
	subvalues[j] = *(W->sbval[j]);
      cells[i] = vec_pos(W->sym, subvalues);
      vals[i] = W->val;
      valency[i] = (W == plit);
    }
    cells[i] = -1;
    new_small_ref(cells,vals,valency,c->softweight,T);
    if (T->novect)
      return false;
  }
  return true;
}




/*
*  Poss_val returns true iff the value assigned to w is possible
*  according to the vector of possible values. Note that this 
*  calls function vec_pos.
*
*  If one of the input values is the dummy, then poss_val is true 
*  iff the value of w is the dummy of the appropriate type.
*/

boolean poss_val(wff w, set info[])
{
  int i;
  int args[ARGMAX];

  for (i=0; i<w->sym->adicity; i++)
    if (*(w->sbval[i]) == w->sym->arg_type[i]->siz)
      return (w->val==w->sym->val_type->siz);
  for (i=0; i<w->sym->adicity; i++)
    args[i] = *(w->sbval[i]);
  return S_is_member(w->val,info[vec_pos(w->sym,args)]);
}



/*
*  For testing purposes, tfb is assumed to be in the posterity
*  of the other open cell tfa. They just might have got assigned
*  the wrong way round, so this function tests for correctness.
*/

void correct_order(wff_stack tf[])
{
  wff_stack wf;
  int i;

  if (!tf[0] || !tf[1]) return;
  for (i = 0; tf[i+1]; i++) {
    for (wf = tf[i]->nxt; wf; wf = wf->nxt)
      if (wf == tf[i+1]) break;
    if (!wf) skipout("Bad pointer sequence in a clause",1);
  }
}



/*
 * Give values where the function symbol is =, <, > or DIF
 */

int builtin(functor f, int sbvals[])
{
  if (f->fname[0] == '=')
    return (sbvals[0] == sbvals[1]? 1: 0);
  if (!strcmp(f->fname,"!="))
    return (sbvals[0] == sbvals[1]? 0: 1);
  if (f->fname[0] == '<')
    return (sbvals[0] < sbvals[1]? 1: 0);
  if (f->fname[0] == '>')
    return (sbvals[0] > sbvals[1]? 1: 0);
  if (!strcmp(f->fname,"DIF"))
    return (sbvals[0] > sbvals[1]? sbvals[0]-sbvals[1]: sbvals[1]-sbvals[0]);
  if (!strncmp(f->fname,"PRED",4))
      return (sbvals[0] && sbvals[0] < f->arg_type[0]->siz? sbvals[0]-1:
	      f->val_type->siz);
  if (!strncmp(f->fname,"SUCC",4))
      return (sbvals[0] < f->arg_type[0]->siz? sbvals[0]+1:
	      f->val_type->siz);
  {
    char report_line[100];
    sprintf(report_line,"Wrong builtin call: %s\nPlease file a bug report",f->fname);
    skipout(report_line,SKIP);
  }
  //  skipout("Wrong builtin call",SKIP);
  return 0;                            // Junk line to avoid warning
}



/*
*  Injectiveness is special to TRANSREF. It has to be informed of
*  the set of cells whose values cannot be identical, of the upper
*  bound on the set of values which cannot be assigned to more 
*  than one of the cells, and of course of the TRS structure to 
*  which all of this pertains. The upper bound on values for this 
*  purpose is to allow partial functions to be injective.
*
* If local search is selected, there are no propagators for global
* constraints, so in that case injectives have to be decomposed into
* sets of binary not-equals.
*
*  THIS MECHANISM WILL BE REPLACED BY A PROPER ALLDIFF ONE WHEN I
*  GET AROUND TO HACKING IT.
*/

void do_injectives(TRS *T)
{
  boolean *dinfo = (boolean*) malloc(T->job->vlength * sizeof(boolean));
  int i, j;
  functor f;
  int argument[ARGMAX];

  for (f = FN; f && !T->job->done; f = f->nextfunct) {
    if (f->injective && f->adicity > 0) {
      for (i = 0; i < tr_parameters.vlength; i++)
	dinfo[i] = false;
      set_dinfo(f,f->adicity-1,argument,dinfo);
      injective(dinfo, f->val_type->siz, T);
      f->leftin = f->rightin = false;
    }
    if (f->rightin && f->adicity==2) {
      for (i = 0; i < tr_parameters.vlength; i++)
	dinfo[i] = false;
      for (i = 0; i < f->arg_type[0]->siz; i++) {
	argument[0] = i;
	for (j = 0; j < f->arg_type[1]->siz; j++) {
	  argument[1] = j;
	  dinfo[vec_pos(f,argument)] = true;
	}
	injective(dinfo, f->val_type->siz, T);
	for (j = 0; j < f->arg_type[1]->siz; j++) {
	  argument[1] = j;
	  dinfo[vec_pos(f,argument)] = false;
	}
      }
    }
    if (f->leftin && f->adicity==2) {
      for (i = 0; i < tr_parameters.vlength; i++)
	dinfo[i] = false;
      for (j = 0; j < f->arg_type[1]->siz; j++) {
	argument[1] = j;
	for (i = 0; i < f->arg_type[0]->siz; i++) {
	  argument[0] = i;
	  dinfo[vec_pos(f,argument)] = true;
	}
	injective(dinfo, f->val_type->siz, T);
	for (i = 0; i < f->arg_type[0]->siz; i++) {
	  argument[0] = i;
	  dinfo[vec_pos(f,argument)] = false;
	}
      }
    }
  }
  if (dinfo) free(dinfo);
}



void set_dinfo(functor f, int x, int argument[], boolean dinfo[])
{
  for (argument[x]=0; argument[x]<f->arg_type[x]->siz; argument[x]++) {
    if (x)
      set_dinfo(f,x-1,argument,dinfo);
    else
      dinfo[vec_pos(f,argument)] = true;
  }
}




/*
*  Surjective functions are also treated specially by TRANSREF.
*  Left and right surjectiveness are treated here, as well as 
*  total surjectiveness.
*/

boolean do_surjectives(TRS *T)
{
  boolean *dinfo = (boolean*) malloc(T->job->vlength * sizeof(boolean));
  set vset = S_new(1,0);
  int i, j, m;
  int valcount, argcount;
  functor f;
  int argument[ARGMAX];

  for (f=FN; f && !T->job->done; f=f->nextfunct)
    if (f->adicity > 0) {
      if (f->rightsur) {
	for (i=0; i < tr_parameters.vlength; i++) dinfo[i] = false;
	for (i=0; i < f->arg_type[0]->siz; i++) {
	  argument[0] = i;
	  for (j=0; j < f->arg_type[1]->siz; j++) {
	    argument[1] = j;
	    dinfo[vec_pos(f,argument)] = true;
	  }
	  S_empty(vset);
	  valcount = 0;
	  for (m=0; m < f->val_type->siz; m++) {
	    S_into(m,vset);
	    valcount++;
	  }
	  if (valcount > f->arg_type[1]->siz) {
	    if (dinfo) free(dinfo);
	    return false;
	  }
	  surjective(dinfo,vset,T);
	  for (j=0; j < f->arg_type[1]->siz; j++) {
	    argument[1] = j;
	    dinfo[vec_pos(f,argument)] = false;
	  }
	}
	f->surjective = false;
      }
      if (f->leftsur) {
	for (i=0; i < tr_parameters.vlength; i++) dinfo[i] = false;
	for (j=0; j < f->arg_type[1]->siz; j++) {
	  argument[1] = j;
	  for (i=0; i < f->arg_type[0]->siz; i++) {
	    argument[0] = i;
	    dinfo[vec_pos(f,argument)] = true;
	  }
	  S_empty(vset);
	  valcount = 0;
	  for (m=0; m < f->val_type->siz; m++) {
	    S_into(m,vset);
	    valcount++;
	  }
	  if (valcount > f->arg_type[0]->siz) {
	    if (dinfo) free(dinfo);
	    return false;
	  }
	  surjective(dinfo,vset,T);
	  for (i=0; i < f->arg_type[0]->siz; i++) {
	    argument[0] = i;
	    dinfo[vec_pos(f,argument)] = false;
	  }
	  f->surjective = false;
	}
      }
      if (f->surjective) {
	for (i=0; i < tr_parameters.vlength; i++) dinfo[i] = false;
	set_dinfo(f,f->adicity-1,argument,dinfo);
	argcount = 1;
	for (i=0; i<f->adicity; i++)
	  argcount *= f->arg_type[i]->siz;
	S_empty(vset);
	valcount = 0;
	for (i=0; i < f->val_type->siz; i++) {
	  S_into(i,vset);
	  valcount++;
	}
	if (valcount > argcount) {
	  free(dinfo);
	  return false;
	}
	surjective(dinfo,vset,T);
      }
    }
  if (dinfo) free(dinfo);
  S_free(vset);
  if (T->job->done) return false;
  return true;
}



/*
*  Local weight is weight minus any nulladic functions with only
*  one possible value. The parameter *plit is set to point to the
*  left subformula of the sole positive literal (if any) which must
*  have identity as its predicate and an open left subformula.
*/ 

int localweight(set info[], clause c, wff *plit, trs T)
{
  wff_stack wf;
  int counter = 0;
  functor f;
  boolean poscon;    /* Positive literal condition */

  poscon = (
	    c->conc &&
	    !c->conc->nxt &&
	    *(c->conc->w->sym->fname) == '=' &&
	    is_open(c->conc->w->sb[0]->sym) &&
	    !occurs_inside(c->conc->w->sb[0], c->conc->w->sb[1]) &&
	    !occurs_in_list(c->conc->w->sb[0], c->prem));
  if (poscon) *plit = c->conc->w->sb[0];
  else *plit = 0;

  for (wf=c->testlist; wf; wf = wf->nxt) {
    f = wf->w->sym;
    if (is_open(f)) {
      if (f->adicity || !(S_size(info[f->fcc]) == 1))
	counter++;
    }
  }
  return counter;
}



/*
*  The first open spot in a list of wffs is the first one which
*  is part of a function that is_open, other than a nulladic one
*  with only one possible value.
*/

wff_stack firstopen(set info[], wff_stack wfl, int opens)
{
  functor f;

  if (!opens) printf("Firstopen called without opens!\n");
  f = wfl->w->sym;
  while (!is_open(f) ||
	 (opens && !f->adicity && S_size(info[f->fcc]) == 1)) {
    wfl = wfl->nxt;
    f = wfl->w->sym;
  }
  return wfl;
}



/*
*  A function is_open if it is really part of the search space 
*  and not one of the constant functions such as identity or a
*  function piped in from another FINDER.
*
*  6/5/08 : let piped functions count as open
*/

boolean is_open(functor f)
{
  if (!f) return false;
 /*
  * if (f->category != Func && f->category != Const) return false;
  */
  if (f->category != Func)          return false;
  if (strchr("+-=<>",f->fname[0]))  return false;
  if (!strcmp(f->fname,"DIF"))      return false;
  if (!strcmp(f->fname,"!="))       return false;
  if (!strncmp(f->fname,"EST_",4))  return false;
  if (!strncmp(f->fname,"PRED_",5)) return false;
  if (!strncmp(f->fname,"SUCC_",5)) return false;
  if (!strcmp(f->fname,"NOT"))      return false;
  if (!strcmp(f->fname,"AND"))      return false;
  if (!strcmp(f->fname,"OR"))       return false;
  if (!strcmp(f->fname,"IMP"))      return false;
  if (!strncmp(f->fname,"LEN_",4))  return false;
  /*
   *  if (f->piped)                     return false;
   */
  return true;
}


/*
*  Formula fmla occurs as a subformula of something in wfl.
*/

boolean occurs_in_list(wff fmla, wff_stack wfl)
{
  if (!wfl) return false;
  do
    if (occurs_inside(fmla,wfl->w)) return true;
  while ((wfl = wfl->nxt));
  return false;
}


/*
*  Formula little occurs as a subformula of formula big.
*/

boolean occurs_inside(wff little, wff big)
{
  int i;

  if (!big) return false;
  if (little == big) return true;
  for (i=0; i<big->sym->adicity; i++)
    if (occurs_inside(little, big->sb[i])) return true;
  return false;
}
