#include "Fvntr.h"

#define ABS_MAX_TIME (7*24*60*60)

static time_t start_sec = 0;                        // Seconds at the start
static suseconds_t start_adjustment;                // Clock ticks to subtract
static int theLimit = ABS_MAX_TIME*LOCAL_CLK_TCK;   // Timeout
static int checkpoint = 0;                          // When to check the clock

void reset_timer(void)
// Set current time as the baseline
{
  start_sec = 0;
  timestamp();
}



void set_maxtime(int s)
// Set the time limit (in clock ticks) to s seconds. Do not exceed the
// absolute time limit - this is also the default if no limit (zero) is
// specified.
{
  if (s && s <= ABS_MAX_TIME)
    theLimit = s * LOCAL_CLK_TCK;
  else
    theLimit = ABS_MAX_TIME * LOCAL_CLK_TCK;
}


int time_limit(void)
// Return the current time limit in seconds
{
  return theLimit / LOCAL_CLK_TCK;
}


int timeout(trs T)
// If the checkpoint (a number of value insertions) is not reached,
// return 0 (false). Otherwise get a timestamp. If the time limit is
// exceeded, return 1 (true) but take no other action. Otherwise set a
// new checkpoint.
{
  int a = T->job->reps[INVALUES].i_report;
  int time_now;
  int prediction;

  if (!a || a >= checkpoint) {
    time_now = timestamp();
    if (time_now >= theLimit)
      return 1;
    if (a && time_now) {
      prediction = (theLimit * a) / time_now;
      if (prediction > 3*(checkpoint+1))
	prediction = 3*(checkpoint+1);          // to avoid OoM errors
      checkpoint += (prediction - checkpoint) /2;
    }
  }
  return 0;
}


int timestamp(void)
// Return the current elapsed time in clock ticks since the first time
// this function was called. It does not matter what clock ticks are.
{
  struct timeval tp;
  struct timezone tzp;

  gettimeofday(&tp,&tzp);
  if (!start_sec) {
    theClockTick = LOCAL_CLK_TCK;
    start_sec = tp.tv_sec;
    start_adjustment = (tp.tv_usec * theClockTick) / 1000000;
    return 0;
  }
  else
    return ((tp.tv_sec - start_sec) * theClockTick) +
      ((tp.tv_usec * theClockTick) / 1000000) - start_adjustment;
}


int timestat(trin Job, int old_t, report_code rc)
// Report the time taken in seconds
{
  int t;

  t = timestamp();
  Job->reps[rc].f_report += ((t-old_t) * 1.0) / theClockTick;
  return t;
}
