#include "FINDER.h"

#ifdef SOFTIE_SCOTT
#include "SOFTIE_SCOTT.h"
#endif

void get_options(int argc, char *argv[],
		 boolean *debugging,
		 char *filename,
		 char *diag_output,
		 boolean *pipe_out
		 )
// These are options for both finder and diagnose
{
  int option;

  while ((option = getopt (argc, argv, "c:def:hmn:o:pst:v:")) != -1)
    switch(option) {
    case 'c': dynstring_strcatc(&command_line,optarg);
      dynstring_strcatc(&command_line," ");
      break;
    case 'd': *debugging = true;
      break;
    case 'e': setting.drilldown = true;
      break;
    case 'f': if (filename) {
	strcpy(filename,optarg);
	piping = true;
	pipe_open = true;
      }
      break;
    case 'h': print_synopsis();
      exit(0);
    case 'm': setting.single_literal = false;
      break;
    case 'n': strcpy(filename,optarg);
      piping = true;
      pipe_open = true;
      negative = true;
      break;
    case 'o': if (diag_output)
	strcpy(diag_output,optarg);
      break;
    case 'p': *pipe_out = true;
      break;
    case 's': setting.single_literal = true;
      break;
    case 't':
      break;           // Temp path name no longer does anything
    case 'v': if (!strcmp(optarg,"comp"))
	setting.function_spec_mode = comp;
      else if (!strcmp(optarg,"math"))
	setting.function_spec_mode = math;
      else
	inerror("Unrecognised function spec mode",optarg,false);
      break;
    }
}


/*
*	Go is the main work loop of FINDER. It calls transref to search
*	every search space, defined by setting the cardinalities of the
*	sorts, within the range stipulated. Go is called from the above 
*	functions, and from itself, and also from main().
*/


static int old_updates;

void go(int x, boolean persists)
{
  if (tr_parameters.done) return;
  if (negative && satisfied) return;
  if (!x) old_updates = updates[current_theory];
  if (!sort[x].sname) {
    set_up_cc();
    firstmodel = true; // perhaps need to change this?
    transref(&tr_parameters);
    if (!persists) clean_up_cc();
    // Multiple searches will call cc, which cleans up ahead of
    // itself. We may want the last one to clear memory, or we may
    // want the current state to persist so that test functions work
    // correctly.
    if (tr_parameters.done) return;
  }
  else if (sort[x].piped && !tr_parameters.done) go(x+1,persists);
  else for (sort[x].siz = sort[x].sizmin;
	    sort[x].siz <= sort[x].sizmax && !tr_parameters.done;
	    sort[x].siz++) {
      if (updates[current_theory]==old_updates || !setting.sortcut)
	go(x+1,persists);
    }
}


/*
 * At present the condition for the clause to be "long" is rather
 * crude. Later it can be made more refined.
 */

#define SHORT_CLAUSE_CUTOFF 4

boolean clause_is_long()
{
#ifdef SOFTIE_SCOTT
  return (latest_clause.weight > SHORT_CLAUSE_CUTOFF);
#else
  return (latest_clause.weight + latest_clause.varcount > 7);
#endif
}


/*
*	Action on signal from alarm.
*/

void timesup()
{
  if (not_searching_yet) {
    //    alarm(0);
    exit(1);
  }
  job_done(&tr_parameters, TIMEOUT);
}


/*
*	Action on signal from ^C or whatever SIGINT may be.
*/

void intrupt()
{
  if (not_searching_yet) {
    //    alarm(0);
    exit(1);
  }
  job_done(&tr_parameters, INTERRUPT);
}


void print_synopsis(void)
{
  printf("Synopsis:\n  finder [-cdefhmnopsv]\n\n");
  printf("    -c <string>: append <string> to input\n");
  print_with_default('d',"dump debugging information",
		     false);
  print_with_default('e',"print detailed error messages",
		     setting.drilldown);
  printf("    -f <path>: pipe partial models from <path> before stdin\n");
  printf("    -h: print this message\n");
  print_with_default('m',"formulae in multi-literal (clause) form",
		     !setting.single_literal);
  printf("    -n <path>: pipe models from <path> and test for constraint failure\n");
  printf("    -o: <command> if diagnosing, execute <command> only\n");
  print_with_default('p',"write output in suitable form for piping to FINDER",
		     false);
  print_with_default('s',"formulae in single-literal (FOF) form",
		     setting.single_literal);
  printf("    -v comp: vocabulary uses \"comp\" style function declarations\n");
  printf("    -v math: use \"math\" style function declarations\n");
  if (setting.function_spec_mode == comp)
    printf("             (default is comp mode)\n");
  else
    printf("             (default is math mode)\n\n");
}


void print_with_default(char c, char *s, boolean b)
{
  char toggle[4];

  if (b)
    strcpy(toggle,"on");
  else 
    strcpy(toggle,"off");
  printf("    -%c: %s (%s by default)\n",c,s,toggle);
}

