#define TRFILE
#include "Fvntr.h"

// Macro to replace the dirty free call
#define Free(mem) {free(mem); mem = 0;}

static int debugprinting = 0;
static int smallrefcounter = 0;
static int tabuchoices;

void turn_on_debugging(void);

void check_cell_for_sanity(cell c, trs T);
// void check_sanity(trs T);


// declared in Fproto.h, but including that causes problems
void print_best_model(int info[], trs T);

/*
*  This is the top level function for this file. The "Job" contains a 
*  succinct description of the problem, consisting of the length of the
*  vectors required, the addresses of the test function etc. and some 
*  miscellaneous settings. The search is performed on the whole space 
*  and then on each subspace created by "cutting" (see above) until 
*  there is no "cut" in the stack.
*/

void transref(trin Job)
{
  trs T = 0;
  int local_time[2];
  int sol_found;
  set *possibles = 0;
  set *impossibles = 0;

#ifdef DEBUG
  init_init_refring();
#endif

  T = tr_initial(Job);      /* Returns null if Job->done */
  if (!T) return;
  do {
    local_time[0] = timestamp();
    T->local_memory = 0;
    if (prepared(T) && !T->novect && !Job->done) {
      if (pre_test(T) && !Job->done) {
	local_time[1] = timestat(Job,local_time[0],PRE_TIME);
	if (Job->configmode) {
	  init_config(T,&possibles,&impossibles);
      	  if (Job->maxtime) {
	  //	    alarm(0);   /* Prevents timeout */
	    reset_timer();
	    set_maxtime(0);
	  }
	}
	EVENT(enterSearch,T->job);
	sol_found = found_solutions(T);
	EVENT(leaveSearch,T->job);
	if (Job->configmode) config_search(T,possibles,impossibles,sol_found);
	timestat(Job,local_time[1],SEARCH_TIME);
	if (Job->awnty_enabled) printf("awnty 1 0\n");
      }
    }
    clean_up(T);
  }
  while (T->cutptr && !T->job->done);

  if (T->best_penalty && T->best_penalty <= Job->max_penalty)
    print_best_model(T->best,T);
  if (Job->configmode)
    (*Job->Printdomains)(possibles,impossibles);
  if (Job->dump_rate) dump_trace(Job,true);

  if (Job->configmode) free_config(possibles,impossibles,T);
  free_trs(T);
}


void tr_init_globals(void)
{
  tr_last_vlen = 0;
  tr_last_vector = 0;
}


int found_solutions(trs T)
{
  int sol;
  int x;

  sol = T->job->reps[SOLUTIONS].i_report;
  T->firstchange = -1;
  T->forkpoint = -1;
  T->this_cell = -1;
  T->unfixt = T->job->vlength-1;
  T->ts = false;
  T->newref = 0;
  /*
   *  T->job->done = false;
   */
  if (!T->job->done)
    T->job->reps[STOP_CODE].i_report = 0;
  T->job->tracepoints = 0;
  for (x=0; x<T->job->vlength; x++) T->vector[x].fixed = false;
  set_total_penalty(T);
#ifdef DEBUG
  init_refring(T);
#endif
  if (T->job->search_is_local)
    localsearch(T);
  else search(T,0,0);
  T->job->reps[SEARCHES].i_report++;
  return T->job->reps[SOLUTIONS].i_report - sol;
}


void init_config(trs T, set **possibles, set **impossibles)
{
  *possibles = (set*) malloc(T->job->vlength * sizeof(set));
  *impossibles = (set*) malloc(T->job->vlength * sizeof(set));
  initial_domains(T,*possibles,*impossibles);
  set_domains(T,*possibles,*impossibles);
  T->configtestcell = -2;         /* Any impossible value */
  T->configstart = false;
}

void initial_domains(trs T, set possibles[], set impossibles[])
{
  int x;

  for (x=0; x<T->job->vlength; x++) {
    if (T->job->configurable[x]) {
      possibles[x] = S_new(T->vector[x].csiz,0);
      impossibles[x] = S_new(T->vector[x].csiz,0);
    }
    else {
      possibles[x] = 0;
      impossibles[x] = 0;
    }
  }
}


void set_domains(trs T, set possibles[], set impossibles[])
{
  int x,y;

  for (x=0; x<T->job->vlength; x++)
    if (T->job->configurable[x])
      for (y=0; y<T->vector[x].csiz; y++) {
	if (T->vector[x].cvals[y].blocks)
	  S_into(y,impossibles[x]);
	if (T->vector[x].cvals[y].known_possible)
	  S_into(y,possibles[x]);
	if (!S_is_disjoint(possibles[x],impossibles[x])) {
	  int i; char line[100];
	  for (i=0; i<T->vector[x].csiz; i++)
	    if (S_is_member(i,possibles[x]) && S_is_member(i,impossibles[x])) {
	      sprintf(line,
		      "%d marked as possible and impossible for cell %d",
		      i,x);
	      skipout(line,SKIP);
	    }
	}
      }
}



void config_search(trs T, set possibles[], set impossibles[], int sol)
{
  int x;
  boolean *tabu = (boolean*) malloc(T->job->vlength * sizeof(boolean));
  FILE *fil = T->job->outf;
  ifile ifil = T->job->outi;

  if (!sol) {                    /* No solution found */
    for (x=0; x<T->job->vlength; x++) 
      if (T->job->configurable[x]) {
	S_empty(possibles[x]);
	S_fill(impossibles[x],T->vector[x].csiz);
      }
  }
  else {
    set_domains(T,possibles,impossibles);
    for (x=0; x<T->job->vlength; x++) tabu[x] = false;
    tabuchoices = 0;
    T->configtestcell = -1;
    while (find_unknown(T,tabu)) {    /* Also sets allknown */
#ifdef DEBUG
      init_refring(T);
#endif
      if (T->job->done && T->job->reps[STOP_CODE].i_report == INTERRUPT)
	return;
      T->job->done = false;
      T->job->reps[STOP_CODE].i_report = 0;
      T->job->reps[STOP_CODE].s_report[0] = 0;
      //      alarm(T->job->maxtime);
      reset_timer();
      set_maxtime(T->job->maxtime);
      if (!found_solutions(T)) {
	if (T->job->reps[STOP_CODE].i_report == EXHAUSTION) {
	  S_into(T->configtestval,impossibles[T->configtestcell]);
	  T->vector[T->configtestcell].cvals[T->configtestval].blocks++;
#ifdef IN_FINDER
	  xprint(fil,ifil,"%%REMOVED%%\nclause { NOT(");
	  print_un_vec_pos(fil,ifil,T->configtestcell);
	  xprint(fil,ifil," = ");
	  print_val_of_cell(fil,ifil,T->configtestcell,T->configtestval);
	  xprint(fil,ifil,"). }\n");
	  fflush(stdout);
#endif
	}
	else {
#ifdef DEBUG
	  fprintf(stderr,"#");
#endif
	  tabu[T->configtestcell] = true;
	}
      }
      set_domains(T,possibles,impossibles);
    }
  }
}


boolean find_unknown(trs T, boolean tabu[])
{
  int x, y;
  boolean found_unknown = false;

  for (x=0; x<T->job->vlength; x++) {
    if (T->job->configurable[x]) {
      T->vector[x].allknown = true;
      for (y=0; y<T->vector[x].csiz; y++)
	if (!T->vector[x].cvals[y].blocks &&
	    !T->vector[x].cvals[y].known_possible) {
	  T->vector[x].allknown = false;
	  if (!found_unknown && !tabu[x]) {
	    T->configtestcell = x;
	    T->configtestval = y;
	    found_unknown = true;
	    T->configstart = true;
	  }
	  break;
	}
    }
    T->vector[x].times_picked = 0;
    T->vector[x].used = false;
  }
  if (!found_unknown) {
    for (x=0; x<T->job->vlength; x++)
      if (tabu[x]) {
	tabu[x] = false;
	tabuchoices++;
	if (!found_unknown && !T->vector[x].allknown) {
	  for (y=0; y<T->vector[x].csiz; y++)
	    if (!T->vector[x].cvals[y].blocks &&
		!T->vector[x].cvals[y].known_possible) {
	      T->configtestcell = x;
	      T->configtestval = y;
	      found_unknown = true;
	      T->configstart = true;
	      T->job->maxtime *= 2;              /* Increase the time */
	      T->job->configheuristic = false;   /* and opt for efficiency */
	    }
	  break;
	}
      }
  }
#ifdef DEBUG
  {
    int theArgs[10];
    un_vec_pos(T->configtestcell,theArgs);
    fprintf(stderr,"Testing %d (f[%d]) %d\n",
	    T->configtestcell, theArgs[0], T->configtestval);
  }
#endif
  return found_unknown;
}


void free_config(set possibles[], set impossibles[], trs T)
{
  int x;

  for (x=0; x<T->job->vlength; x++) {
    if (possibles) S_free(possibles[x]);
    if (impossibles) S_free(impossibles[x]);
    possibles[x] = impossibles[x] = 0;
  }
  if (possibles) free(possibles);
  if (impossibles) free(impossibles);
  possibles = impossibles = 0;
}



void free_trs(trs T)
{
  int x;

  if (T->vector) {
    for (x=0; x<T->job->vlength; x++)
      if (T->vector[x].cvals) Free(T->vector[x].cvals);
    Free(T->vector);
  }
  if (T->scomvec) {
    for (x=0; x<T->job->vlength; x++)
      S_free(T->scomvec[x]);
    Free(T->scomvec);
  }
  if (T->icomvec) Free(T->icomvec);
  if (T->inorder) Free(T->inorder);
  if (T->coinorder) Free(T->coinorder);
  if (T->priority) Free(T->priority);
  if (T->best) Free(T->best);
  for (x=0; x<SURJ_MAX; x++)
    if (T->surjection[x].last_seen) Free(T->surjection[x].last_seen);
  if (T->awnty_past) Free(T->awnty_past);
  if (T->awnty_future) Free(T->awnty_future);
  if (T->awnty_sofar) Free(T->awnty_sofar);
  Free(T);
}



/*
*  This is not part of the search. It just initialises a job description.
*/

void init_trin(trin tr)
{
  int i, j;

  tr->Test = 0;
  tr->Onerefs = 0;
  tr->Smallrefs = 0;
  tr->Sizes = 0;
  tr->batches = REF_DEFAULT;
  tr->extra_batches = EXTRA_DEFAULT;
  tr->maxref = MAX_REF_SIZE;
  tr->max_cellvals = CELLVAL_MAX;
  tr->branches = 0;
  tr->Printdomains = 0;
  tr->forbidden = 0;
  tr->configurable = 0;
  tr->dump_rate = 0;
  tr->verb = 0;
  tr->Nosecs = 0;
  tr->topsoff = true;
  tr->report_breaks = false;
  tr->print_headers = false;
  tr->random = false;
  tr->dumpdepth = false;
  tr->awnty_enabled = false;
  tr->cgv_enabled = false;
  tr->debug_enabled = false;
  tr->profile = false;
  tr->configmode = false;
  tr->configheuristic = false;
  tr->search_is_local = false;
  tr->maxbak = 0;
  tr->maxinvals = 0;
  tr->max_penalty = MAX_PENALTY;
  tr->hardweight = MAX_PENALTY;
  tr->done = false;
  tr->outf = 0;
  tr->outi = 0;

  for (i = 0; i < MAX_REPORT; i++) {
    tr->reps[i].report_type = 0;
    for (j=0; j<80; j++)
      tr->reps[i].s_report[j] = 0;
    tr->reps[i].i_report = 0;
    tr->reps[i].f_report = 0;
    tr->reps[i].do_report = 0;
  }

  for (i = 0; i < RRA; i++)
    for (j = 0; j < RRB; j++)
      tr->ref_record[i][j] = 0;    
}



/*
*  Clean up after a loop through the search by deleting the database and
*  freeing the memory it used. This is mostly just garbage collection.
*
*  After a search, there may be a dangling active cut, or the need for a
*  cut may have been signalled but not acted upon. This function puts all
*  such things back into their proper state, and activates the cut for the
*  next loop if there is one.
*/

void clean_up(trs T)
{
  int i;

  for (i = 0; T->stak[i] && i < T->job->batches; i++) {
    free(T->stak[i]);
    T->stak[i] = 0;
  }
  for (i = 0; T->il_stak[i] && i < T->stakmax; i++) {
    free(T->il_stak[i]);
    T->il_stak[i] = 0;
  }
  for (i = 0; T->rl_stak[i] && i < T->job->batches+T->stakmax; i++) {
    free(T->rl_stak[i]);
    T->rl_stak[i] = 0;
  }
  if (!T->cutptr) return;
  if (T->cutptr->active) delete_cut(T);
  T->cut_flag = NOCUT;
  if (T->cutptr) T->cutptr->active = true;
  if (T->local_memory > T->job->reps[LOC_MEMORY].i_report)
    T->job->reps[LOC_MEMORY].i_report = T->local_memory;
}



/*
*  It is possible to report many statistics on the execution. The routine
*  for printing these out is somewhere else in the code, but the definitions
*  of the printable quantities are here. Each report has a type (integer, 
*  floating point or function) and an identifying string.
*/

void report_initial(trs T)
{
  int i;

  switch(T->job->verb) {
  case 0:
    T->job->reps[0].report_type = 0;
    return;

  case 1:
    for (i = 0; i < SOLUTIONS; i++)
      T->job->reps[i].report_type = NO_REPORT;
    strcpy(T->job->reps[SOLUTIONS].s_report,  "Solutions found:       ");
    T->job->reps[SOLUTIONS].report_type = INT_REPORT;
    T->job->reps[SOLUTIONS+1].report_type = 0;
    return;

  case 2:
    strcpy(T->job->reps[SOLUTIONS].s_report,  "Solutions found:       ");
    T->job->reps[SOLUTIONS].report_type = INT_REPORT;

    strcpy(T->job->reps[SEARCHES].s_report,   "Subspaces searched:    ");
    T->job->reps[SEARCHES].report_type = INT_REPORT;

    strcpy(T->job->reps[SETPOSSES].s_report,  "Subspaces prepared:    ");
    T->job->reps[SETPOSSES].report_type = INT_REPORT;

    strcpy(T->job->reps[DOSMALLS].s_report,   "Subspaces preprocessed:");
    T->job->reps[DOSMALLS].report_type = INT_REPORT;

    strcpy(T->job->reps[BAD_TESTS].s_report,  "Bad candidates tested: ");
    T->job->reps[BAD_TESTS].report_type = INT_REPORT;

    strcpy(T->job->reps[SURJ_BACKTRACKS].s_report,
	   "Surjection backtracks: ");
    T->job->reps[SURJ_BACKTRACKS].report_type = INT_REPORT;

    strcpy(T->job->reps[NULL_BACKTRACKS].s_report,
	   "Other backtracks:      ");
    T->job->reps[NULL_BACKTRACKS].report_type = INT_REPORT;

    strcpy(T->job->reps[F_SUBSUMED].s_report, "Forward subsumptions:  ");
    T->job->reps[F_SUBSUMED].report_type = INT_REPORT;

    strcpy(T->job->reps[B_SUBSUMED].s_report, "Back subsumptions:     ");
    T->job->reps[B_SUBSUMED].report_type = INT_REPORT;

    strcpy(T->job->reps[B1_SUBSUMED].s_report,"   Back subs by units: ");
    T->job->reps[B1_SUBSUMED].report_type = INT_REPORT;

    strcpy(T->job->reps[SOFT_DOUBLE].s_report,"Soft doublings:        ");
    T->job->reps[SOFT_DOUBLE].report_type = INT_REPORT;

    strcpy(T->job->reps[LOC_MEMORY].s_report, "Maximum database size: ");
    T->job->reps[LOC_MEMORY].report_type = INT_REPORT;

    strcpy(T->job->reps[MEMORY].s_report,     "Total database entries:");
    T->job->reps[MEMORY].report_type = INT_REPORT;

    strcpy(T->job->reps[VECMAX].s_report,     "Decision variables:    ");
    T->job->reps[VECMAX].report_type = INT_REPORT;

    strcpy(T->job->reps[PRE_TIME].s_report,   "Preprocess time:       ");
    T->job->reps[PRE_TIME].report_type = FLOAT_REPORT;

    strcpy(T->job->reps[SEARCH_TIME].s_report,"Search time:           ");
    T->job->reps[SEARCH_TIME].report_type = FLOAT_REPORT;

    strcpy(T->job->reps[PREREFS].s_report,    "Initial constraints:   ");
    T->job->reps[PREREFS].report_type = INT_REPORT;

    strcpy(T->job->reps[INVALUES].s_report,   "Values inserted:       ");
    T->job->reps[INVALUES].report_type = INT_REPORT;

    strcpy(T->job->reps[FLIPS].s_report,      "Flips:                 ");
    T->job->reps[FLIPS].report_type = INT_REPORT;

    strcpy(T->job->reps[REFSIZE].s_report,    "Nogoods learned:       ");
    T->job->reps[REFSIZE].report_type = INT_REPORT | DO_REPORT;
    T->job->reps[REFSIZE].do_report = ref_size_report;

    strcpy(T->job->reps[TOPLESS].s_report,    "Nogoods deleted:       ");
    T->job->reps[TOPLESS].report_type = INT_REPORT;

    strcpy(T->job->reps[BRANCHES].s_report, "Branches");
    T->job->reps[BRANCHES].report_type = DO_REPORT;
    T->job->reps[BRANCHES].do_report = report_branches;

    strcpy(T->job->reps[STOP_CODE].s_report, "");
    T->job->reps[STOP_CODE].report_type = DO_REPORT;
    T->job->reps[STOP_CODE].do_report = stopped_by;

    T->job->reps[MAX_REPORT].report_type = 0;
  }
}



/*
 * This sets the strings for reporting purposes in case we need to
 * print the reason why the search was terminated. The line says
 * "Search stopped by <string>").
 */

void termination_initial(trs T)
{
  strcpy(T->job->tcode[EXHAUSTION],"exhaustion");
  strcpy(T->job->tcode[TIMEOUT],"time limit");
  strcpy(T->job->tcode[ENOUGH_SOLUTIONS],"limit on number of solutions");
  strcpy(T->job->tcode[ASSIGNMENT_OK],"consistent assignment");
  strcpy(T->job->tcode[INSERTIONS],"limit on insertions");
  strcpy(T->job->tcode[BACKTRACKS],"limit on backtracks");
  strcpy(T->job->tcode[INTERRUPT],"interrupt");
}



/*
*   This is the initialisation function for a TRS structure, which contains
*   all the variables used by functions in this file to communicate with each
*   other. It is really a subroutine of transref above. The job description 
*   becomes a field of the TRS so that every function can refer to it.
*/

trs tr_initial(trin Job)
{
  trs T;
  int i;

  if (Job->done) return 0;
  if (!(Job->vlength) && !(Job->piping)) {
    EVENT(abortSearch,Job);
    fprintf(WARNING_STREAM,"\nNo problem specified, so no results\n");
    return 0;
  }
  if (Job->vlength > Job->reps[VECMAX].i_report)
    Job->reps[VECMAX].i_report = Job->vlength;
  if (Job->batches < 1) {
    fprintf(WARNING_STREAM,"Setting stack size to 1\n");
    Job->batches = 1;
  }
  if (Job->extra_batches < 1) {
    fprintf(WARNING_STREAM,"Setting marginal stack size to 1\n");
    Job->extra_batches = 1;
  }
  if (Job->batches >= REF_MAX) {
    //    fprintf(WARNING_STREAM,"Setting stack size to %d\n", REF_MAX-1);
    Job->batches = REF_MAX-1;
  }
  if (Job->maxref < 1) {
    fprintf(WARNING_STREAM,"Setting ref-length to 1\n");
    Job->maxref = 1;
  }
  if (Job->maxref > MAX_REF_SIZE)
    fprintf(WARNING_STREAM,
	    "Warning: ref-length %d exceeds notional maximum %d\n",
	    Job->maxref, MAX_REF_SIZE);

  T = (trs) malloc(sizeof(TRS));
  if (T == NULL) {
      fprintf(stderr, "Error allocating %d bytes\n", (int)sizeof(TRS));
      exit(100);
  }
  T->vector = (cell) malloc(Job->vlength * sizeof(CELL));
  for (i=0; i<Job->vlength; i++) {
    T->vector[i].cvals = 0;
    T->vector[i].last_flip = lrand48()%LAST_FLIP_INIT;
  }
  T->scomvec = (set*) malloc(Job->vlength * sizeof(set));
  for (i=0; i<Job->vlength; i++)
    T->scomvec[i] = S_new(1,0);
  T->icomvec = (int*) malloc(Job->vlength * sizeof(int));
  T->inorder = (int*) malloc(Job->vlength * sizeof(int));
  T->coinorder = (int*) malloc(Job->vlength * sizeof(int));
  T->priority = (int*) malloc((Job->vlength+1) * sizeof(int));
  T->best = (int*) malloc(Job->vlength * sizeof(int));
  T->awnty_past = (int*) malloc(Job->vlength * sizeof(int));
  T->awnty_future = (int*) malloc(Job->vlength * sizeof(int));
  T->awnty_sofar = (int*) malloc(Job->vlength * sizeof(int));
  T->job = Job;
  T->max_batches = T->job->batches - 1;
  T->stakmax = (Job->maxref < 4? 4: Job->maxref) * Job->batches;
  report_initial(T);
  termination_initial(T);
  T->cutptr = 0;
  T->cut_flag = NOCUT;
  T->absolute = true;
  T->violations = 0;
  T->sdc_violations = 0;
  T->viol_list = 0;
  T->configtestcell = -2;         /* Any impossible value */
  T->configtestval = 0;
  T->configstart = false; 
  for (i=0; i < T->job->vlength; i++) T->priority[i] = 0;
  for (i=0; i < SURJ_MAX; i++) {
    T->surjection[i].valset = 0;
    T->surjection[i].cellset = 0;
    T->surjection[i].last_seen = 0;
  }
  for (i=0; i<T->job->batches; i++) T->stak[i] = 0;
  for (i=0; i<T->stakmax; i++) T->il_stak[i] = 0;
  for (i=0; i<T->stakmax+T->job->batches; i++) T->rl_stak[i] = 0;
  T->stak_ptr = 0;
  T->il_free = 0;
  T->rl_free = 0;
  T->best_penalty = Job->max_penalty+1;
  for (i=0; i<T->job->vlength; i++) {
    T->awnty_past[i] = 0;
    T->awnty_future[i] = 0;
    T->awnty_sofar[i] = 0;
  }
  T->flips = 0;
  T->this_cell = T->susguy = T->firstchange = T->forkpoint = 0;
  T->maxval = T->unfixt = 0;

  return T;
}




/******************************************************************************
*
*  This is where the going gets serious. The search function calls itself
*  recursively to search the subspaces determined by setting the value of
*  the picked cell successively in each possible way. Value wy is inserted
*  in cell ex. If the picked cell is -1 this means the vector is complete
*  and should be tested.
*
*  When the program is re-constituting its search after a "cut", it has to 
*  take account of the values that cells had before the cut. This is done 
*  by attending to the starting point for the loop of assignments to wy.
*
* Return the number of nodes in the subtree rooted at this node.
*/

int search(trs T, int dpth, int bdpth)
{
  int choice;
  int rc = 1;

  if (!T->job->done && !T->novect) {
    choice = pick_cell(T);
    if (choice < 0)
      test_possible_solution(T);
    else
      rc = make_search_tree_node(T,dpth,bdpth,choice);
    do_dumps(T,dpth,bdpth,choice,rc);
  }
  return rc;
}


int remaining(cell c, int wyord[], int bound)
{
  int x;
  int count = 0;

  for (x=0; x < bound; x++)
    if (!(c->cvals[wyord[x]].blocks))
      count++;
  return count;
}



/*
*  The next routine picks the next cell to have its value fixed.
*  If there is a cell with only one possible value, that is picked.
*  Otherwise, pick the cell with the smallest number of possible 
*  values among those with the highest priority.
*  If there is no open cell, return -1.
*
*  Things are a little different when there is an active cut, since
*  the change order up to the cut point must be reproduced exactly
*  in order to maintain integrity.
*/

int pick_cell(trs T)
{
  int rc = -1;             /* Return code               */
  int smallest = F_MAXINT; /* Cardinality of cell rc    */
  int best = 0;            /* Priority of cell rc       */
  int rarest = F_MAXINT;   /* Times cell rc picked      */
  int penalty = 0;         /* Least penalty for cell rc */
  int i;
  
  if (T->cutptr && T->cutptr->active) {
    for (i = 0; i < T->job->vlength && T->inorder[i] >= 0; i++);
    if (i < T->job->vlength && (rc = T->cutptr->inorder[i]) >= 0) {
      return set_fork(T, rc);
    }
    delete_cut(T);
  }
  
  if (T->cutptr && T->cutptr->active) {
    for (i=0; i<T->job->vlength; i++)
      if (T->cutptr->inorder[i] >= 0 &&
	  !(T->vector[T->cutptr->inorder[i]].has_value))
	return T->cutptr->inorder[i];
    delete_cut(T);
  }

  if (T->configstart) {
    T->configstart = false;
    return T->configtestcell;
  }

  while (T->unfixt >= 0 && T->vector[T->unfixt].has_value) T->unfixt--;
  if (T->unfixt < 0) return -1;
  
  for (i=T->unfixt; i >= 0 && smallest > 1; i--)
    if (!(T->vector[i].has_value)) {
      if (rc < 0 ||
	  T->vector[i].poss == 1 ||
	  (T->job->configheuristic && T->job->configurable[i] &&
	   (!T->job->configurable[rc] || T->vector[rc].allknown) &&
	   !T->vector[i].allknown) ||
	  ((!T->job->configurable[i] || T->job->configurable[rc] ||
	    !T->job->configheuristic ||
	    T->vector[i].allknown == T->vector[rc].allknown) &&
	   (T->priority[i] > best ||
	    (T->priority[i] == best &&
	     (T->vector[i].poss < smallest ||
	      (T->vector[i].poss == smallest &&
	       (T->vector[i].least_penalty > penalty ||
		(T->vector[i].least_penalty == penalty &&
		 T->vector[i].times_picked < rarest)))))))) {
	rc = i;
	smallest = T->vector[i].poss;
	best = T->priority[i];
	rarest = T->vector[i].times_picked;
	penalty = T->vector[i].least_penalty;
      }
    }
  if (rc < 0) {
    EVENT(abortSearch,T->job);
    skipout("Error in picking cell",SKIP);
  }

  if (smallest == 1) {
    if (T->absolute) T->vector[rc].fixed = true;
    T->job->branches[1]++;
    return rc;
  }

  T->vector[rc].times_picked++;
  if (smallest > T->vector[rc].csiz && !T->job->done) {
    EVENT(abortSearch,T->job);
    skipout("Invalid branching factor",SKIP);
  }
  T->job->branches[smallest]++;
  if (smallest > 1 && T->absolute) {
    T->absolute = false;
    print_constraint_graph(T);
  }
  return set_fork(T,rc);
}


/*
*  Set_fork is a subroutine of pick_cell. It sets the forkpoint just 
*  before returning the return code.
*/

int set_fork(trs T, int rc)
{
  if (T->forkpoint < 0 && T->vector[rc].poss > 1)
    T->forkpoint = rc;
  return rc;
}



void test_possible_solution(trs T)
{
  int i;

  if (T->job->Test) {
    EVENT(enterTest,T->job);
    if ((*T->job->Test)(T->icomvec,T)) {
      EVENT(solution,T->job);
      T->job->reps[SOLUTIONS].i_report++;
      for (i=0; i<T->job->vlength; i++)
	T->vector[i].cvals[T->icomvec[i]].known_possible = true;
      T->firstchange = T->job->vlength;
      if (T->job->cgv_enabled)
	cgv_state_report(T,0,0);
      if (T->job->configmode)
	T->job->done = true;
    }
    else {
      setref(T);
      T->job->reps[BAD_TESTS].i_report++;
    }
    EVENT(leaveTest,T->job);
  }
  else job_done(T->job,ASSIGNMENT_OK);

}


int make_search_tree_node(trs T, int dpth, int bdpth, int ex)
{
  int rc = 1;
  cell c;
  int wy;
  int jay;
  int kay = -1;
  int *wyord = 0;
  boolean inok;
  boolean ccut = false;
  boolean fch = false;
  boolean nochoice = false;
  void timesup();

  c = T->vector + ex;
  if (c->poss == 1) {
    nochoice = true;
    EVENT(unitPropagate,T->job);
  }
  else EVENT(enterNode,T->job);
  if (T->job->awnty_enabled) {
    T->awnty_past[bdpth] = 0;         /* Number of subtrees searched */
    T->awnty_future[bdpth] = c->poss; /* Number (possibly) to be searched */
    T->awnty_sofar[bdpth] = 0;        /* Cumulative subtree size */
  }
  if (timeout(T)) timesup();
  S_empty(T->scomvec[ex]);
  wyord = (int*) malloc((c->csiz+1) * sizeof(int));
  jay = get_possible_values(T,ex,wyord);
  while (set_wy(T,wyord,jay,ex,bdpth,&kay,&wy)) {
    inok = invalue(T,c,wy);
    if (T->cut_flag == FOUND && T->forkpoint >= 0) record_cut(T);
    else if (inok && !T->ts)
      rc += recursive_search_call(T,nochoice,dpth,bdpth);   /* Search */
    if (T->ts && !T->job->done) totsus(T);
    if (T->cut_flag == RECORDED && ex == T->forkpoint) ccut = true;
    if (T->coinorder[ex] <= T->firstchange) fch = true;
    if (fch) T->firstchange = T->coinorder[ex];
    outvalue(T,c);
    if (T->job->done) {
      if (wyord) Free(wyord);
      if (!nochoice) EVENT(leaveNode,T->job);
      return rc;
    }
    if (T->job->awnty_enabled)
      T->awnty_future[bdpth] = remaining(c,wyord+kay+1,jay-(kay+1));
    if (T->ts && T->vector[T->susguy].poss) T->ts = false;
    if (ccut) complete_cut(T);
    if (T->newref) {
      if (c->used) {
	addref(ex,T);
	if (T->ts) totsus(T);
      }
      else break;
    }
    if (fch && T->job->forbidden[ex]) break;
  }
  if (wyord) Free(wyord);
  if (!nochoice) EVENT(leaveNode,T->job);
  return rc;
}


int get_possible_values(trs T, int ex, int wyord[])
{
  int jay = 0;
  int wy;
  cell c = T->vector+ex;

  if (T->cutptr && T->cutptr->active && T->cutptr->icomvec[ex] < 0)
    delete_cut(T);
  if (T->cutptr && T->cutptr->active) {
    S_copy(T->cutptr->valsleft[ex],T->scomvec[ex]);
    wyord[jay++] = T->cutptr->icomvec[ex];
    for (wy=0; wy < c->csiz; wy++)
      if (wy != wyord[0] && S_is_member(wy,T->scomvec[ex]))
	wyord[jay++] = wy;
  }
  else
    for (wy=0; wy < c->csiz; wy++)
      if (!(c->cvals[wy].blocks)) {
	wyord[jay++] = wy;
	S_into(wy,T->scomvec[ex]);
      }
  wyord[jay] = -1;
  if (T->job->random) {
    if (T->cutptr && T->cutptr->active)
      randomise(wyord+1,jay-1);          /* First choice is determined */
    else
      randomise(wyord,jay);              /* All choices up for grabs */
  }
  if (T->job->configmode) {
    if (ex == T->configtestcell) {
      wyord[0] = T->configtestval;
      jay = 1;
      S_singleton(wyord[0],T->scomvec[ex]);  /* OK: set already assigned */
    }
    else if (T->job->configheuristic && T->job->configurable[ex])
      unknowns_first(wyord,jay,c->cvals); /* Move unknown values up */
  }
  return jay;
}


boolean set_wy(trs T, int wyord[], int jay, int ex, int bdpth,
	       int *kay, int *wy)
{
  cell c = T->vector+ex;

  if (T->novect || T->cut_flag)
    return false;
  while (++(*kay) < jay) {
    if (T->job->dump_rate) dump_trace(T->job,false);
    greedy_order(wyord,jay,*kay,c,T);     /* Lowest penalty value next */
    *wy = wyord[*kay];
    S_outof(*wy,T->scomvec[ex]);
    if (!(c->cvals[*wy].blocks)) {
      if (T->job->cgv_enabled && *kay<jay-1)
	cgv_state_report(T,c,*wy);
      if (T->cutptr && T->cutptr->active 
	  && *wy != T->cutptr->icomvec[ex]) delete_cut(T);
      return true;
    }
    else if (T->job->awnty_enabled)
      T->awnty_future[bdpth]--;
  }
  return false;
}


/*
 * Call search on the next subtree, and return the number of nodes
 */

int recursive_search_call(trs T, boolean nochoice, int dpth, int bdpth)
{
  int subtree;

  if (nochoice)
    return search(T,dpth+1,bdpth);
  else {
    subtree = search(T,dpth+1,bdpth+1);
    if (T->job->awnty_enabled) {
      T->awnty_sofar[bdpth] += subtree;
      T->awnty_past[bdpth]++;
      T->awnty_future[bdpth]--;
    }
    return subtree;
  }
}


void do_dumps(trs T, int dpth, int bdpth, int choice, int rc)
{

  if (T->job->awnty_enabled) T->awnty_future[bdpth] = 0;
  if (rc == 1) {
    if (T->job->dumpdepth) dump_depth(dpth);
    if (T->job->awnty_enabled)
      dump_awnty(T->awnty_past,T->awnty_future,T->awnty_sofar,bdpth,1);
  }
  else if (T->job->awnty_enabled && T->vector[choice].poss > 1)
    dump_awnty(T->awnty_past,T->awnty_future,T->awnty_sofar,bdpth,0);
}


/*
 * Shuffle an array into random order. If bound<2 this does nothing.
 */

void randomise(int arr[], int bound)
{
  int i,j,k,r,b;
  int *temporary = (int*) malloc(bound * sizeof(int));

  for (i=0; i<bound; i++) temporary[i] = F_MAXINT;
  for (i=0,b=bound; i<bound; i++) {
    r = lrand48() % b;
    for (j=0,k=0; j<bound; j++) {
      if (temporary[j]==F_MAXINT) {
	if (k==r) {
	  temporary[j] = arr[i];
	  b--;
	  break;
	}
	else k++;
      }
    }
  }
  for (i=0; i<bound; i++) arr[i] = temporary[i];
  Free(temporary);
}



/*
 * Jump the cell with lowest penalty up to the next array position.
 * If there are no soft constraints, this has no effect.
 * If there are, it tends to work against randomness by breaking ties.
 */

void greedy_order(int arr[], int bound, int pos, cell c, trs T)
{
  int x,y,z;

  if (T->cutptr && T->cutptr->active) return;
  x = pos;
  z = c->cvals[arr[x]].penalty;
  for (y = pos+1; y < bound && arr[y] >= 0; y++)
    if (c->cvals[arr[y]].penalty < z) {
      x = y;
      z = c->cvals[arr[x]].penalty;
    }
  if (x > pos) {
    y = arr[pos];
    arr[pos] = arr[x];
    arr[x] = y;
  }
}


/*
 * Put the values which are not yet known to be possible before those
 * which are known to be possible, keeping the internal order of each
 * sub-list unclanged
 */

void unknowns_first(int wyord[], int length, cellvalue cv)
{
  int *knowns = (int*) malloc(length * sizeof(int));
  int *unknowns = (int*) malloc(length * sizeof(int));
  int x, kx, ux;

  for (x=0,kx=0,ux=0; x < length; x++)
    if (cv[wyord[x]].known_possible)
      knowns[kx++] = wyord[x];
    else
      unknowns[ux++] = wyord[x];
  for (x=0; x<ux; x++)
    wyord[x] = unknowns[x];
  for (x=0; x<kx; x++)
    wyord[ux+x] = knowns[x];
  if (knowns) Free(knowns);
  if (unknowns) Free(unknowns);
}



/*
*  Old cuts are trashed to stop them interfering with the search.
*/

void delete_cut(trs T)
{
  cutt ftc;
  int x;

  ftc = T->cutptr;
  T->cutptr = ftc->lastone;
  Free(ftc->icomvec);
  Free(ftc->inorder);
  for (x=0; x<T->job->vlength; x++) {
    S_free(ftc->possvals[x]);
    S_free(ftc->valsleft[x]);
  }
  Free(ftc->possvals);
  Free(ftc->valsleft);
  Free(ftc);
}



/*
*  The next few functions have to do with allocation and de-allocation of
*  refutations, refutation-list entries and integer-pairs. Some of this is
*  a little intricate, but none of it is deep.
*
*  First, to get a refutation list entry.
*/

reflist get_rl(trs T)
{
  reflist r;

  if (!T->rl_free->nextref) more_rls(T);
  r = T->rl_free;
  T->rl_free = T->rl_free->nextref;
  r->nextref = r->lastref = T->rl_free->lastref= 0;
  return r;
}


/*
*  Conversely, to put a dead one to rest.
*/

void put_rl(trs T, reflist dead)
{
  if (!dead) return;
  dead->rf = 0;
  if (dead->lastref) dead->lastref->nextref = dead->nextref;
  if (dead->nextref) dead->nextref->lastref = dead->lastref;
  dead->lastref = 0;
  dead->nextref = T->rl_free;
  if (dead->nextref) dead->nextref->lastref = dead;
  T->rl_free = dead;
}


/*
*  Removing a refutation is more than just putting the stake through its 
*  heart. It may occur in several lists. Here is the routine to delete 
*  the reference to it from a given list and to kill that rl.
*/

void remove_rl(trs T, ref r, reflist *source)
{
  reflist rs;

  for (rs = *source; rs; rs = rs->nextref)
    if (rs->rf == r) {
      if (rs == *source) 
	*source = rs->nextref;
      put_rl(T, rs);
      return;
    }
}



/*
*  Now there is analogous stuff for integer-pair lists. When one of these
*  is allocated, it has its initial values kindly filled in here.
*/

intpair get_il(trs T, boolean b, int x, int y, intpair nx)
{
  intpair ip;

  if (!T->il_free->p) more_ils(T);
  ip = T->il_free;
  T->il_free = T->il_free->p;
  ip->positive = b;
  ip->v = x;
  ip->w = y;
  ip->p = nx;
  return ip;
}


/*
*  This is how to put down a pair of integers.
*/

void put_il(trs T, intpair dead)
{
  if (!dead) return;
  dead->v = -1;
  dead->p = T->il_free;
  T->il_free = dead;
}



/*
*  Getting a ref involves filling it with default values.
*/

ref get_ref(trs T)
{
  ref r;

  if (!T->stak_ptr->nxtr) more_refs(T);
  r = T->stak_ptr;
  T->stak_ptr = T->stak_ptr->nxtr;
  r->cardinality = 0;
  r->softweight = 0;
  r->notop = false;
  r->cellvals = 0;
  r->last_flip = lrand48()%LAST_FLIP_INIT;
  r->violated = false;
  T->job->reps[MEMORY].i_report++;
  T->local_memory++;
  if (batches_used(T) > T->max_batches) {
    printf("%d batches used, so triggering cut\n",batches_used(T));
    T->cut_flag = FOUND;
  }
  return r;
}



/*
*  And losing it again involves first getting rid of the cell-value pairs
*  which constitute the refutation proper. For this, call recursive put_il.
*/

void put_ref(trs T, ref dead, boolean flag)
{
  if (dead) {
    rec_put_il(T, dead->cellvals, dead, flag);
    dead->cardinality = -1;
    dead->nxtr = T->stak_ptr;
    T->stak_ptr = dead;
    T->job->reps[MEMORY].i_report--;
    T->local_memory--;
  }
}


/*
*  Here's the recursive put_il, which deletes a whole list of integer pairs.
*/

void rec_put_il(trs T, intpair ip, ref r, boolean flag)
{
  if (!ip) return;

  rec_put_il(T, ip->p, r, flag);
  if (flag) {
    if (ip->positive) 
      remove_rl(T, r, &(T->vector[ip->v].cvals[ip->w].posrefs));
    else
      remove_rl(T, r, &(T->vector[ip->v].cvals[ip->w].cvrefs));
    if (T->vector[ip->v].cvals[ip->w].back == r) {
      EVENT(abortSearch,T->job);
      skipout("Ghost in the back field",SKIP);
    }
  }
  put_il(T, ip);
}



/*
*  In order to know when to cut-and-guess, we need to know how many 
*  batches of NEWSLENGTH refutations have been used. So we count them!
*/

int batches_used(trs T)
{
  int i;
  
  for (i = 0; i <= T->job->batches; i++)
    if (!T->stak[i]) return i;
  EVENT(abortSearch,T->job);
  skipout("Too many refutations requested",BUST);
  return 0;
}



/*
*  The next three routines are for grabbing new chunks of memory to hold
*  refutations, rls or ils. The relevant data structure is an array of 
*  pointers to such chunks. That is, T->stak[x] is a pointer to a structure
*  which consists simply of an array of refs.
*/

void more_refs(trs T)
{
  int i;
  ref r;

  i = batches_used(T);
  T->stak[i] = (rbatch) malloc(sizeof(RBATCH));
  for (r = T->stak[i]->news; r-T->stak[i]->news < NEWSLENGTH; r++) {
    r->nxtr = (r == T->stak[i]->news+(NEWSLENGTH-1)? 0: r+1);
    r->cardinality = -1;
  }
  if (i) T->stak_ptr->nxtr = T->stak[i]->news;
  else T->stak_ptr = T->stak[i]->news;
  return;
}



/*
*  Getting more rls is pretty much like getting more refs.
*/

void more_rls(trs T)
{
  int i;
  reflist rp;

  for (i = 0; i <= T->stakmax+T->job->batches; i++)
    if (!T->rl_stak[i]) {
      T->rl_stak[i] = (rlbatch) malloc(sizeof(RLBATCH));
      for (rp = T->rl_stak[i]->news;
	   rp-T->rl_stak[i]->news < NEWSLENGTH; rp++) {
	rp->nextref = (rp == T->rl_stak[i]->news+(NEWSLENGTH-1)? 0: rp+1);
	rp->lastref = (rp == T->rl_stak[i]->news? 0: rp-1);
      }
      if (i) {
	T->rl_free->nextref = T->rl_stak[i]->news;
	T->rl_stak[i]->news->lastref = T->rl_free;
      }
      else {
	T->rl_free = T->rl_stak[i]->news;
	T->rl_free->lastref = 0;
      }
      return;
    }
}



/*
*  And getting more ils is similar.
*/

void more_ils(trs T)
{
  int i;
  intpair ip;

  for (i = 0; i <= T->stakmax; i++)
    if (!T->il_stak[i]) {
      T->il_stak[i] = (ilbatch) malloc(sizeof(ILBATCH));
      for (ip = T->il_stak[i]->news; ip-T->il_stak[i]->news < NEWSLENGTH; ip++)
	ip->p = (ip == T->il_stak[i]->news+(NEWSLENGTH-1)? 0: ip+1);
      if (i) T->il_free->p = T->il_stak[i]->news;
      else T->il_free = T->il_stak[i]->news;
      return;
    }
}



/*************************************************************************
*
*  Now back to the plot. The next few functions are where a healthy run of
*  any vntr program spends most of its time.
*
*  First we have the routine to insert a new value v in a cell c.
*/

boolean invalue(trs T, cell c, int v)
{
  reflist r;
  intpair ip;
  int i;

  if (T->job->done) return false;
  T->job->reps[INVALUES].i_report++;
  if (T->job->maxinvals &&
       T->job->reps[INVALUES].i_report >= T->job->maxinvals) {
    job_done(T->job,INSERTIONS);
    return false;
  }
  if (T->job->maxbak &&
       T->job->reps[NULL_BACKTRACKS].i_report +
       T->job->reps[SURJ_BACKTRACKS].i_report +
       T->job->reps[SOLUTIONS].i_report +
       T->job->reps[BAD_TESTS].i_report >= T->job->maxbak) {
    job_done(T->job,BACKTRACKS);
    return false;
  }
  c->has_value = true;
  T->icomvec[c->vpos] = v;
  T->inorder[++T->this_cell] = c->vpos;
  T->coinorder[c->vpos] = T->this_cell;

  for (r=c->cvals[v].cvrefs; r; r=r->nextref)
    transfer_ref(T,r->rf);

  for (i=0; i<c->csiz; i++) if (i != v)
    for (r=c->cvals[i].posrefs; r; r=r->nextref)
      transfer_ref(T,r->rf);

  for (r=c->inject; r; r=r->nextref)
    if (v < r->rf->cardinality) {
      FORR(r->rf,ip)
	if (!(T->vector[ip->v].has_value))
	  blockval(T,T->vector+ip->v,v,r->rf,NULL_BACKTRACKS);
    }
  /*
   * Because the value for c has just been fixed, the least (= only)
   * penalty for c is that for c->cvals[v], so update accordingly
   * and change T->total_penalty if necessary
   */

  if (c->least_penalty != c->cvals[v].penalty)
    find_min_penalty(c,T);
  if (T->total_penalty && T->total_penalty >= T->best_penalty)
    return false;

  return surjok(T);
}


/*
*  This is just a subroutine of invalue. It decrements the cardinality 
*  of a refutation. If the result is cardinality 1 then either there is 
*  a value inserted in some cell other than the value involved at that 
*  cell, in which case no action need be taken, or the refutation now 
*  blocks the insertion of a value. In the latter case, the value is 
*  marked as blocked and if this newly removes it the `poss' count for 
*  its cell is decremented. If this empties the possval set, we are 
*  ready to backtrack.
*/

void transfer_ref(trs T, ref rf)
{
  intpair ip, iq;
  int j;
  cell c_to;

  if (--(rf->cardinality) == 1) {
    iq = 0;
    FORR(rf, ip) {
      c_to = T->vector + ip->v;
      if (!(c_to->has_value)) {
        if (ip->positive) {
          for (j = 0; j < c_to->csiz; j++)
	    if (j != ip->w) {
	      if (rf->softweight) {
		add_penalty(ip->v,j,rf->softweight,T);
		if (c_to->least_penalty == 
		    c_to->cvals[j].penalty - rf->softweight)
		  find_min_penalty(c_to,T);
	      }
	      else
		blockval(T, c_to, j, rf, NULL_BACKTRACKS);
	    }
        }
        else {
	  if (rf->softweight) {
	    add_penalty(ip->v,ip->w,rf->softweight,T);
	    if (c_to->least_penalty == 
		c_to->cvals[ip->w].penalty - rf->softweight)
	      find_min_penalty(c_to,T);
	  }
	  else
	    blockval(T, c_to, ip->w, rf, NULL_BACKTRACKS);
	}
        if (iq) {
          iq->p = ip->p;
          ip->p = rf->cellvals;
          rf->cellvals = ip;
        }
        return;
      }
      else if ((ip->positive && T->icomvec[ip->v] == ip->w)
	       || (!ip->positive && T->icomvec[ip->v] != ip->w)) {
        if (iq) {
          iq->p = ip->p;
          ip->p = rf->cellvals;
          rf->cellvals = ip;
        }
        return;
      }
      iq = ip;
    }
  }
}



/*
*  The subroutine to block a value in the light of a transferred refutation
*  could be left in-line, but since it is called from three places, it is
*  better isolated like this.
*/

void blockval(trs T, cell c, int v, ref rrf, report_code RC)
{
  c->cvals[v].blocks++;
  if (c->cvals[v].blocks == 1) {
    c->poss--;
    if (!(c->poss) && !T->ts) {
      T->ts = true;
      T->susguy = c->vpos;
      if (RC) T->job->reps[RC].i_report++;
    }
    c->cvals[v].back = rrf;
    if (c->cvals[v].penalty == c->least_penalty)
      find_min_penalty(c,T);
  }
}



/*
*  Injective and surjective fragments of the vector are given special 
*  treatment. This routine checks the surjections to ensure that every 
*  value is still possible for at least one of the cells. The trick of
*  remembering the cell where the value was last seen and looking there
*  first keeps time spent on this to something rather small, and where 
*  surjectiveness is important, surjok pays its way handsomely.
*  
*/

boolean surjok(trs T)
{
  int i, j, ls;
  intpair ip;
  static boolean onoff;

  if (T->ts) return true;
  if (onoff && T->this_cell < T->job->vlength-3) {
    onoff = false;
    return true;
  }
  onoff = true;
  for (i=0; T->surjection[i].valset; i++) // For each surjection
    for (j = 0; j < T->maxval; j++)
      if (S_is_member(j,T->surjection[i].valset)) {        // For each value in it
	ls = T->surjection[i].last_seen[j];
	if (T->vector[ls].cvals[j].blocks || 
	     (T->vector[ls].has_value && T->icomvec[ls] != j)) {
	  FOR(T->surjection[i].cellset, ip)
	    if (j < T->vector[ip->v].csiz &&
		!(T->vector[ip->v].cvals[j].blocks) &&
		(!T->vector[ip->v].has_value || T->icomvec[ip->v] == j)) {
	      T->surjection[i].last_seen[j] = ip->v;
/* Try this */
	      ip->v = T->surjection[i].cellset->v;
	      T->surjection[i].cellset->v = T->surjection[i].last_seen[j];
/* for efficiency */
	      break;
	    }
	  if (!ip) {
	    T->job->reps[SURJ_BACKTRACKS].i_report++;
	    return false;
	  }
	}
      }
  return true;
}



void find_min_penalty(cell c, trs T)
{
  int local;
  int j;

  if (c->has_value) local = c->cvals[T->icomvec[c->vpos]].penalty;
  else {
    local = MAX_PENALTY;
    for (j=0; j< c->csiz; j++)
      if (!(c->cvals[j].blocks) && (c->cvals[j].penalty < local))
	local = c->cvals[j].penalty;
  }
  T->total_penalty += (local - c->least_penalty);
  c->least_penalty = local;
}



/*
 * Add up the minimum penalties possible for the cells. This is for the
 * branch and bound algorithm. Called from transref before searching.
 */

void set_total_penalty(trs T)
{
  int i;

  T->total_penalty = 0;
  for (i=0; i<T->job->vlength; i++) {
    T->vector[i].least_penalty = 0;
    find_min_penalty(T->vector+i,T);
  }
}



/*
*  The converse of putting a value in is taking one out. Why do I have to 
*  point that out to you? Note that these functions maintain T->inorder[x] 
*  (the x-th cell to get a value) and T->coinorder[x] (the y such that x 
*  is T->inorder[y]). These arrays are set to -1 where undefined.
*/

void outvalue(trs T, cell c)
{
  intpair ip;
  reflist r;
  int i;
  int val;

  if (!c->has_value)
    return;                       // Previous invalue failed

  val = T->icomvec[c->vpos];

  for (r=c->cvals[val].cvrefs; r; r=r->nextref)
    untransfer_ref(T,c,r->rf);

  for (i=0; i < c->csiz; i++) if (i != val)
    for (r=c->cvals[i].posrefs; r; r=r->nextref)
      untransfer_ref(T,c,r->rf);

  for (r=c->inject; r; r=r->nextref) if (val < r->rf->cardinality) {
      FORR(r->rf,ip)
	if (!(T->vector[ip->v].has_value))
	  unblockval(T,T->vector+ip->v,val,r->rf);
    }

  if (c->vpos == T->forkpoint && !T->job->configmode) {
    c->cvals[val].blocks++;
    if (c->cvals[val].blocks == 1)
      if (--c->poss == 1) T->forkpoint = -1;
  }
  c->has_value = false;
  find_min_penalty(c,T);
  T->icomvec[c->vpos] = 0;
  T->inorder[T->this_cell--] = -1;
  T->coinorder[c->vpos] = -1;
  if (c->vpos > T->unfixt) T->unfixt = c->vpos;
  delete_duds(T);
}


/*
*  The converse of transfer_ref. This unblocks any blocked cell after 
*  incrementing the cardinality of the refutation.
*/

void untransfer_ref(trs T, cell c, ref rf)
{
  int i;
  intpair ip;
  
  rf->cardinality++;
  if (rf->cardinality == 2 && !T->vector[rf->cellvals->v].has_value) {
    ip = rf->cellvals;
    if (ip->positive) {
      for (i=0; i<T->vector[ip->v].csiz; i++) 
	if (i != ip->w) {
	  if (rf->softweight) {
	    subtract_penalty(ip->v,i,rf->softweight,T);
	    if (T->vector[ip->v].least_penalty >
		T->vector[ip->v].cvals[i].penalty)
	      find_min_penalty(T->vector+ip->v,T);
	  }
	  else
	    unblockval(T,T->vector+ip->v,i,rf);
	}
    }
    else {
      if (rf->softweight) {
	subtract_penalty(ip->v,ip->w,rf->softweight,T);
	if (T->vector[ip->v].least_penalty >
	    T->vector[ip->v].cvals[ip->w].penalty)
	  find_min_penalty(T->vector+ip->v,T);
      }
      else
	unblockval(T,T->vector+ip->v,ip->w,rf);
    }
  }
  
  if (rf->notop) {
    FORR(rf,ip)
      if (ip->v != c->vpos && T->vector[ip->v].has_value)
	break;
    if (!ip) {
      mark_delete(T,rf);
      T->job->reps[TOPLESS].i_report++;
    }
  }
}



/*
*  Unblockval is the converse of blockval: re-transfer a refutation by 
*  deleting the block and adjusting the poss and back fields as necessary.
*/

void unblockval(trs T, cell c, int v, ref rrf)
{
  if (!(--c->cvals[v].blocks)) {
    c->poss++;
    if (c->least_penalty > c->cvals[v].penalty)
      find_min_penalty(c,T);
  }
  if (v == 0)
  if (c->cvals[v].blocks < 0 && !T->job->done) {
    EVENT(abortSearch,T->job);
    skipout("Negative blocks",SKIP);
  }
  if (c->cvals[v].back == rrf) c->cvals[v].back = 0;
}




/*
*  Refutations are deleted when they are no longer meaningful. Those
*  to be deleted are marked for death while other things are going on
*  and later executed all together.
*/

void mark_delete(trs T, ref r)
{
  reflist rr;

  rr = get_rl(T);
  rr->rf = r;
  rr->nextref = T->duds;
  if (rr->nextref) rr->nextref->lastref = rr;
  T->duds = rr;
}


void delete_duds(trs T)
{
  reflist delete_now;
  reflist delete_next;

  for (delete_now = T->duds; delete_now; delete_now = delete_next) {
    delete_next = delete_now->nextref;
    put_ref (T, delete_now->rf, true);
    put_rl(T, delete_now);
  }
  T->duds = 0;
}



/*
*  When a new refutation has been produced, it is sent to setref to
*  have any unnecessary parts removed and generally to be smartened up
*  for its entry into the database. First any cells which have
*  absolutely only one possible value are not really involved in the
*  refutation and can be removed. Secondly, if the refutation involves
*  the two or more most significant cells (not including the
*  absolutely fixed ones) then all but the least significant of these
*  can be deleted, since when the search backtracks far enough to
*  change that cell, this refutation can never again be active. In
*  such a case it is marked as topless so that outvalue knows to
*  delete it when its useful life is over. Finally, any refutation
*  which is unreasonably large occupies a lot of the database and
*  yields little information, so we can pretend it is a small
*  refutation until such time as this pretence is no longer safe, then
*  throw it away. We do this by treating it as though it involved all
*  of the cells more significant than a chosen one and applying the
*  "tops off" strategy above.
*/

void setref(trs T)
{
  int size_before, size_after, i, j;
  int latest, del;

  if (!T->newref) {
    T->novect = true;
    return;
  }

#ifdef DEBUG
  printf("Setref\n");
  report_constraint(T->newref);
  ref2ring(T,2,T->newref);
#endif
  size_before = size_after = T->newref->cardinality;

  /* Absolutely fixed cells do not count in the refutation */

  for (latest = 0;
       latest < T->job->vlength &&
	 T->inorder[latest] >= 0 &&
	 T->vector[T->inorder[latest]].fixed;
       latest++) {
    if (T->vector[T->inorder[latest]].used &&
	size_after > 2) {
      T->vector[T->inorder[latest]].used = false;
      size_after--;
    }
  }
#ifdef DEBUG
  ref2ring(T,3,T->newref);
#endif
  if (size_after < size_before)
    reduceref(T);
  if (!T->newref->cardinality) {
    T->novect = true;
    return;
  }
  size_before = size_after;
#ifdef DEBUG
  ref2ring(T,4,T->newref);
#endif

  /*
   * If the k+1 most significant cells are involved, the most
   * significant k can be deleted. At this point, T->inorder[latest]
   * should be the first non-fixed cell to have a value. Caveat: don't
   * reduce to a topless singleton, as that can be awkward for the
   * book-keeping.
   */

  for (del=-1;
       T->job->topsoff &&
	 size_after > 2 &&
	 latest < T->job->vlength &&
	 T->inorder[latest] >= 0 &&
	 T->vector[T->inorder[latest]].used;
       latest++) {
    if (del >= 0) {
      T->vector[T->inorder[del]].used = false;
      size_after--;
    }
    del = latest;
  }
#ifdef DEBUG
  ref2ring(T,5,T->newref);
#endif
  if (size_after < size_before) {
    reduceref(T);
    T->newref->notop = true;
  }
#ifdef DEBUG
  ref2ring(T,5,T->newref);
#endif

  /*
   * Just record the size before and after, for the stats
   */

  i = size_before;  j = size_after;
  if (i >= RRA) i = RRA-1;
  if (j >= RRB) j = RRB-1;
  T->job->ref_record[i][j]++;

  no_topless_singles(T);
  size_before = size_after = T->newref->cardinality;

  /*
   * Finally, if the refutation is still too big, lop its head off
   */

  if (!T->newref->softweight) {
    while (size_after > T->job->maxref) {
      if (T->vector[T->inorder[latest]].used) {
	if (del >= 0) {
	  T->vector[T->inorder[del]].used = false;
	  size_after--;
	}
	del = latest;
      }
      latest++;
    }
  }
#ifdef DEBUG
  ref2ring(T,6,T->newref);
#endif
  if (size_after < size_before) {
    reduceref (T);
    T->newref->notop = true;
  }
}



/*
*  This is the routine to take unused cell-value pairs out of a refutation.
*  It comes in two stages: one to take care of an unused starter pair and 
*  one for the other cases.
*/

void reduceref(trs T)
{
  intpair ip, iq;

  if (!T->newref->cellvals) return;
  while (!T->vector[T->newref->cellvals->v].used) {
    ip = T->newref->cellvals->p;
    put_il(T, T->newref->cellvals);
    T->newref->cellvals = ip;
    T->newref->cardinality--;
    if (!ip) {
      T->novect = true;
      return;
    }
  }

  FORR(T->newref, ip)
    while (ip->p && !(T->vector[ip->p->v].used)) {
      iq = ip->p->p;
      put_il(T, ip->p);
      ip->p = iq;
      T->newref->cardinality--;
    }
}



/*
*  Adding a refutation to the database is reasonably straightforward.
*  The call to this comes when backtracking has reached the point where
*  the least significant cell involved has just lost its value. The other
*  cells involved still have their values.
*/

void addref(int x, trs T)
{
  intpair ip;
  reflist rp;
  int i;
  int ipv, ipw;
  boolean ipb;

  if (!T->newref) {
    EVENT(abortSearch,T->job);
    skipout("Null refutation",NULL_REF);
  }
#ifdef DEBUG
  /*
  printf("Addref (x = %d)\n",x);
  report_constraint(T->newref);
  */
  ref2ring(T,1,T->newref);
#endif
  if (T->newref->cardinality==1 && T->newref->notop) {
    EVENT(abortSearch,T->job);
    skipout("Topless domain reduction shouldn't happen",SKIP);
  }
  if (!(T->newref->cardinality)) {
    T->novect = true;
    return;
  }
  T->job->reps[REFSIZE].i_report++;

  FORR(T->newref,ip)
    if (!ip->positive && T->vector[ip->v].cvals[ip->w].blocks) {
      if (T->newref->softweight)
	add_penalty(x,T->newref->cellvals->w,T->newref->softweight,T);
      FORR(T->newref,ip)
	T->vector[ip->v].used = false;
      put_ref(T,T->newref,false);
      T->newref = 0;
      return;
    }

  FORR(T->newref,ip) {
    T->vector[ip->v].used = false;
    rp = get_rl(T);
    rp->rf = T->newref;
    if (ip->positive) {
      rp->nextref = T->vector[ip->v].cvals[ip->w].posrefs;
      if (rp->nextref) rp->nextref->lastref = rp;
      T->vector[ip->v].cvals[ip->w].posrefs = rp;
    }
    else {
      rp->nextref = T->vector[ip->v].cvals[ip->w].cvrefs;
      if (rp->nextref) rp->nextref->lastref = rp;
      T->vector[ip->v].cvals[ip->w].cvrefs = rp;
    }
    if (ip->v == x) {
      if (ip->positive) {
        for (i=0; i<T->vector[x].csiz; i++) if (i != ip->w) {
	  if (T->newref->softweight)
	    add_penalty(x,i,T->newref->softweight,T);
	  else
	    blockval(T,T->vector+x,i,rp->rf,0);
	}
      }
      else {
	if (T->newref->softweight)
	  add_penalty(x,ip->w,T->newref->softweight,T);
	else
	  blockval(T,T->vector+x,ip->w,rp->rf,0);
      }
      ipv = ip->v;  ipw = ip->w;  ipb = ip->positive;
      ip->v = T->newref->cellvals->v;
      ip->w = T->newref->cellvals->w;
      ip->positive = T->newref->cellvals->positive;
      T->newref->cellvals->v = ipv;
      T->newref->cellvals->w = ipw;
      T->newref->cellvals->positive = ipb;
    }
    else T->newref->cardinality--;
  }
  if (T->newref->cardinality != 1) {
    EVENT(abortSearch,T->job);
    skipout("Refutation cardinality not correctly set",SKIP);
  }
  for (i=0; i<T->job->vlength; i++)
    if (T->vector[i].used) {
      EVENT(abortSearch,T->job);
      skipout("Used not reset",SKIP);
    }
  T->newref = 0;
}



void print_ref(ref r)
{
  intpair ip;

  if (!r) return;
  FORR(r,ip) printf(" %d:%d%c ",ip->v, ip->w, ip->positive? '*': ' ');
  if (r->notop) printf(" +");
  printf(" [%d]", r->softweight);
}


void print_sus(trs T, cell c)
{
  intpair ip;
  int i;

  puts("");
  for (i = 0; i < c->csiz; i++) {
    printf("    %d:%d",c->vpos,i);
    if (c->cvals[i].back)
      FORR(c->cvals[i].back,ip) if (ip->v == c->vpos)
	printf("%c  ",ip->positive? '*': ' ');
    if (c->cvals[i].back)
      FORR(c->cvals[i].back,ip) if (ip->v != c->vpos)
	printf("%d:%d%c  ", ip->v, ip->w, ip->positive? '*': ' ');
    puts("");
  }
}


void totsus(trs T)
{
  int i, addval;
  cell c;
  intpair ip, jp, kp = 0;
  boolean *pos_used = 0;

  if (T->job->Nosecs) {
    if (T->vector[T->susguy].poss) T->ts = false;
    return;
  }
  pos_used = (boolean*) malloc(T->job->vlength*sizeof(boolean));
  T->ts = false;
  c = T->vector + T->susguy;
#ifdef DEBUG
  if (0) print_sus(T,c);
#endif
  T->susguy = 0;
  T->newref = get_ref(T);
  for (i = 0; i<c->csiz; i++) if (c->cvals[i].back) {
    if (c->cvals[i].back->notop) T->newref->notop = true;
    FORR(c->cvals[i].back, ip) pos_used[ip->v] = false;
  }

  for (i = 0; i<c->csiz; i++) if (c->cvals[i].back)
    FORR(c->cvals[i].back, ip)
      if (ip->v != c->vpos) {
	if (ip->w < 0) {  /* The case of an injection */
	  if (T->vector[ip->v].has_value
	       && T->icomvec[ip->v] == i)
	    addval = i;
	  else addval = -1;
	}
	else addval = ip->w;
	if (ip->positive) pos_used[ip->v] = true;      
	if ((pos_used[ip->v] || !T->vector[ip->v].used)
	     && addval >= 0) {
	  jp = get_il(T, ip->positive, ip->v, addval, 0);
	  if (!T->newref->cellvals)
	    T->newref->cellvals = jp;
	  else kp->p = jp;
	  kp = jp;
	  T->vector[jp->v].used = true;
	  T->newref->cardinality++;
	}
      }
  make_coherent(T);

  if (T->newref->notop) {
    ip = T->newref->cellvals;
    FOR(ip->p, jp)
      if (T->coinorder[jp->v] < T->coinorder[ip->v])
	T->vector[jp->v].used = false;
      else {
	T->vector[ip->v].used = false;
	ip = jp;
      }
    reduceref(T);
    no_topless_singles(T);
  }
  else setref(T);
  if (pos_used) Free(pos_used);
#ifdef DEBUG
  if (0) {
    print_ref(T->newref); printf("\n");
  }
#endif
}


/*
 * Mainly for book-keeping reasons, singleton nogoods may not have
 * their tops removed. This function adds a dummy extra cell to such a
 * nogood, making its cardinality 2, so that it gets deleted smoothly
 * when the search backs up. Sorry if you thought the name of the
 * function was a bit naughty.
 */

void no_topless_singles(trs T)
{
  intpair ip;
  int i;

  T->newref->cardinality = 0;
  FOR(T->newref->cellvals,ip) T->newref->cardinality++;
  if (T->newref->cardinality == 1 && T->newref->notop) {
    if (T->inorder[0] == T->newref->cellvals->v)
      T->newref->notop = false;
    else {
      i = T->inorder[T->coinorder[T->newref->cellvals->v]-1];
      ip = get_il(T,false,i,T->icomvec[i],0);
      T->newref->cellvals->p = ip;
      T->vector[i].used = true;
      T->newref->cardinality++;
      reduceref(T);
    }
  }
}





/*
*  At present, the program is not set up to deal with two positive cell-value
*  pairs in a refutation which both pertain to the same cell. That is, we
*  cannot represent non-horn constraints. The following subroutine of totsus
*  checks for such difficulties and if they are present substitutes for the
*  offending set of pairs the negative one that the cell should not have its
*  current value.  */

void make_coherent(trs T)
{
  intpair ip, iq, ir;

  FORR(T->newref,ip) if (ip->positive) {
    while (T->newref->cellvals != ip
	    && T->newref->cellvals->v == ip->v) {
      ir = T->newref->cellvals->p;
      put_il(T, T->newref->cellvals);
      T->newref->cellvals = ir;
      T->newref->cardinality--;
      ip->w = -1;
    }
    for (iq = T->newref->cellvals; iq != ip && iq->p != ip; iq = iq->p)
      while (iq->p != ip && iq->p->v == ip->v) {
	ir = iq->p;
	iq->p = ir->p;
	put_il(T, ir);
	T->newref->cardinality--;
	ip->w = -1;
      }
    for (iq = ip; iq && iq->p; iq = iq->p)
      while (iq->p && iq->p->v == ip->v) {
	if (iq->p->w != ip->w) ip->w = -1;
	ir = iq->p;
	iq->p = ir->p;
	put_il(T, ir);
	T->newref->cardinality--;
      }
    if (ip->w < 0) {
      ip->positive = false;
      ip->w = T->icomvec[ip->v];
    }
  }
}





void new_small_ref(int cells[], int vals[], boolean valency[],
		    int softweight, trs T)
{
  ref the_ref;
  reflist rp, rrp, *tvc;
  int i, j, k, m;
  int refcard, maxrefcard;
  boolean b;

  /* Cludge at this stage: don't allow refutations with more than
     one positive member. This restriction can be removed once the
     basic program is running OK. */

  smallrefcounter++;
  for (i = 0; cells[i] >= 0; i++)
    if (valency[i])
      for (j = i+1; cells[j] >= 0; j++)
	if (valency[j]) {
	  if (cells[i] != cells[j] || vals[i] != vals[j]) {
	    EVENT(abortSearch,T->job);
	    skipout("Constraint is non-Horn",SKIP);
	  }
	}

  /* Check for subsumption by a one-refutation 
     and remove any fixed cell-value pairs. Then
     skip out of the search if the refutation is null */

  for (maxrefcard = 0; cells[maxrefcard] >= 0; maxrefcard++);
  for (i = 0; cells[i] >= 0; i++) {
    if (valency[i]) {
      if (T->vector[cells[i]].cvals[vals[i]].blocks) {
	for (j = i; cells[j] >= 0; j++) {
	  cells[j] = cells[j+1];
	  vals[j] = vals[j+1];
	  valency[j] = valency[j+1];
	}
	i--;  
      }
      else if (T->vector[cells[i]].poss == 1) return;
    }
    else {
      if (T->vector[cells[i]].cvals[vals[i]].blocks) return;
      if (T->vector[cells[i]].poss == 1) {
	for (j = i; cells[j] >= 0; j++) {
	  cells[j] = cells[j+1];
	  vals[j] = vals[j+1];
	  valency[j] = valency[j+1];
	}  
	i--;
      }
    }
  }
  if (!(refcard = i)) {
    if (!softweight) T->novect = true;
    return;
                                /* Impossible soft constraints are ignored */
  }

  /* Sort the refutation, removing repetitions. Remember that 
     refcard is likely to be about 2 or 3 (4 at most) so 
     the crudest sorting method is as good as anything!       */

  b = true; while (b) {
    b = false;
    for (i = 0; cells[i] >= 0 && cells[i+1] >= 0; i++)
      if (cells[i] > cells[i+1]) {
	b = true;
	j = cells[i];  k = vals[i];  m = valency[i];
	cells[i] = cells[i+1];  vals[i] = vals[i+1];
	valency[i] = valency[i+1];
	cells[i+1] = j;  vals[i+1] = k; valency[i+1] = m; 
      }
      else if (cells[i] == cells[i+1]) {
	if (vals[i] == vals[i+1]) {
	  if (valency[i] != valency[i+1]) return;  /* Tautology */
	  for (j = i; cells[j] >= 0; j++) {
	    cells[j] = cells[j+1];
	    vals[j] = vals[j+1];
	    valency[j] = valency[j+1];
	  }
	  i--;
	  refcard--;
	  b = true;
	}
	else {
	  if (!valency[i] && !valency[i+1]) return;  /* Tautology */
	  for (j = (valency[i]? i: i+1); cells[j] >= 0; j++) {
	    cells[j] = cells[j+1];
	    vals[j] = vals[j+1];
	    valency[j] = valency[j+1];
	  }
	  i--;
	  refcard--;
	  b = true;
	}
      }
  }

  if (refcard == 1) {
    if (valency[0]) {
      for (i = 0; i<T->vector[cells[0]].csiz; i++) if (i != vals[0])
	if (!T->vector[cells[0]].cvals[i].blocks) {
	  if (softweight) add_penalty(cells[0],i,softweight,T);
	  else remove_value(T, T->vector+cells[0], i);
	}
    }
    else {
      if (softweight) add_penalty(cells[0],vals[0],softweight,T);
      else remove_value(T, T->vector+cells[0], vals[0]);
    }
  }
  else if (!T->cut_flag) {
    the_ref = get_ref(T);
    the_ref->cardinality = refcard;
    the_ref->softweight = softweight;
    the_ref->cellvals = get_all_ils(cells, vals, valency, 0, T);
    if (subsumed(the_ref, T)) {
      put_ref(T, the_ref, false);
      T->job->reps[F_SUBSUMED].i_report++;
      return;
    }
    if (softweight && identical_ref(the_ref,T)) {
      put_ref(T, the_ref, false);
      T->job->reps[SOFT_DOUBLE].i_report++;
      return;
    } 
    T->job->reps[PREREFS].i_report++;
    for (i = 0; cells[i] >= 0; i++) {
      rp = get_rl(T);
      rp->rf = the_ref;
      if (valency[i])
	tvc = &(T->vector[cells[i]].cvals[vals[i]].posrefs);
      else tvc = &(T->vector[cells[i]].cvals[vals[i]].cvrefs);
      if (!*tvc || ((*tvc)->rf->cellvals->v != cells[i])
	   || the_ref->cellvals->v == cells[i]) {
	rp->nextref = *tvc;
	if (rp->nextref) rp->nextref->lastref = rp;
	*tvc = rp;
      }
      else {
	for (rrp=*tvc; rrp->nextref; rrp=rrp->nextref)
	  if (rrp->nextref->rf->cellvals->v != cells[i])
	    break;
	rp->nextref = rrp->nextref;
	if (rp->nextref) rp->nextref->lastref = rp;
	rrp->nextref = rp;
	rp->lastref = rrp;
      }
    }
    if (refcard<maxrefcard && !softweight) back_sub(T, the_ref);
  }
#ifdef DEBUG
  if (0) {              /* Insert more interesting conditon if desired */
    if (refcard==1) {
      the_ref = get_ref(T);
      the_ref->cardinality = refcard;
      the_ref->cellvals = get_all_ils(cells, vals, valency, 0, T);
    }
    report_constraint(the_ref);
  }
#endif
}



/*
 * Really remove a value because it is (absolutely) a singleton badguy
 */

void remove_value(trs T, cell c, int x)
{
  reflist rp;
  intpair ip, iq;
  int i;

  c->cvals[x].blocks++;
  if (c->cvals[x].blocks == 1) {
    if (!--c->poss) T->novect = true;
  }
  else {
    EVENT(abortSearch,T->job);
    skipout("Nonexistent value removed by 1-refutation",SKIP);
  }
  for (rp = c->cvals[x].cvrefs; rp; rp = rp->nextref) {
    FORR(rp->rf, ip) if (ip->v == c->vpos) break;
    if (!ip) {
      EVENT(abortSearch,T->job);
      skipout("Error in refutation",SKIP);
    }
    if (ip->positive) {
      if (ip == rp->rf->cellvals) rp->rf->cellvals = ip->p;
      else {
        FORR(rp->rf, iq)
	  if (iq->p == ip) break;
        iq->p = ip->p;
      }
      put_il(T, ip);
      remove_rl(T, rp->rf, &(c->cvals[x].cvrefs));
      rp->rf->cardinality--;
      if (rp->rf->cardinality == 1) {
        if (rp->rf->cellvals->positive) {
          for (i = 0; i < T->vector[rp->rf->cellvals->v].csiz; i++)
	    if (i != rp->rf->cellvals->w)
	      if (!T->vector[rp->rf->cellvals->v].cvals[i].blocks)
		remove_value(T, T->vector+rp->rf->cellvals->v, i);
        }
        else remove_value(T, T->vector+rp->rf->cellvals->v,
			   rp->rf->cellvals->w);
      }
      else back_sub(T, rp->rf);
    }
    else {
      mark_delete(T, rp->rf);
      T->job->reps[B_SUBSUMED].i_report++;
      T->job->reps[B1_SUBSUMED].i_report++;
      T->job->reps[PREREFS].i_report--;
    }
  }
  delete_duds(T);
  S_outof(x,T->scomvec[c->vpos]);
}




intpair get_all_ils(int cells[], int vals[], boolean valency[], int x, trs T)
{
  if (cells[x] < 0) return 0;
  return get_il(T, valency[x], cells[x], vals[x],
		 get_all_ils(cells, vals, valency, x+1, T));
}




boolean subsumed(ref the_ref, trs T)
{
  reflist rp;
  intpair ip;

  FORR(the_ref, ip) if (ip->p)
    for (rp = T->vector[ip->v].cvals[ip->w].cvrefs; rp; rp = rp->nextref) {
      if (rp->rf->cellvals->v != ip->v) break;
      if ((rp->rf->cardinality < the_ref->cardinality ||
	    (ip == the_ref->cellvals &&
	     rp->rf->cardinality == the_ref->cardinality))
	   && subsumes(rp->rf, the_ref)) return true;
    }
  return false;
}


boolean subsumes(ref little_ref, ref big_ref)
{
  intpair il, ib;

  if (little_ref->softweight) return false;
  ib = big_ref->cellvals;
  FORR(little_ref, il) {
    while (ib && (il->v != ib->v)) ib = ib->p;
    if (!ib) return false;
    if (ib->positive && !il->positive) return false;
    if (!ib->positive && il->positive) {
      if (il->w == ib->w) return false;
    }
    else if (il->w != ib->w) return false;
  }
  return true;
}


void back_sub(trs T, ref the_ref)
{
  reflist rp;
  intpair ip;

  ip = the_ref->cellvals;
  for (rp = T->vector[ip->v].cvals[ip->w].cvrefs; rp; rp = rp->nextref)
  if (rp->rf->cardinality > the_ref->cardinality)
  if (subsumes(the_ref, rp->rf)) {
    mark_delete(T, rp->rf);
    T->job->reps[B_SUBSUMED].i_report++;
    T->job->reps[PREREFS].i_report--;
  }
  delete_duds(T);
}


/*
 * Checks for a soft constraint involving exactly the same cells and
 * values. If one is found, adds the two softweights so that there is
 * no need to kep two copies of the same thing.
 */

boolean identical_ref(ref the_ref, trs T)
{
  reflist rp;
  intpair ip;

  if (!the_ref->softweight) return false;
  ip = the_ref->cellvals;
  for (rp = T->vector[ip->v].cvals[ip->w].cvrefs; rp; rp = rp->nextref)
    if (is_the_same(the_ref,rp->rf)) {
      rp->rf->softweight += the_ref->softweight;
      return true;
    }
  return false;
}


boolean is_the_same(ref the_new_ref, ref the_old_ref)
{
  intpair ipnew, ipold;

  if (!the_new_ref || !the_old_ref)
    return false;
  if (!the_new_ref->softweight || !the_old_ref->softweight)
    return false;
  if (the_new_ref->cardinality != the_old_ref->cardinality)
    return false;

  for (
       ipnew = the_new_ref->cellvals, ipold = the_old_ref->cellvals;
       ipnew || ipold;
       ipnew = ipnew->p, ipold = ipold->p
      )
    if (!ipnew || !ipold || 
	ipnew->v != ipold->v ||
	ipnew->w != ipold->w ||
	ipnew->positive != ipold->positive)
      return false;
  return true;
}




void Ref(int x, trs T)
{
  intpair it;

  if (T->vector[x].used) return;
  if (!(T->newref)) T->newref = get_ref(T);
  it = get_il(T, false, x, T->icomvec[x], T->newref->cellvals);
  T->newref->cellvals = it;
  T->newref->cardinality++;
  T->vector[x].used = true;
}



/*
*  This is done when a forkpoint exists and has a value, as do zero or
*  more less significant cells which next have to be outvalued.
*
*  If there is currently an active cut it must be deleted, as this one
*  subsumes it. Failure to do this may produce looping.
*/

void record_cut(trs T)
{
  if (T->cutptr && T->cutptr->active)
    delete_cut(T);
  record_part_cut(T->scomvec[T->forkpoint],T);
  T->cut_flag = RECORDED;
  T->ts = false;
  T->novect = true;
}



/*
*  Record the cuts for values from x up.
*/

void record_part_cut(set s, trs T)
{
  int i;
  cutt newcut;
  set s_reduced = 0;
  int x;

  if (!s || S_size(s) == 0) return;
  x = S_first_member(s,-1);
  s_reduced = S_copy(s,0);
  S_outof(x,s_reduced);
  record_part_cut(s_reduced, T);
  S_free(s_reduced);
  if (T->vector[T->forkpoint].cvals[x].blocks) return;
  newcut = (cutt) malloc(sizeof(CUTT));
  newcut->icomvec = (int*) malloc(T->job->vlength * sizeof(int));
  newcut->inorder = (int*) malloc(T->job->vlength * sizeof(int));
  newcut->possvals = (set*) malloc(T->job->vlength * sizeof(int));
  newcut->valsleft = (set*) malloc(T->job->vlength * sizeof(int));
  for (i=0; i <= T->job->vlength; i++) {
    newcut->valsleft[i] = 0;
    newcut->possvals[i] = 0;
  }
  newcut->lastone = T->cutptr;
  T->cutptr = newcut;

  if (x == T->icomvec[T->forkpoint])
    for (i = 0; i <= T->job->vlength; i++) {
      T->cutptr->icomvec[i] = T->icomvec[i];
      T->cutptr->inorder[i] = T->inorder[i];
      if (T->coinorder[i] < 0) T->cutptr->valsleft[i] = S_new(1,0);
      else {
	T->cutptr->valsleft[i] = S_copy(T->scomvec[i],0);
	S_into(T->icomvec[i],T->cutptr->valsleft[i]);
      }
    }
  else {
    for (i=0; i<=T->job->vlength; i++) {
      if (T->coinorder[i] < T->coinorder[T->forkpoint]) {
	T->cutptr->icomvec[i] = T->icomvec[i];
	T->cutptr->inorder[T->coinorder[i]] = i;
	T->cutptr->valsleft[i] = S_singleton(T->icomvec[i],0);
      }
      else {
	T->cutptr->icomvec[i] = -1;
	T->cutptr->valsleft[i] =
	  S_new(T->vector[i].csiz,T->cutptr->valsleft[i]);
      }
    }
    while (++i <= T->job->vlength);
  }
  T->cutptr->forkpoint = T->forkpoint;
  T->cutptr->active = false;
  T->cutptr->cutval = x;
  T->cutptr->valsleft[T->forkpoint] =
    S_singleton(x,T->cutptr->valsleft[T->forkpoint]);
}




/*
*  This is the action when the last cell in inorder to have a value 
*  is the forkpoint, and there is a cut being recorded.
*/

void complete_cut(trs T)
{
  int i, j;
  cutt cp;

  for (cp=T->cutptr; cp && cp->forkpoint == T->forkpoint; cp=cp->lastone) {
    for (i = 0; i < T->job->vlength; i++) {
      cp->possvals[i] = S_new(T->vector[i].csiz,cp->possvals[i]);
      for (j = 0; j < T->vector[i].csiz; j++)
	if (!T->vector[i].cvals[j].blocks)
	  S_into(j,cp->possvals[i]);
    }
    cp->possvals[T->forkpoint] =
      S_singleton(cp->cutval,cp->possvals[T->forkpoint]);
  }
  T->cut_flag = COMPLETED;
}



boolean prepared(trs T)
{
  int i, j;
  cell c;
  int actual_cellvals = 0;

  T->novect = false;
  if (T->cutptr) {
    for (i=0; i<T->job->vlength; i++)
      T->scomvec[i] = S_copy(T->cutptr->possvals[i],T->scomvec[i]);
  }
  else {
    (*(T->job->Sizes))(T->icomvec,T->job->vlength);
    T->maxval=0;
    for (i=0; i<T->job->vlength; i++) {
      if (T->icomvec[i] > T->maxval)
	T->maxval = T->icomvec[i];    /* USED TO BE T->icomvec[i]-1 */
      c = T->vector + i;
      c->vpos = i;
      c->inject = 0;
      c->csiz = T->icomvec[i];
      actual_cellvals += T->icomvec[i];
      if (T->job->max_cellvals && actual_cellvals > T->job->max_cellvals) {
	EVENT(abortSearch,T->job);
	skipout("Job too big: too many values for too many variables",SKIP);
      }
      c->cvals = (cellvalue) malloc(c->csiz * sizeof(CELLVALUE));
      for (j=0; j<c->csiz; j++) {
	c->cvals[j].penalty = 0;
	c->cvals[j].vpos = i;
	c->cvals[j].posrefs = c->cvals[j].cvrefs = 0;
	c->cvals[j].posrefs_eol = c->cvals[j].cvrefs_eol = 0;
	c->cvals[j].back = 0;
	c->cvals[j].blocks = 0;
	c->cvals[j].known_possible = false;
	c->cvals[j].last_flip = lrand48()%LAST_FLIP_INIT;
      }
      T->inorder[i] = -1;
      T->coinorder[i] = -1;
    }
    for (i=0; i<T->job->vlength; i++)
      S_fill(T->scomvec[i],T->vector[i].csiz);
  }

  if (T->job->Onerefs) {
    T->job->reps[SETPOSSES].i_report++;
    if (!(*(T->job->Onerefs))(T->scomvec,T)) {
      T->novect = true;
      return false;
    }
  }

  for (i = 0; i < T->job->vlength; i++) {
    if (S_size(T->scomvec[i]) == 0) {
      T->novect = true;
      return false;
    }
  }

  for (i = 0; i < T->job->batches; i++) T->stak[i] = 0;
  for (i = 0; i < T->stakmax; i++) T->il_stak[i] = 0;
  for (i = 0; i < T->stakmax+T->job->batches; i++) T->rl_stak[i] = 0;
  T->stak_ptr = 0;
  T->il_free = 0;
  T->rl_free = 0;
  more_refs(T);
  more_ils(T);
  more_rls(T);

  setinx(T);

  T->duds = 0;
  if (T->job->branches) Free(T->job->branches);
  T->job->branches = (int*) malloc((T->maxval+1) * sizeof(int));
  for (i=0; i <= T->maxval; i++) T->job->branches[i] = 0;
  T->job->maxval = T->maxval;
  return true;
}



void setinx(trs T)
{
  int i, j;
  cell c;

  for (i=0; i<SURJ_MAX; i++) {
    T->surjection[i].valset = 0;
    T->surjection[i].cellset = 0;
    T->surjection[i].last_seen = 0;
  }

  for (i=0; i<T->job->vlength; i++) {
    c = T->vector+i;
    c->poss = 0;
    for (j=0; j<c->csiz; j++) {
      if (S_is_member(j,T->scomvec[i])) {
        c->poss++;
        c->cvals[j].blocks = 0;
      }
      else {
	c->cvals[j].blocks = 1;
      }
    }
    c->used = false;
    c->has_value = false;
    c->times_picked = 0;
    c->fixed = false;
  }
}




boolean  pre_test(trs T)
{
  int i, j;

  if (!T->job->Smallrefs) return true;
  T->job->reps[DOSMALLS].i_report++;

  for (i=0; i<T->job->vlength; i++) {
    S_empty(T->scomvec[i]);
    for (j=0; j < T->vector[i].csiz; j++)
      if (!T->vector[i].cvals[j].blocks)
	S_into(j,T->scomvec[i]);
  }

  if (!(*(T->job->Smallrefs))(T->scomvec,T)) return false;
  if (!remove_singles(T)) return false;
  if (T->novect) return false;

  for (i=0; i<T->job->vlength; i++) {
    if (!(T->vector[i].poss)) {
      T->novect = true;
      return false;
    }
    S_empty(T->scomvec[i]);
    T->icomvec[i] = 0;
  }
  T->max_batches = batches_used(T) + T->job->extra_batches;
  if (T->max_batches >= T->job->batches)
    T->max_batches = T->job->batches - 1;
  return true;
}




void surjective(boolean vec[], set vset, trs T)
{
  int i, j;
  int count = 0;

  for (i=0; T->surjection[i].valset; i++)
    if (i == SURJ_MAX) {
      EVENT(abortSearch,T->job);
      skipout("Too many surjections",SURJES);
    }

  T->surjection[i].valset = S_copy(vset,T->surjection[i].valset);
  for (j=0; j<T->job->vlength; j++)
    if (vec[j]) {
      T->surjection[i].cellset =
	get_il(T,false, j,-1,T->surjection[i].cellset);
      count++;
    }
  T->surjection[i].last_seen = (int*) malloc(count * sizeof(int));
  for (j=0; j < T->maxval; j++)
    if (S_is_member(j,vset))
      T->surjection[i].last_seen[j] = T->surjection[i].cellset->v;
  /* Now each value has an arbitrary "witness" as a starter */
}


/*
 * Set up alldiff sections of the vector, for whatever propagator we
 * have. If local search is in operation, there are no propagators for
 * global constraints, so in that case alldiffs have to be decomposed
 * into sets of not-equals.
 *
 * Array vec (of length vlength) says which cells have to be different.
 * Integer x is the number of possible values (the same for all cells).
 */

void injective(boolean vec[], int x, trs T)
{
  int i, j, k;
  ref r;
  reflist rp;
  int cells[3] = {0,0,-1};
  int vals[3] = {0,0,-1};
  boolean valency[3] = {false,false,false};

  if (T->job->search_is_local) {
    for (i=1; i<T->job->vlength; i++)
      if (vec[i])
	for (j=0; j<i; j++)
	  if (vec[j]) {
	    for (k=0; k<x; k++) {
	      /* cells and vals have to be reset every time because
		 they can be changed in various ways by
		 new_small_ref(). valency should be OK. */
	      cells[0] = j; cells[1] = i;
	      vals[0] = vals[1] = k;
	      new_small_ref(cells,vals,valency,0,T);
	    }
	  }
  }
  else {
    r = get_ref(T);
    r->cardinality = x;
    r->notop = false;
    
    for (i=0; i<T->job->vlength; i++)
      if (vec[i]) {
	r->cellvals = get_il(T,false,i,-1,r->cellvals);
	rp = get_rl(T);
	rp->rf = r;
	rp->nextref = T->vector[i].inject;
	if (rp->nextref) rp->nextref->lastref = rp;
	T->vector[i].inject = rp;
      }
  }
}



void co_priority(int x, int y, trs T)
{
  T->priority[x] = y;
}




void job_done(trin t, termination_code tc)
{
  if (!t->configmode && t->report_breaks)
    fprintf(WARNING_STREAM,"\nStopped by %s\n", t->tcode[tc]);
  else
    strcpy(t->reps[STOP_CODE].s_report,t->tcode[tc]);
  t->reps[STOP_CODE].i_report = (int) tc;
  t->done = true;
}



void skipout(char s[], int exitcode)
{
  if (s) {
    fprintf(WARNING_STREAM,"\n\n\n Aborting on detection of an error");
    fprintf(WARNING_STREAM,(*s? ":": "."));
    fprintf(WARNING_STREAM,"\n\n %s.\n\n", s);
#ifdef DEBUG
    dump_refring();
#endif
  }
  exit(exitcode);
}


void test_backs(trs T)
{
  int i, j;
  cellvalue cv;

  for (i=0; i<T->job->vlength; i++)
  for (j=0; j < T->vector[i].csiz; j++) {
    cv = &(T->vector[i].cvals[j]);
    /*
    * if (cv->blocks && !cv->back)
    * skipout("Block without back",SKIP);
    */
    if (cv->back && !cv->blocks) {
      EVENT(abortSearch,T->job);
      skipout("Back without block",SKIP);
    }
  }
}



void add_penalty(int vp, int val, int x, trs T)
{
  T->vector[vp].cvals[val].penalty += x;
}



void subtract_penalty(int vp, int val, int x, trs T)
{
  T->vector[vp].cvals[val].penalty -= x;
}



void set_best(trs T)
{
  int i;

  for (i=0; i<T->job->vlength; i++) T->best[i] = T->icomvec[i];
  T->best_penalty = T->total_penalty;
}


/*
 * Perform unit propagation and unit subsumption with any cells having
 * only one value.
 */

boolean remove_singles(trs T)
{
  int x, y;
  reflist rp;
  boolean more_to_come = true;

  while (more_to_come) {
    more_to_come = false;
    for (x=0; x<T->job->vlength; x++)
      if (T->vector[x].poss == 1) {
	for (y=0; y<T->vector[x].csiz; y++)
	  if (T->vector[x].cvals[y].blocks) {
	    for (rp=T->vector[x].cvals[y].cvrefs; rp; rp=rp->nextref)
	      mark_delete(T,rp->rf);
	    for (rp=T->vector[x].cvals[y].posrefs; rp; rp=rp->nextref)
	      if (apply_unit_propagation(T,x,y,rp->rf))
		more_to_come = true;
	    while (T->vector[x].cvals[y].posrefs) {
	      rp = T->vector[x].cvals[y].posrefs->nextref;
	      put_rl(T,T->vector[x].cvals[y].posrefs);
	      T->vector[x].cvals[y].posrefs = rp;
	    }
	  }
	  else {
	    for (rp=T->vector[x].cvals[y].posrefs; rp; rp=rp->nextref)
	      mark_delete(T,rp->rf);
	    for (rp=T->vector[x].cvals[y].cvrefs; rp; rp=rp->nextref)
	      if (apply_unit_propagation(T,x,y,rp->rf))
		more_to_come = true;
	    while (T->vector[x].cvals[y].cvrefs) {
	      rp = T->vector[x].cvals[y].cvrefs->nextref;
	      put_rl(T,T->vector[x].cvals[y].cvrefs);
	      T->vector[x].cvals[y].cvrefs = rp;
	    }
	  }
	if (T->novect) return false;
	delete_duds(T);
      }
  }
  return true;
}


boolean apply_unit_propagation(trs T, int x, int y, ref r)
{
  boolean more_to_come = false;
  intpair ip, iq;
  int v, w, z;

  for(ip=r->cellvals,iq=0; ip; iq=ip,ip=ip->p)
    if (ip->v == x)
      break;
  if (!ip) {
    EVENT(abortSearch,T->job);
    skipout("Constraint has missing variable",SKIP);
  }
  if (ip == r->cellvals)
    r->cellvals = ip->p;
  else
    iq->p = ip->p;
  put_il(T,ip);
  r->cardinality--;
  if (r->cardinality == 0) {
    T->novect = true;
    return false;
  }
  if (r->cardinality == 1) {
    v = r->cellvals->v;
    w = r->cellvals->w;
    if (r->cellvals->positive) {
      for (z=0; z<T->vector[v].csiz; z++)
	if (z != w && !T->vector[v].cvals[z].blocks) {
	  T->vector[v].cvals[z].blocks = 1;
	  T->vector[v].cvals[z].back = 0;
	  T->vector[v].poss--;
	  if (T->vector[v].poss == 0) {
	    T->novect = true;
	    return false;
	  }
	  if (T->vector[v].poss == 1 && v < x)
	    more_to_come = true;
	}
    }
    else {
      if (!T ->vector[v].cvals[w].blocks) {
	T->vector[v].cvals[w].back = 0;
	T->vector[v].cvals[w].blocks = 1;
	T->vector[v].poss--;
	if (T->vector[v].poss == 0) {
	  T->novect = true;
	  return false; 
	} 
	if (T->vector[v].poss == 1 && v < x)
	  more_to_come = true;
      }
    }
  }
  return more_to_come;
}



void dump_depth(int d)
{
  printf("BT depth %d\n",d);
}


void dump_awnty(int past[], int future[], int sofar[], int depth, boolean term)
{
  int x;

  printf("awnty %d %d ",(term? 1: 0), depth+1);
  for (x=0; x<=depth; x++)
    printf(" %d %d %d ",past[x],future[x],sofar[x]);
  printf("\n");
}

#ifdef DEBUG
void init_init_refring(void)
{
  int x;

  for (x=0; x<dbringmax; x++)
    REFRING[x].r.cellvals = 0;
}

void init_refring(trs T)
{
  int x;

  for (x=0; x<dbringmax; x++) {
    REFRING[x].where = 0;
    REFRING[x].refid = 0;
    if (x == dbringmax-1) REFRING[x].next = REFRING;
    else REFRING[x].next = REFRING+x+1;
    if (x == 0) REFRING[x].last = REFRING+(dbringmax-1);
    else REFRING[x].last = REFRING+(x-1);
    rec_put_il(T, REFRING[x].r.cellvals, 0, false);
    REFRING[x].r.cellvals = 0;
    REFRING[x].r.cardinality = 0;
    REFRING[x].r.notop = false;
    REFRING[x].r.softweight = 0;
    REFRING[x].r.nxtr = 0;
  }
  refring = REFRING;
}

void ref2ring(trs T, int w, ref rf)
{
  refring = refring->next;
  refring->where = w;
  refring->refid = (int)rf;
  copy_ref(T,&(refring->r),rf);
}

void copy_ref(trs T, ref dest, ref source)
{
  if (!dest || !source) return;
  dest->cardinality = source->cardinality;
  dest->notop = source->notop;
  dest->softweight = source->softweight;
  dest->nxtr = 0;
  rec_put_il(T, dest->cellvals, 0, false);
  dest->cellvals = rec_copy_ils(T,source->cellvals);
}

intpair rec_copy_ils(trs T, intpair source)
{
  if (!source) return 0;
  return get_il(T,
		source->positive,
		source->v,
		source->w,
		rec_copy_ils(T,source->p));
}

void dump_refring(void)
{
  rec_dump_refring(refring->next,refring);
}

void rec_dump_refring(dbref current, dbref final)
{
  intpair ip;

  if (current->refid) {
    printf("%2d: REF %d = {", current->where, current->refid);
    for (ip=current->r.cellvals; ip; ip=ip->p) {
      printf(" ");
      print_un_vec_pos(ip->v);
      printf(" %c ", (ip->positive? '=': '#'));
      print_val_of_cell(ip->v,ip->w);
    }
    printf("}  card %d%s\n", current->r.cardinality, 
	   (current->r.notop? "(N)": ""));
  }
  if (current != final)
    rec_dump_refring(current->next,final);
}
#endif



/*
int sanitycheck(trs T)
{
  int x;
  int y;
  reflist r;

  for (x=0; x<T->job->vlength; x++)
    for (y=0; y<T->vector[x].csiz; y++) {
      for (r=T->vector[x].cvals[y].cvrefs; r; r=r->nextref) {
	if (!r->rf)
	  skipout("Null rf field in cvrefs",SKIP);
	if (r->lastref && r->lastref->nextref != r)
	  skipout("Incorrect pointer sequence 1",SKIP);
	if (r->nextref && r->nextref->lastref != r)
	  skipout("Incorrect pointer sequence 2",SKIP);
      }
      for (r=T->vector[x].cvals[y].posrefs; r; r=r->nextref) {
	if (!r->rf)
	  skipout("Null rf field in posrefs",SKIP);
	if (r->lastref && r->lastref->nextref != r)
	  skipout("Incorrect pointer sequence 3",SKIP);
	if (r->nextref && r->nextref->lastref != r)
	  skipout("Incorrect pointer sequence 4",SKIP);
      }
    }
}
*/


/*
void check_cell_for_sanity(cell c, trs T)
{
  int posscount = 0;
  int x;

  for (x=0; x<c->csiz; x++)
    if (!(c->cvals[x].blocks)) {
      if (c->cvals[x].back)
	skipout("Insanity: back without block",SKIP);
      else
	posscount++;
    }
  if (posscount != c->poss)
    skipout("Insanity: bad poss count",SKIP);
}
*/

/*
void check_sanity(trs T)
{
  int x;

  for (x=0; x<T->job->vlength; x++)
    check_cell_for_sanity(T->vector+x,T);
}

int dump_the_problem(trs T)
{
  int x, y;

  printf("Problem:\n");
  for (x=0; x<T->job->vlength; x++) {
    printf("v%d: ",x);
    if (T->vector[x].has_value)
      printf(" Value %d   ",T->icomvec[x]);
    for (y=0; y<T->vector[x].csiz; y++)
      if (T->vector[x].cvals[y].blocks == 0)
	printf(" %d",y);
    printf("\n");
  }
  exit(2);
}
*/

/*
 * Record the current vector of values (which may represent a
 * solution) for future reference. The recorded vector survives
 * between calls to transref.
 * 
 * cf set_last_model() in ftest.c
 */

void set_last_vector(int thisvector[], int vlength)
{
  int i;

  if (tr_last_vector) free(tr_last_vector);
  tr_last_vector = (int*) malloc(vlength * sizeof(int));
  tr_last_vlen = vlength;
  for (i=0; i<vlength; i++)
    tr_last_vector[i] = thisvector[i];
}

void turn_on_debugging(void)
{
  debugprinting = 1;
}
