#include "Fvntr.h"

#define AGE_LIMIT 1000
#define TRY_TO_IMPROVE 5


void sanity_check_viols(trs T);
#ifdef IN_FINDER
void set_last_model(int thisvector[]);
#endif


void localsearch(trs T)
{
  int alt = 0;
  int reps = 1;

  if (!T->job->done && !T->novect) {
    set_initial_assignment(T);
    while (still_searching(T)) {
      set_parameters(&alt,&reps);
      random_step(T,reps,alt);
    }
  }
  if (T->best_penalty == 0)
    test_possible_solution(T);
}


void set_parameters(int *alt, int *reps)
{
  static int counter;

  if (counter < 0) counter = 0;
  if (counter >= AGE_LIMIT) counter = 0; else counter++;
  *reps = TRY_TO_IMPROVE;
  *alt = !counter;
}



void set_initial_assignment(trs T)
{
  int x;
  int y;

  T->viol_list = 0;
  T->violations = 0;
  for (x=0; x<T->job->vlength; x++) T->icomvec[x] = -1;
  set_eols(T);
  for (x=0; x<T->job->vlength; x++) {    /* For each decision variable c */
    y = choose_random_value(T->vector+x,-1);
    local_setvalue(T,T->vector+x,y);
  }
  add_positive_violations(T);
  calculate_penalty(T);
  set_best(T);
  T->flips = 0;
}


boolean still_searching(trs T)
{
  if (T->novect || T->job->done)
    return false;
  if (T->violations == 0)
    return false;
  if (T->flips >= T->job->maxflips)
    return false;
  return true;
}


void random_step(trs T, int reps, int age)
{
  ref r;
  cell c;
  int v;
  int oldv;
  int oldviolations = T->violations;
  //  int x;
  //  INTPAIR IP;

  if (!T->job->done && !T->novect) {
    if (age) {
      r = choose_oldest_constraint(T);
      c = T->vector+choose_oldest_variable(T,r);
      c->times_picked++;
      oldv = T->icomvec[c->vpos];
      v = choose_oldest_value(T,c,oldv);
      local_setvalue(T,c,v);
      c->last_flip = c->cvals[v].last_flip = T->flips+LAST_FLIP_INIT;
      if (r) r->last_flip = T->flips+LAST_FLIP_INIT;
      calculate_penalty(T);
      if (T->total_penalty < T->best_penalty) {
	set_best(T);
#ifdef IN_FINDER
	set_last_model(T->icomvec);
#else
	set_last_vector(T->icomvec,T->job->vlength);
#endif
	//	printf("A: %d violations, badness %d   %d flips\n",
	//	       T->violations,T->total_penalty,T->flips+1);
	//	fflush(stdout);
      }
    }
    else do {
	r = choose_random_constraint(T);
	c = T->vector+choose_random_variable(T,r);
	c->times_picked++;
	oldv = T->icomvec[c->vpos];
	v = choose_random_value(c,oldv);
	local_setvalue(T,c,v);
	calculate_penalty(T);
	if (T->total_penalty < T->best_penalty) {
	  set_best(T);
#ifdef IN_FINDER
	  set_last_model(T->icomvec);
#else
	  set_last_vector(T->icomvec,T->job->vlength);
#endif
	  //	  printf("A: %d violations, badness %d   %d flips\n",
	  //		 T->violations,T->total_penalty,T->flips+1);
	  //	  if (T->violations==1) {
	  //	    if (T->viol_list) report_constraint(T->viol_list->rf);
	  //	    for (x=0; x<T->job->vlength; x++)
	  //	      if (T->vector[x].cvals[T->icomvec[x]].penalty) {
	  //		printf("Bad value: ");
	  //		IP.v = x;  IP.w = T->icomvec[x];  IP.positive = false;
	  //#ifdef IN_FINDER
	  //		report_literal(&IP);
	  //#else
	  //		report_cell_val(&IP);
	  //#endif
	  //	      }
	  //	  }
	  //	  fflush(stdout);
	}
	if (reps < 2 || T->violations <= oldviolations) {
	  c->last_flip = c->cvals[v].last_flip = T->flips+LAST_FLIP_INIT;
	  if (r) r->last_flip = T->flips+LAST_FLIP_INIT;
	  break;
	}
	local_setvalue(T,c,oldv);
	c->times_picked--;
	calculate_penalty(T);
	if (T->violations != oldviolations)
	  skipout("Strange violation count",SKIP);
      }
      while (reps-- && !T->job->done && !T->novect);
  }
  T->flips++;
  T->job->reps[FLIPS].i_report++;
}


/*
 * Return pointer to a randomly chosen violated constraint.
 * Return 0 if this will be a soft domain constraint.
 */

ref choose_random_constraint(trs T)
{
  int x;
  reflist rp;

  if (T->sdc_violations) {
    if (T->violations == T->sdc_violations) return 0;
    if (drand48() <= (T->sdc_violations*1.0)/T->violations)
      return 0;
  }
  if (T->violations > T->sdc_violations) {
    x = lrand48() % (T->violations-T->sdc_violations);
    for (rp=T->viol_list; rp; rp=rp->nextref) {
      if (x) x--;
      else return rp->rf;
    }
  }
  skipout("Failed to choose a constraint",SKIP);
  return 0;        // Unreachable code to kill warning message
}


ref choose_oldest_constraint(trs T)
{
  ref thisone = 0;
  reflist rp;
  int x;

  for (rp=T->viol_list; rp; rp=rp->nextref)
    if (!thisone || rp->rf->last_flip < thisone->last_flip)
      thisone = rp->rf;
  if (T->sdc_violations) {
    for (x=0; x<T->job->vlength; x++)
      if (T->vector[x].cvals[T->icomvec[x]].penalty &&
	  (!thisone || T->vector[x].last_flip < thisone->last_flip))
	return 0;
  }
  if (thisone) return thisone;
  skipout("Failed to choose a constraint",SKIP);
  return 0;      // Unreachable statement to kill warning message
}


int choose_random_variable(trs T, ref r)
{
  int x,y;
  intpair ip;

  if (r) {
    x = lrand48() % r->cardinality;
    FORR(r,ip) {
      if (x) x--;
      else return ip->v;
    }
  }
  else if (T->sdc_violations) {
    x = lrand48() % T->sdc_violations;
    for (y=0; y<T->job->vlength; y++)
      if (T->vector[y].cvals[T->icomvec[y]].penalty) {
	if (x) x--;
	else return y;
    }
  }
  skipout("Failed to choose a variable",SKIP);
  return 0;      // Unreachable statement to kill warning message
}


int choose_oldest_variable(trs T, ref r)
{
  int bestv, bestage = -1;
  intpair ip;
  int x;

  if (r) {
    FORR(r,ip)
      if (bestage < 0 || T->vector[ip->v].last_flip < bestage) {
	bestv = ip->v;
	bestage =  T->vector[ip->v].last_flip;
      }
  }
  else {
    for (x=0; x<T->job->vlength; x++)
      if (T->vector[x].cvals[T->icomvec[x]].penalty &&
	  (bestage < 0 || T->vector[x].last_flip < bestage)) {
	bestv = x;
	bestage = T->vector[x].last_flip;
      }
  }
  return bestv;
}


int choose_random_value(cell c, int current_value)
{
  int x, y, z;

  if (c->poss <= 1) {
    for (y=0; y<c->csiz; y++)
      if (!c->cvals[y].blocks)
	return y;
    return -1;
  }

  if (current_value < 0) x = c->poss;
  else x = c->poss - 1;
  z = lrand48() % x;
  for (y=0; y<c->csiz; y++)
    if (!c->cvals[y].blocks && y != current_value) {
      if (z) z--;
      else return y;
    }
  /* Cannot get to here, as z will reach zero eventually */
  return 0;
}


int choose_oldest_value(trs T, cell c, int current_value)
{
  int x;
  int bestval, bestage = -1;

  if (c->poss <= 1) {
    for (x=0; x<c->csiz; x++)
      if (!c->cvals[x].blocks)
	return x;
    return -1;
  }

  for (x=0; x<c->csiz; x++)
    if (!c->cvals[x].blocks && x != current_value) {
      if (bestage < 0 || c->cvals[x].last_flip < bestage) {
	bestval = x;
	bestage = c->cvals[x].last_flip;
      }
    }
  return bestval;
}


void calculate_penalty(trs T)
{
  reflist rp;
  int x;

  T->violations = 0;
  T->sdc_violations = 0;
  T->total_penalty = 0;
  for (rp=T->viol_list; rp; rp=(rp? rp->nextref: T->viol_list))
    if (rp->rf->violated) {
      T->violations++;
      if (rp->rf->softweight)
	T->total_penalty += rp->rf->softweight;
      else
	T->total_penalty += T->job->hardweight;
    }
    else {
      if (rp == T->viol_list)
	T->viol_list = rp->nextref;
      if (rp->lastref) {
	rp = rp->lastref;
	put_rl(T,rp->nextref);
      }
      else {
	put_rl(T,rp);
	rp = 0;
      }
    }
  for (x=0; x<T->job->vlength; x++)
    if (T->vector[x].cvals[T->icomvec[x]].penalty) {
      T->sdc_violations++;
      T->violations++;
      T->total_penalty += T->vector[x].cvals[T->icomvec[x]].penalty;
    }
  sanity_check_viols(T);
}



void local_setvalue(trs T, cell c, int v)
{
  cellvalue cv;
  reflist rp, rq;
  int oldv;

  if (v < 0) {
    T->novect = true;
    return;
  }
  oldv = T->icomvec[c->vpos];
  T->icomvec[c->vpos] = v;
  if (oldv >= 0) {
    cv = c->cvals+oldv;
    for (rp=cv->cvrefs; rp; rp=rp->nextref)
      rp->rf->violated = false;
    for (rp=cv->posrefs; rp; rp=rp->nextref)
      if (is_violated(T->icomvec,rp->rf)) {
	check_viol_list(T,rp->rf);
	rp->rf->violated = true;
	rq = get_rl(T);
	rq->rf = rp->rf;
	rq->nextref = T->viol_list;
	if (rq->nextref) rq->nextref->lastref = rq;
	T->viol_list = rq;
      }
  }
  cv = c->cvals+v;
  for (rp=cv->posrefs; rp; rp=rp->nextref)
    rp->rf->violated = false;
  for (rp=cv->cvrefs; rp; rp=rp->nextref)
    if (is_violated(T->icomvec,rp->rf)) {
      check_viol_list(T,rp->rf);
      rp->rf->violated = true;
      rq = get_rl(T);
      rq->rf = rp->rf;
      rq->nextref = T->viol_list;
      if (rq->nextref) rq->nextref->lastref = rq;
      T->viol_list = rq;
    }
}


/*
 * At the start of the search, after all the values for the first
 * assignment have been inserted, go through all refutations in the
 * entire problem and collect the violated ones. There is no need to
 * consider purely negative ones, as these will have been added by
 * local_setvalue().
 */

void add_positive_violations(trs T)
{
  int x, y;
  reflist r, temp;

  for (x=0; x<T->job->vlength; x++)
    for (y=0; y<T->vector[x].csiz; y++)
      for (r=T->vector[x].cvals[y].posrefs; r; r=r->nextref)
	if (!r->rf->violated &&
	    is_violated(T->icomvec,r->rf)) {
	  r->rf->violated = true;
	  temp = get_rl(T);
	  temp->rf = r->rf;
	  temp->nextref = T->viol_list;
	  if (temp->nextref) temp->nextref->lastref = temp;
	  T->viol_list = temp;
	}
}


/*
 * Decide whether the given vector of values violates the given
 * refutation. Since the refutastion is a clause, it is violated iff
 * every negative literal corresponds to a value in the vector and
 * every positive literal does not.
 */

boolean is_violated(int vec[], ref r)
{
  intpair ip;

  FORR(r,ip)
    if (vec[ip->v] < 0 ||
	(ip->positive && vec[ip->v] == ip->w) ||
	(!ip->positive && vec[ip->v] != ip->w))
      return false;
  return true;
}


/*
 * Set the end-of-list markers to point at the ends of the lists
 */

void set_eols(trs T)
{
  int x,y;
  cellvalue cv;

  for (x=0; x<T->job->vlength; x++)
    for (y=0; y<T->vector[x].csiz; y++) {
      cv = (T->vector[x].cvals)+y;
      for (cv->cvrefs_eol = cv->cvrefs;
	   cv->cvrefs_eol && cv->cvrefs_eol->nextref;
	   cv->cvrefs_eol = cv->cvrefs_eol->nextref) ;
      for (cv->posrefs_eol = cv->posrefs;
	   cv->posrefs_eol && cv->posrefs_eol->nextref;
	   cv->posrefs_eol = cv->posrefs_eol->nextref) ;
    }
}


/*
 * Check for violations of an alldiff constraint, given that #ch has
 * just changed from v1 to v2. Return the results in the list supp.
 */

reflist alldiff_violations(trs T, ref r, int ch, int v1, int v2, reflist supp)
{
  intpair ip;
  reflist rp, rq;

  for (rp=supp; rp; rp=(rp? rp->nextref: supp))
    if (rp->rf->cellvals->v == v1 || rp->rf->cellvals->p->v == v1) {
      put_ref(T,rp->rf,false);
      rq = rp->lastref;
      if (rp==supp) supp = rp->nextref;
      put_rl(T,rp);
      rp = rq;
    }
  FORR(r,ip)
    if (ip->v != ch && ip->w == v2) {
      rp = get_rl(T);
      rp->rf = get_ref(T);
      rp->rf->cellvals = get_il(T,false,ch,v2,get_il(T,false,ip->v,v2,0));
      rp->rf->cardinality = 2;
      rp->nextref = supp;
      if (supp) supp->lastref = rp;
      supp = rp;
    }
  return supp;
}


reflist initial_alldiff_violations(trs T, ref r)
{
  int *c = (int*) malloc(r->cardinality * sizeof(int));
  int *v = (int*) malloc(r->cardinality * sizeof(int));
  int x;
  intpair ip;
  reflist supp = 0;

  x=0;
  FORR(r,ip) {
    c[x] = ip->v;
    v[x++] = ip->w;
  }
  merge_sort(c,v,x);
  for (x=1; x<r->cardinality; x++)
    if (v[x] == v[x-1]) {
      if (supp) {
	supp->lastref = get_rl(T);
	supp->lastref->nextref = supp;
	supp = supp->lastref;
      }
      else supp = get_rl(T);
      supp->rf = get_ref(T);
      supp->rf->cellvals = 
	get_il(T,false,c[x-1],v[x],get_il(T,false,c[x],v[x],0));
      supp->rf->cardinality = 2;
    }
  if (c) free(c);
  if (v) free(v);
  return supp;
}


void merge_sort(int c[], int v[], int length)
{
  int *copyc = (int*) malloc(length*sizeof(int));;
  int *copyv = (int*) malloc(length*sizeof(int));;
  int x, y, z;
  int h1, h2;

  if (length > 1) {
    h1 = length/2;
    h2 = length-h1;
    for (x=0; x<length; x++) {
      copyc[x] = c[x];
      copyv[x] = v[x];
    }
    merge_sort(copyc,copyv,h1);
    merge_sort(copyc+h1,copyv+h1,h2);
    for (x=0,y=0,z=h1; x<length; x++) {
      if (y==h1) {
	c[x] = copyc[z];
	v[x] = copyv[z++];
      }
      else if (z==length) {
	c[x] = copyc[y];
	v[x] = copyv[y++];
      }
      else if (copyv[y] < copyv[z]) {
	c[x] = copyc[y];
	v[x] = copyv[y++];
      }
      else {
	c[x] = copyc[z];
	v[x] = copyv[z++];
      }
    }
  }
  free(copyc);
  free(copyv);
}

int check_viol_list(trs T, ref r)
{
  reflist rp;

  for (rp=T->viol_list; rp; rp=rp->nextref)
    if (rp->rf == r)
      skipout("Duplication in violations list",SKIP);
  return 0;
}


void sanity_check_viols(trs T)
{
  cell c;
  cellvalue cv;
  reflist rp;
  int refcount = 0;
  int sdc_refcount = 0;
  int x,y;
  intpair ip;
  boolean isfirst;

  for (x=0; x<T->job->vlength; x++) {
    c = T->vector+x;
    cv = c->cvals+T->icomvec[x];
    if (cv->penalty) {
      sdc_refcount++;
      refcount++;
    }
    for (rp=cv->cvrefs; rp; rp=rp->nextref) {
      isfirst = true;
      FORR(rp->rf,ip) if (ip->v < x) isfirst = false;
      if (isfirst) {
	if (is_violated(T->icomvec,rp->rf)) {
	  refcount++;
	  if (!rp->rf->violated)
	    skipout("Failed to flag violation",SKIP);
	}
	else if (rp->rf->violated)
	  skipout("Incorrectly flagged violation",SKIP);
      }
    }
    for (y=0; y<c->csiz; y++)
      for (rp=c->cvals[y].posrefs; rp; rp=rp->nextref) {
	isfirst = true;
	FORR(rp->rf,ip) if (ip->v < x) isfirst = false;
	if (isfirst) {
	  if (is_violated(T->icomvec,rp->rf)) {
	    refcount++;
	    if (!rp->rf->violated)
	      skipout("Failed to flag violation",SKIP);
	  }
	  else if (rp->rf->violated)
	    skipout("Incorrectly flagged violation",SKIP);
	}
      }
  }
  if (sdc_refcount != T->sdc_violations)
    skipout("Incorrect sdc violation count",SKIP);
  if (refcount != T->violations)
    skipout("Incorrect violation count",SKIP);
}
