make_number
make_compound

/*
 * A formula is a sequence of minimal formulae separated by binary
 * (infix) operators. Minimal formulae are parsed by sub_parse. The
 * flat list is made into a tree, obeying precedence conventions, by
 * make_compound. The terminating character (term) is either ")" or
 * "." depending on whether the call is from sub_parse or not.
 */

wff parse(char *term)
{
  wff_stack subs = 0;
  f_stack ops = 0;
  functor f;
  wff w;
  dynstring *s;
  char message[100];

  sprintf(message,"found where %s or a binary operator expected",term);
  for (;;) {
    subs = push_wff(sub_parse(),subs);
    next_word(s);
    if (!strcmp(s->c_str,term))
      break;
    f = s2f(s->c_str);
    if (f && f->adicity == 2)
      ops = push_op(f,ops);
    else
      inerror(s->c_str,message,0);
  }
  w = make_compound(subs,ops);
  free_wff_list(subs);
  free_op_list(ops);
  return w;
}


/*
 * Parse and return the smallest formula starting at the present point
 * in the input stream. All prefixed operators have minimal scope. The
 * arguments are a parenthesised, comma-separated list except in the
 * case of a quantifier, whose first argument must be a variable.
 */

wff sub_parse(void)
{
  dynstring *s;
  functor f;
  int x;
  wff *sb = 0;                 // Allocated here, freed elsewhere

  next_word(s);
  if (!strcmp(s->c_str,"("))
    return parse(")");
  f = s2f(s->c_str);                  // There is probably a function for this
  if (f) {
    if (f->adicity) {
      sb = (wff*) malloc(f->adicity * sizeof(wff));
      if (is_quantifier(s)) {
	sb[0] = sub_parse();
	if (sb[0]->symbol)
	  inerror("Expected variable after",s->c_str,0);
	sb[1] = sub_parse();
      }
      else {
	next_word(s);
	if (strcmp(s->c_str,"("))
	  inerror("Expected left parenthesis after",s->c_str,0);
	for (x=0; x<f->adicity; x++) {
	  sb[x] = sub_parse();
	  if (x < f->adicity-1) {
	    next_word(s);
	    if (strcmp(s->c_str,","))
	      inerror("Expected comma in argument list after",s->c_str,0);
	  }
	}
	next_word(s);
	if (strcmp(s->c_str,")"))
	  inerror("Expected right parenthesis after arguments of",s->c_str,0);
      }
    }
    return make_wff(f,sb);     // Also frees sb
  }
  if (isdigit(*(s->c_str)))
    return make_number(s->c_str);
  if (isalpha(*(s->c_str)))
    return make_var(s->c_str);
  inerror(s,"found where variable or numeral expected",0);
}



wff make_wff(functor f, wff sb[])
{
  wff w;
  int x;

  w = found_wff(f,sb);
  if (!w) {
    w = new_wff();
    w->sym = f;
    for (x=0; x<f->adicity; x++)
      w->sb[x] = sb[x];
  }
  if (sb) free(sb);
  return w;
}


wff make_var(char *s)
{
  return make_wff(newvar(s,0),0);
}


functor s2f(char *s)
{
  functor f;

  for (f=FN; f; f=f->nextfunct)
    if (!strcmp(s,f->fname))
      break;
  return f;
}


wff_stack push_wff(wff w, wff_stack dest)
{
  wff_stack ptr;

  if (dest) {
    for (ptr=dest; ptr->nxt; ptr=ptr->nxt) ;
    ptr->nxt = (wff_stack) malloc(sizeof(WFF_STACK));
    ptr = ptr->nxt;
  }
  else
    ptr = (wff_stack) malloc(sizeof(WFF_STACK));
  ptr->nxt = 0;
  ptr->w = w;
  return ptr;
}

f_stack push_op(functor f, f_stack dest)
{
  f_stack ptr;

  if (dest) {
    for (ptr=dest; ptr->nxt; ptr=ptr->nxt) ;
    ptr->nxt = (f_stack) malloc(sizeof(F_STACK));
    ptr = ptr->nxt;
  }
  else
    ptr = (f_stack) malloc(sizeof(F_STACK));
  ptr->nxt = 0;
  ptr->f = f;
  return ptr;
}

void free_wff_list(wff_stack ws)
{
  if (ws) {
    free_wff_list(ws->nxt);
    free(ws);
  }
}

void free_op_list(op_stack os)
{
  if (os) {
    free_op_list(os->nxt);
    free(os);
  }
}
